/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "LikelihoodCalculations.hpp"
#include "RandomNumberGenerator.hpp"

namespace TreeTime_Core
{

	TreeTime_Utilities::MemoryAllocator*
			LikelihoodCalculations::s_pMemoryAllocator = 0;
	map<int,realf****>				LikelihoodCalculations::s_mIdToMemory;
	int								LikelihoodCalculations::s_nLoci;
	vector<int>						LikelihoodCalculations::s_vnSites;
	vector<int>						LikelihoodCalculations::s_vnRates;
	int								LikelihoodCalculations::s_nStates;
	vector<vector<int> > 			LikelihoodCalculations::s_vvMultiplicities;
	vector<vector<vector<int> > >  LikelihoodCalculations::s_vvvInvariantStates;

	void LikelihoodCalculations::init(TreeTime_NexusProcessing::Data& p_Data)
	{
		s_mIdToMemory.clear();
		delete s_pMemoryAllocator;
		s_nLoci = p_Data.getInitialMolecularModels().size();
		s_vnSites.reserve(s_nLoci);
		for(int i=0; i<s_nLoci; ++i)
			s_vnSites.push_back(((p_Data.getId2DnaData()[0])[i]).size());
		s_vnRates			= p_Data.getRateDimensions();
		s_nStates			= 4;
		s_vvMultiplicities  = p_Data.getDnaMultipliers();
		s_vvvInvariantStates= p_Data.getInvariantNucleotides();
		int* temp_nSites = new int[s_nLoci];
		int* temp_nRates = new int[s_nLoci];
		for(int i=0; i<s_nLoci; ++i)
		{
			temp_nSites[i] = s_vnSites[i];
			temp_nRates[i] = s_vnRates[i];
		}
		s_pMemoryAllocator	= new TreeTime_Utilities::MemoryAllocator(
				s_nLoci, temp_nSites, temp_nRates, s_nStates);
		delete[] temp_nSites;
		delete[] temp_nRates;
		map<int,vector4d>& mIdToData = p_Data.getId2DnaData();
		for(map<int,vector4d>::iterator I = mIdToData.begin();
				I!=mIdToData.end(); ++I)
		{
			realf**** pMemory = s_pMemoryAllocator->allocateMemory();
			s_mIdToMemory.insert(pair<int,realf****>(I->first,pMemory));
			for(int loci=0; loci<s_nLoci; ++loci)
			{
				int ratesAtLoci = s_vnRates[loci];
				for(int site=0; site<s_vnSites[loci]; ++site)
				{
					for(int nucl=0; nucl<s_nStates; ++nucl)
					{
						for(int rate=0; rate<ratesAtLoci; ++rate)
						{
							pMemory[loci][rate][site][nucl] =
									(I->second)[loci][site][nucl][0];
						}
					}
				}
			}
		}
	}

	void LikelihoodCalculations::initRateMultipliers(Tree_Iterator p_iterator, int p_nLoci)
	{
		p_iterator->m_globalRateMultiplier = 1.0;
		p_iterator->m_vLocalRateMultiplier = vector<double>(p_nLoci,1.0);
		if(p_iterator.hasChildNodes())
		{
			initRateMultipliers(p_iterator.getFirstChild(),p_nLoci);
			initRateMultipliers(p_iterator.getLastChild(),p_nLoci);
		}
	}

	void LikelihoodCalculations::allocateMemory(
			Tree_Iterator p_iterator)
	{
		p_iterator->m_molecularData = s_pMemoryAllocator->allocateMemory();
	}

	void LikelihoodCalculations::allocateAndCopyMemory(
			Tree_Iterator p_iterator)
	{
		p_iterator->m_molecularData =
				s_pMemoryAllocator->allocateMemory(p_iterator->m_molecularData);
	}

	void LikelihoodCalculations::allocateSubtreeMemory(
			Tree_Iterator p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			allocateMemory(p_iterator);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	void LikelihoodCalculations::allocateUpwardsMemory(
			Tree_Iterator p_iterator)
	{
		Tree_Iterator I = p_iterator;
		while(I)
		{
			allocateMemory(I);
			I = I.getFather();
		}
	}

	void LikelihoodCalculations::deallocateMemory(
			Tree_Iterator p_iterator)
	{
		s_pMemoryAllocator->freeMemory(p_iterator->m_molecularData);
	}

	void LikelihoodCalculations::deallocateMemory(realf**** p_memory)
	{
		s_pMemoryAllocator->freeMemory(p_memory);
	}

	void LikelihoodCalculations::assignLeafMemory(Tree_Tree* p_pTree)
	{
		Tree_Iterator I = p_pTree->getRoot();
		stack<Tree_Iterator> sIterator;
		sIterator.push(I);
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
			else
			{
				I->m_molecularData = s_mIdToMemory[I->m_id];
			}
		}
	}

	void LikelihoodCalculations::calcDnaProbabilities(
			Tree_Iterator p_iterator,
			vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&
			p_vpDnaModel, vector<double>& p_vRateScaling,
			double p_firstChildTime, double p_lastChildTime)
	{
		for(int loci=0; loci<s_nLoci; ++loci)
		{
			calcDnaProbabilities(p_iterator, p_vpDnaModel, p_vRateScaling,
					loci, p_firstChildTime, p_lastChildTime);
		}
	}

	void LikelihoodCalculations::calcDnaProbabilities(
			Tree_Iterator p_iterator,
			vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&
			p_vpDnaModel, vector<double>& p_vRateScaling, int p_loci,
			double p_firstChildTime, double p_lastChildTime)
	{
		int loci		= p_loci;
		int sitesAtLoci	= s_vnSites[loci];
		double rateMulti;
		realf**** __restrict probs = p_iterator->m_molecularData;
		realf**** __restrict first = p_iterator.getFirstChild()
				->m_molecularData;
		realf**** __restrict last  = p_iterator.getLastChild()
				->m_molecularData;
		realf*** __restrict pppProb = probs[loci];;
		realf*** __restrict pppFirs = first[loci];
		realf*** __restrict pppLast = last[loci];
		realf**  __restrict ppProb;
		realf**  __restrict ppFirs;
		realf**  __restrict ppLast;
		realf*   __restrict pProb;
		realf*   __restrict pFirs;
		realf*   __restrict pLast;
		double** __restrict tM;
		for(int rate=0; rate<s_vnRates[loci];
				++rate, ++pppProb, ++pppFirs, ++pppLast)
		{
			rateMulti = p_vRateScaling[loci] *
					p_vpDnaModel[loci]->getGammaRates()[rate];
			tM = p_vpDnaModel[loci]->getTransitionMatrix(
					p_firstChildTime*rateMulti);
			register const realf first_tM[4][4] = {
					{tM[0][0],tM[0][1],tM[0][2],tM[0][3]},
					{tM[1][0],tM[1][1],tM[1][2],tM[1][3]},
					{tM[2][0],tM[2][1],tM[2][2],tM[2][3]},
					{tM[3][0],tM[3][1],tM[3][2],tM[3][3]}};
			tM  = p_vpDnaModel[loci]->getTransitionMatrix(
					p_lastChildTime*rateMulti);
			register const realf last_tM[4][4] = {
					{tM[0][0],tM[0][1],tM[0][2],tM[0][3]},
					{tM[1][0],tM[1][1],tM[1][2],tM[1][3]},
					{tM[2][0],tM[2][1],tM[2][2],tM[2][3]},
					{tM[3][0],tM[3][1],tM[3][2],tM[3][3]}};
			for(ppProb = *pppProb, ppFirs = *pppFirs, ppLast = *pppLast;
					ppProb != (*pppProb + sitesAtLoci);)
			{
				pProb = *ppProb++;
				pFirs = *ppFirs++;
				pLast = *ppLast++;
				for(int n=0; n<4; ++n)
				{  
					*pProb++ = (
							(first_tM[n][0]* (* pFirs  )) +
							(first_tM[n][1]* (*(pFirs+1)))+
							(first_tM[n][2]* (*(pFirs+2)))+
							(first_tM[n][3]* (*(pFirs+3))))*(
							(last_tM[n][0] * (*(pLast  ))) +
							(last_tM[n][1] * (*(pLast+1))) +
							(last_tM[n][2] * (*(pLast+2))) +
							(last_tM[n][3] * (*(pLast+3))));
				}      
			}
		}
	}

	// use the following function
	void LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
			Tree_Iterator p_iterator,
			vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&
			p_vpDnaModel, vector<double>& p_vRateScaling)
	{
		if(p_iterator.hasChildNodes())
		{
			calcScaledSubtreeDnaProbabilities(p_iterator.getFirstChild(),
					p_vpDnaModel, p_vRateScaling);
			calcScaledSubtreeDnaProbabilities(p_iterator.getLastChild(),
					p_vpDnaModel, p_vRateScaling);
			for(unsigned int loci = 0; loci < p_vpDnaModel.size(); ++loci)
			{
				Tree_Iterator it = p_iterator.getFirstChild();
				double firstChildLength = it->m_length * it->m_globalRateMultiplier *
						it->m_vLocalRateMultiplier[loci];
				it = p_iterator.getLastChild();
				double lastChildLength = it->m_length * it->m_globalRateMultiplier *
						it->m_vLocalRateMultiplier[loci];
				calcDnaProbabilities(p_iterator, p_vpDnaModel, p_vRateScaling,
						loci, firstChildLength, lastChildLength);
			}
		}
	}

	void LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
			Tree_Iterator p_iterator,
			vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>& p_vpDnaModel,
			vector<double>& p_vRateScaling, int p_loci)
	{
		if(p_iterator.hasChildNodes())
		{
			calcScaledSubtreeDnaProbabilities(p_iterator.getFirstChild(),
					p_vpDnaModel, p_vRateScaling, p_loci);
			calcScaledSubtreeDnaProbabilities(p_iterator.getLastChild(),
					p_vpDnaModel, p_vRateScaling, p_loci);
			Tree_Iterator it = p_iterator.getFirstChild();
			double firstChildLength = it->m_length * it->m_globalRateMultiplier *
					it->m_vLocalRateMultiplier[p_loci];
			it = p_iterator.getLastChild();
			double lastChildLength = it->m_length * it->m_globalRateMultiplier *
					it->m_vLocalRateMultiplier[p_loci];
			calcDnaProbabilities(p_iterator, p_vpDnaModel, p_vRateScaling,
					p_loci, firstChildLength, lastChildLength);
		}
	}

	void LikelihoodCalculations::calcScaledUpwardsDnaProbabilities(
			Tree_Iterator p_iterator,
			vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>& p_vpDnaModel,
			vector<double>& p_vRateScaling)
	{
		for(unsigned int loci = 0; loci < p_vpDnaModel.size(); ++loci)
		{
			Tree_Iterator it = p_iterator.getFirstChild();
			double firstChildLength = it->m_length * it->m_globalRateMultiplier *
					it->m_vLocalRateMultiplier[loci];
			it = p_iterator.getLastChild();
			double lastChildLength = it->m_length * it->m_globalRateMultiplier *
					it->m_vLocalRateMultiplier[loci];
			calcDnaProbabilities(p_iterator, p_vpDnaModel, p_vRateScaling,
					loci, firstChildLength, lastChildLength);
		}
		if(p_iterator.getFather())
		{
			calcScaledUpwardsDnaProbabilities(p_iterator.getFather(),
					p_vpDnaModel, p_vRateScaling);
		}
	}

	double LikelihoodCalculations::getLogLikelihoodTree( Tree_Tree* p_pTree,
			vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&
			p_vpDnaModel)
	{
		double		logLikelihood	= 0;
		realf****	probs			= p_pTree->getRoot()->m_molecularData;
		for(int loci=0; loci<s_nLoci; ++loci)
		{
			const double* statDist = p_vpDnaModel[loci]->
					getStationaryDistribution();
			double lociLogLikelihood = 0;
			if(p_vpDnaModel[loci]->hasInvariantSites())
			{
				double propInvSites	=
						p_vpDnaModel[loci]->getProportionOfInvariantSites();
				double weights = (1.0-propInvSites)/
						p_vpDnaModel[loci]->getNGammaRates();
				for(int site=0; site<s_vnSites[loci]; ++site)
				{
					double siteLikelihood = 0;
					for(int rate=0; rate < s_vnRates[loci]; ++rate)
					{
						siteLikelihood += (
							(probs[loci][rate][site][0]*statDist[0]) +
							(probs[loci][rate][site][1]*statDist[1]) +
							(probs[loci][rate][site][2]*statDist[2]) +
							(probs[loci][rate][site][3]*statDist[3])
						);
					}
					siteLikelihood *= weights;
					siteLikelihood += (
						((s_vvvInvariantStates[loci][site][0]*statDist[0]) +
						(s_vvvInvariantStates[loci][site][1]*statDist[1]) +
						(s_vvvInvariantStates[loci][site][2]*statDist[2]) +
						(s_vvvInvariantStates[loci][site][3]*statDist[3]))
						* propInvSites
					);
					lociLogLikelihood += (s_vvMultiplicities[loci][site]*
							log(siteLikelihood));
				}
				logLikelihood += lociLogLikelihood;
			}
			else
			{
				double weights = 1.0 / p_vpDnaModel[loci]->getNGammaRates();
				for(int site=0; site<s_vnSites[loci]; ++site)
				{
					double siteLikelihood = 0;
					for(int rate=0; rate < s_vnRates[loci]; ++rate)
					{
						siteLikelihood += (
							(probs[loci][rate][site][0]*statDist[0]) +
							(probs[loci][rate][site][1]*statDist[1]) +
							(probs[loci][rate][site][2]*statDist[2]) +
							(probs[loci][rate][site][3]*statDist[3])
						);
					}
					siteLikelihood *= weights;
					lociLogLikelihood += (s_vvMultiplicities[loci][site]*
							log(siteLikelihood));
				}
				logLikelihood += lociLogLikelihood;
			}
		}
		//todo : define a Limit
		if(logLikelihood < -100000.0 || logLikelihood > 0)
		{
			return -110000.0;
		}
		return logLikelihood;
	}

	double LikelihoodCalculations::getLogLikelihoodConstraints(
			Tree_Tree* p_pTree, vector<TreeTime_Constraint::
					TopologicalConstraint*> p_vpConstraint)
	{
		double logProb = 0;
		for(vector<TreeTime_Constraint::TopologicalConstraint*>::iterator I =
				p_vpConstraint.begin(); I != p_vpConstraint.end(); ++I)
		{
			if(!(*I)->isTimeConstraint())
				continue;
			set<int> ids;
			set_union((*I)->getSpeciesIds1().begin(),
					(*I)->getSpeciesIds1().end(),(*I)->getSpeciesIds2().begin(),
					(*I)->getSpeciesIds2().end(),inserter(ids,ids.begin()));
			Tree_Iterator mrca = getMrca(p_pTree,ids);
			double timepoint = mrca->m_timepoint;
			logProb += (*I)->getLogDensity(timepoint);
		}
		if(logProb < -100000.0)
		{
			return -110000.0;
		}
		return logProb;
	}

	double LikelihoodCalculations::getTreeLength(Tree_Tree* p_pTree)
	{
		double treeLength = 0;
		Tree_Iterator I = p_pTree->getRoot();
		stack<Tree_Iterator> sIterator;
		sIterator.push(I);
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			treeLength += I->m_length;
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		return treeLength;
	}

	int LikelihoodCalculations::getSubtreeSize(Tree_Iterator p_I)
	{
		int size = 0;
		if(!p_I.hasChildNodes())
		{
			return 1;
		}
		else
		{
			size += getSubtreeSize(p_I.getFirstChild());
			size += getSubtreeSize(p_I.getLastChild());
		}
		return size;
	}

	vector<double> LikelihoodCalculations::getBranchingTimepoints(Tree_Iterator
			p_I)
	{
		vector<double> BranchingTimepoints;
		getBranchingTimepoints(p_I, BranchingTimepoints);
		sort(BranchingTimepoints.begin(), BranchingTimepoints.end() );
		return BranchingTimepoints; 
	}

	void LikelihoodCalculations::getBranchingTimepoints(Tree_Iterator
			p_I, vector<double>& p_vBranchingTimepoints)
	{
		if(p_I.hasChildNodes())
		{
			p_vBranchingTimepoints.push_back(p_I->m_timepoint);
			getBranchingTimepoints(p_I.getFirstChild(), p_vBranchingTimepoints);
			getBranchingTimepoints(p_I.getLastChild(), p_vBranchingTimepoints);
		}
	}

	Tree_Iterator LikelihoodCalculations::getMrca(
			Tree_Tree* p_pTree, set<int>& p_sIds)
	{
		Tree_Iterator temp = p_pTree->getRoot();
		int size = p_sIds.size();
		getMrca(temp,size,p_sIds,temp);
		return temp;
	}

	int LikelihoodCalculations::getMrca(Tree_Iterator p_I, int p_size,
			set<int>& p_sIds, Tree_Iterator& p_result)
	{
		int temp = 0;
		if(!p_I.hasChildNodes())
		{
			if(p_sIds.find(p_I->m_id) != p_sIds.end())
			{
				temp = 1;
			}
		}
		else
		{
			temp += getMrca(p_I.getFirstChild(),p_size,p_sIds,p_result);
			temp += getMrca(p_I.getLastChild(),p_size,p_sIds,p_result);
		}
		if(temp == p_size)
		{
			p_result = p_I;
			temp -= 1;
		}
		return temp;
	}

	// todo more efficient with root,
	// and caching of pointers to leafes.
	set<Tree_Iterator> LikelihoodCalculations::getRotateableNodes(
					Tree_Tree* p_pTree,
					vector<TreeTime_Constraint::TopologicalConstraint*>
					p_vpConstraint, bool p_hasOutgroup)
	{
		set<Tree_Iterator> Nodes;
		Tree_Iterator I     = p_pTree->getRoot();
		Tree_Iterator first = I.getFirstChild();
		Tree_Iterator last  = I.getLastChild();
		stack<Tree_Iterator> sIterator;
		if(p_hasOutgroup)
		{
			if(first.hasChildNodes())
			{
				if(first.getFirstChild().hasChildNodes())
				{
					sIterator.push(first.getFirstChild().getFirstChild());
					sIterator.push(first.getFirstChild().getLastChild());
				}
				if(first.getLastChild().hasChildNodes())
				{
					sIterator.push(first.getLastChild().getFirstChild());
					sIterator.push(first.getLastChild().getLastChild());
				}
			}
			if(last.hasChildNodes())
			{
				if(last.getFirstChild().hasChildNodes())
				{
					sIterator.push(last.getFirstChild().getFirstChild());
					sIterator.push(last.getFirstChild().getLastChild());
				}
				if(last.getLastChild().hasChildNodes())
				{
					sIterator.push(last.getLastChild().getFirstChild());
					sIterator.push(last.getLastChild().getLastChild());
				}
			}
		}
		else
		{
			if(first.hasChildNodes())
			{
				sIterator.push(first.getFirstChild());
				sIterator.push(first.getLastChild());
			}
			if(last.hasChildNodes())
			{
				sIterator.push(last.getFirstChild());
				sIterator.push(last.getLastChild());
			}
		}
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			Nodes.insert(I);
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		for(vector<TreeTime_Constraint::TopologicalConstraint*>::iterator iC =
				p_vpConstraint.begin(); iC != p_vpConstraint.end(); ++iC)
		{
			set<int> nr1 = (*iC)->getSpeciesIds1();
			set<int> nr2 = (*iC)->getSpeciesIds2();
			set<int> all = nr1;
			all.insert(nr2.begin(),nr2.end());
			Tree_Iterator iNr1 = getMrca(p_pTree, nr1);
			Tree_Iterator iNr2 = getMrca(p_pTree, nr2);
			Tree_Iterator iAll = getMrca(p_pTree, all);
			if(iNr1.getFather() == iAll)
			{
				Nodes.erase(iNr1.getFirstChild());
				Nodes.erase(iNr1.getLastChild());
			}
			if(iNr2.getFather() == iAll)
			{
				Nodes.erase(iNr2.getFirstChild());
				Nodes.erase(iNr2.getLastChild());
			}
		}
		return Nodes;
	}

	void LikelihoodCalculations::scaleSubTree(Tree_Iterator p_I,
			double p_factor)
	{
		p_I->m_length    *= p_factor;
		p_I->m_timepoint *= p_factor;
		for(map<int, list<pair<double,double> > >::iterator M =
				p_I->m_mlChangepoints.begin();
				M != p_I->m_mlChangepoints.end(); ++M)
		{
			for(list<pair<double,double> >::iterator I = (++M->second.begin());
					I != M->second.end(); ++I)
			{
				I->first *= p_factor;
			}
		}
		if(p_I.hasChildNodes())
		{
			scaleSubTree(p_I.getFirstChild(),p_factor);
			scaleSubTree(p_I.getLastChild() ,p_factor);
		}
	}

	void LikelihoodCalculations::scaleInnerBranch( Tree_Iterator p_I,
			double p_factor)
	{
		p_I->m_length    *= p_factor;
		for(map<int, list<pair<double,double> > >::iterator M =
				p_I->m_mlChangepoints.begin();
				M != p_I->m_mlChangepoints.end(); ++M)
		{
			for(list<pair<double,double> >::iterator I = (++M->second.begin());
					I != M->second.end(); ++I)
			{
				I->first *= p_factor;
			}
		}
	}

	void LikelihoodCalculations::initTimepoints(Tree_Iterator p_I)
	{
		if(p_I.hasChildNodes())
		{
			initTimepoints(p_I.getFirstChild());
			initTimepoints(p_I.getLastChild());
			p_I->m_timepoint = p_I.getFirstChild()->m_timepoint +
					p_I.getFirstChild()->m_length;
		}
		else
		{
			p_I->m_timepoint = 0.0;
		}
	}

	void LikelihoodCalculations::initRates(Tree_Iterator p_I, int p_nLoci)
	{
		p_I->m_globalRateMultiplier = 1.0;
		p_I->m_vLocalRateMultiplier = vector<double>(p_nLoci,1.0);
		if(p_I.hasChildNodes())
		{
			initRates(p_I.getFirstChild(), p_nLoci);
			initRates(p_I.getLastChild(), p_nLoci);
		}
	}
	

	void LikelihoodCalculations::rootWithOutgroup(Tree_Tree* p_pTree,
			int p_outgroup)
	{
		set<int> sOutgroup;
		sOutgroup.insert(p_outgroup);
		Tree_Iterator I = getMrca(p_pTree, sOutgroup);
		while(!(I.getFather() == p_pTree->getRoot()))
		{
			Tree_Iterator	grandchild = I;
			Tree_Iterator	child  = grandchild.getFather();
			Tree_Iterator	father = child.getFather();
			Tree_Iterator	uncle  = father.getFirstChild();
			if(uncle == child)
			{
				uncle = father.getLastChild();
			}
			if(father.getFirstChild() == child)
				father.setLastChild(grandchild);
			else
				father.setFirstChild(grandchild);
			if(child.getFirstChild() == grandchild)
				child.setFirstChild(uncle);
			else
				child.setLastChild(uncle);
			grandchild.setFather(father);
			uncle.setFather(child);
			double scaling = father->m_timepoint/child->m_timepoint;
			scaleSubTree(grandchild, scaling);
			scaleSubTree(uncle, 1.0/scaling);
		}
	}

}// TreeTime_Core
