/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef LIKELIHOODCALCULATIONS_HPP_
#define LIKELIHOODCALCULATIONS_HPP_

#include "Settings.hpp"
#include "../TreeTime_MolecularTree/Tree.hpp"
#include "../TreeTime_MolecularTree/iterator.hpp"
#include "../TreeTime_Utilities/MemoryAllocator.hpp"
#include "../TreeTime_MolecularEvolution/DnaEvolutionModel.hpp"
#include "../TreeTime_Constraint/TopologicalConstraint.hpp"
#include "../TreeTime_Constraint/GaussianConstraint.hpp"
#include "../TreeTime_Constraint/GammaConstraint.hpp"
#include "NodeData.hpp"
#include "../TreeTime_NexusProcessing/Data.hpp"

namespace TreeTime_Core
{

	typedef TreeTime_MolecularTree::iterator<NodeData> Tree_Iterator;
	typedef TreeTime_MolecularTree::Tree<NodeData> Tree_Tree;

	class LikelihoodCalculations
	{

		public:

			static void init(TreeTime_NexusProcessing::Data&);

			static void initRateMultipliers(Tree_Iterator p_iterator,
					int p_nLoci);

			static void allocateMemory(Tree_Iterator p_iterator);

			static void allocateAndCopyMemory(Tree_Iterator p_iterat);

			static void allocateSubtreeMemory(
					Tree_Iterator p_iterator);

			static void allocateUpwardsMemory(
					Tree_Iterator p_iterator);

			static void deallocateMemory(Tree_Iterator p_iterator);

			static void deallocateMemory(realf**** p_memory);

			static void assignLeafMemory(Tree_Tree* p_pTree);

			static void calcDnaProbabilities(
					Tree_Iterator p_iterator,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>& p_vpDnaModel,
					vector<double>& p_vRateScaling,
					double p_firstChildTime,
					double p_lastChildTime);

			static void calcDnaProbabilities(
					Tree_Iterator p_iterator,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>& p_vpDnaModel,
					vector<double>& p_vRateScaling,
					int p_loci,
					double p_firstChildTime,
					double p_lastChildTime);

			static void calcScaledSubtreeDnaProbabilities(
					Tree_Iterator p_iterator,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>& p_vpDnaModel,
					vector<double>& p_vRateScaling);

			static void calcScaledSubtreeDnaProbabilities(
					Tree_Iterator p_iterator,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>& p_vpDnaModel,
					vector<double>& p_vRateScaling,
					int p_loci);

			static void calcScaledUpwardsDnaProbabilities(
					Tree_Iterator p_iterator,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>& p_vpDnaModel,
					vector<double>& p_vRateScaling);

//			static void calcScaledUpwardsDnaProbabilities(
//					Tree_Iterator p_iterator,
//					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&,
//					vector<double>& p_vRateScaling,
//					int p_loci);

			static double getLogLikelihoodTree(Tree_Tree* p_pTree,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&
					p_vpDnaModel);

			static double getLogLikelihoodConstraints(Tree_Tree* p_pTree,
					vector<TreeTime_Constraint::TopologicalConstraint*>
					p_vpConstraint);

			static double getTreeLength(Tree_Tree* p_pTree);

//			static double getUnitRateTreeLength(Tree_Tree* p_pTree);

			static int getSubtreeSize(Tree_Iterator p_I);

			static vector<double> getBranchingTimepoints(Tree_Iterator p_I);

			static Tree_Iterator getMrca(Tree_Tree* p_pTree,
					set<int>& p_sIds);

			static int getMrca(Tree_Iterator p_I, int p_size,
					set<int>& p_sIds, Tree_Iterator& p_result);

			static set<Tree_Iterator> getRotateableNodes(Tree_Tree* p_pTree,
					vector<TreeTime_Constraint::TopologicalConstraint*>
					p_vpConstraint, bool p_hasOutgroup);

			static void scaleSubTree(Tree_Iterator p_I, double p_factor);

			static void scaleInnerBranch(Tree_Iterator p_I, double p_factor);

			static void initTimepoints(Tree_Iterator p_I);

			static void initRates(Tree_Iterator p_I, int p_nLoci);

			static void rootWithOutgroup(Tree_Tree* p_pTree,
					int p_outgroup);

		private:

//			static double getScaledLength(Tree_Iterator p_I);
			
			static void getBranchingTimepoints(Tree_Iterator p_I,
					vector<double>& p_vBranchingTimepoints);

			static void calcDnaProbabilities_loops(
					Tree_Iterator p_iterator,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&,
					vector<double>& p_vRateScaling,
					double p_firstChildTime,
					double p_lastChildTime);

			static void calcDnaProbabilities_unroll(
					Tree_Iterator p_iterator,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&,
					vector<double>& p_vRateScaling,
					double p_firstChildTime,
					double p_lastChildTime);

			static void calcDnaProbabilities_pointer(
					Tree_Iterator p_iterator,
					vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&,
					vector<double>& p_vRateScaling,
					double p_firstChildTime,
					double p_lastChildTime);

		private:

			static TreeTime_Utilities::MemoryAllocator*	s_pMemoryAllocator;
			static map<int,realf****>				s_mIdToMemory;
			static int								s_nLoci;
			static vector<int>						s_vnSites;
			static vector<int>						s_vnRates;
			static int								s_nStates;
			static vector<vector<int> >				s_vvMultiplicities;
			static vector<vector<vector<int> > > 	s_vvvInvariantStates;

	};//LikelihoodCalculations

}//TreeTime_Core

#endif /*LIKELIHOODCALCULATIONS_HPP_*/
