/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "ProposeChange.hpp"
#include "RandomNumberGenerator.hpp"
#include "StateChanges/ChangeInitialRates.hpp"
#include "StateChanges/ChangeNodeTimepoint.hpp"
#include "StateChanges/ChangeRootTimepoint.hpp"
#include "StateChanges/ChangeTopology.hpp"
#include "StateChanges/ChangeMemGammaRates.hpp"
#include "StateChanges/ChangeMemInvSites.hpp"
#include "StateChanges/ChangeMemStatDist.hpp"
#include "StateChanges/ChangeMemTransParam.hpp"
#include "../TreeTime_RateChangeModel/RateChangeModel.hpp"

namespace TreeTime_Core
{

	ProposeChange::ProposeChange(double p_estimatedPemTime,
			bool p_hasTimeConstraints, double p_estimatedRootTime,
			double p_nSpecies, bool p_hasOutgroup,
			double p_weightTopology, double p_weightNodes,
			double p_weightEvolModels, double p_weightRateChangeModel,
			double p_weightInitialRates,
			double p_changeNodeTime, double p_changeTimeSynchronisation,
			double p_changeRootTime,
			State* p_pState)
	: m_changeRootTimepointFactor	(p_estimatedRootTime)
	, m_changeInitialRatesFactor	(p_estimatedPemTime/p_estimatedRootTime)
	, m_hasOutgroup					(p_hasOutgroup)
	, m_changeNodeTime				(p_changeNodeTime)
	, m_changeTimeSynchronisation	(p_changeTimeSynchronisation)
	, m_changeRootTime				(p_changeRootTime)
	, m_nLoci						(p_pState->m_vpDnaEvolutionModel.size())
	, m_vRcmWeigths					()
	, m_vvMemWeigths				()
	, m_vGroupWeights				()
	{
		// init m_vRcmWeigths
		m_vRcmWeigths = vector<double>(m_nLoci+1, 0.0);
		double tempRcm = 0;
		if(p_pState->m_pGlobalRateChangeModel->getId() == 1)
		{
			m_vRcmWeigths[m_nLoci] = 1.0;
			tempRcm += m_vRcmWeigths[m_nLoci];
		}
		else if(p_pState->m_pGlobalRateChangeModel->getId() >= 2)
		{
			m_vRcmWeigths[m_nLoci] = 0.5;
			tempRcm += m_vRcmWeigths[m_nLoci];
		}
		for(int i = 0; i < m_nLoci; ++i)
		{
			if(p_pState->m_vpLocalRateChangeModel[i]->getId() == 1)
			{
				m_vRcmWeigths[i] = 1.0;
				tempRcm += m_vRcmWeigths[i];
			}
			else if(p_pState->m_vpLocalRateChangeModel[i]->getId() >= 2)
			{
				m_vRcmWeigths[i] = 0.5;
				tempRcm += m_vRcmWeigths[i];
			}
		}
		if(tempRcm < 1.0E-09)
		{
			p_weightRateChangeModel = 0;
		}
		else
		{
			double tempAggMem = 0;
			for(int i = 0; i < (m_nLoci+1); ++i)
			{
				m_vRcmWeigths[i] = m_vRcmWeigths[i]/tempRcm + tempAggMem;
				tempAggMem = m_vRcmWeigths[i];
			}
		}
		// init m_vvMemWeigths
		m_vvMemWeigths = vector<vector<double> >(m_nLoci,vector<double>(4,0.0));
		for(int i = 0; i < m_nLoci; ++i)
		{
			if(p_pState->m_vpDnaEvolutionModel[i]->hasInvariantSites())
			{
				m_vvMemWeigths[i][0] = 1.0;
			}
			if(p_pState->m_vpDnaEvolutionModel[i]->hasGammaRates())
			{
				m_vvMemWeigths[i][1] = 1.0;
			}
			if(p_pState->m_vpDnaEvolutionModel[i]->
					getNIndependentStatDistrParameters() > 0)
			{
				m_vvMemWeigths[i][2] = p_pState->m_vpDnaEvolutionModel[i]->
						getNIndependentStatDistrParameters();
			}
			if(p_pState->m_vpDnaEvolutionModel[i]->
					getNIndependentTransModelParameters() > 0)
			{
				m_vvMemWeigths[i][3] = p_pState->m_vpDnaEvolutionModel[i]->
						getNIndependentTransModelParameters();
			}
		}
		double tempMem = 0;
		for(int i = 0; i < m_nLoci; ++i)
		{
			for(int j = 0; j < 4; ++j)
			{
				if(m_vvMemWeigths[i][j] > 0)
				{
					m_vvMemWeigths[i][j] = 1.0 + log(m_vvMemWeigths[i][j]);
					tempMem += m_vvMemWeigths[i][j];
				}
			}
		}
		if(tempMem < 1.0E-09)
		{
			p_weightEvolModels = 0;
		}
		else
		{
			double tempAggMem = 0;
			for(int i = 0; i < m_nLoci; ++i)
			{
				for(int j = 0; j < 4; ++j)
				{
					double t = m_vvMemWeigths[i][j]/tempMem;
					m_vvMemWeigths[i][j] = t + tempAggMem;
					tempAggMem += t;
				}
			}
		}
		// init m_vGroupWeights
		p_weightTopology		= fabs(p_weightTopology);
		p_weightNodes			= fabs(p_weightNodes);
		p_weightInitialRates	= fabs(p_weightInitialRates);
		p_weightEvolModels		= fabs(p_weightEvolModels);
		p_weightRateChangeModel	= fabs(p_weightRateChangeModel);
		double sum = p_weightTopology + p_weightNodes + p_weightEvolModels +
				p_weightRateChangeModel + p_weightInitialRates;
		m_vGroupWeights = vector<double>(6, 1.0);
		m_vGroupWeights[0] = p_weightTopology/sum;
		if(p_hasTimeConstraints)
		{
			m_vGroupWeights[1] = m_vGroupWeights[0] + ((p_weightNodes/sum) *
					((p_nSpecies-2)/(p_nSpecies-1)));
			m_vGroupWeights[2] = m_vGroupWeights[0] + p_weightNodes/sum;
		}
		else
		{
			m_vGroupWeights[1] = m_vGroupWeights[0] + p_weightNodes/sum;
			m_vGroupWeights[2] = m_vGroupWeights[1];
		}
		m_vGroupWeights[3] = m_vGroupWeights[2] + p_weightInitialRates/sum;
		m_vGroupWeights[4] = m_vGroupWeights[3] + p_weightEvolModels/sum;
	}

	ProposeChange::~ProposeChange()
	{
	}

	TreeTime_Mcmcmc::Change* ProposeChange::proposeNewChange(State* p_pState)
			const
	{
		double uniform = RandomNumberGenerator::getRNG().getUniform();
		if(uniform < m_vGroupWeights[0])
		{//cout << "Top" << endl;
			return new ChangeTopology(p_pState, 1.0, m_hasOutgroup);
		}
		else if(uniform < m_vGroupWeights[1])
		{//cout << "Nod" << endl;
			return new ChangeNodeTimepoint(p_pState, 1.0, m_changeNodeTime);
		}
		else if(uniform < m_vGroupWeights[2])
		{//cout << "Root" << endl;
			return new ChangeRootTimepoint(p_pState, 1.0,
					m_changeRootTimepointFactor, m_changeRootTime);
		}
		else if(uniform < m_vGroupWeights[3])
		{//cout << "InRates" << endl;
			return new ChangeInitialRates(p_pState, 1.0,
					m_changeInitialRatesFactor, m_changeTimeSynchronisation);
		}
		else if(uniform < m_vGroupWeights[4])
		{//cout << "Mems" << endl;
			return getMolecularEvolutionModelChange(p_pState, (uniform-
				m_vGroupWeights[3])/(m_vGroupWeights[4]-m_vGroupWeights[3]));
		}
		else
		{//cout << "Rcms" << endl;
			return getRateChangeModelChange(p_pState, (uniform-
					m_vGroupWeights[4])/(1.0-m_vGroupWeights[4]));
		}
	}

	TreeTime_Mcmcmc::Change* ProposeChange::getMolecularEvolutionModelChange(
			State* p_pState, double p_uniform) const
	{
		for(unsigned int loci = 0; loci < m_vvMemWeigths.size(); ++loci)
		{
			if(p_uniform < m_vvMemWeigths[loci][0])
			{
				return new ChangeMemInvSites(p_pState, 1.0, loci);
			}
			else if(p_uniform < m_vvMemWeigths[loci][1])
			{
				return new ChangeMemGammaRates(p_pState, 1.0, loci);
			}
			else if(p_uniform < m_vvMemWeigths[loci][2])
			{
				return new ChangeMemStatDist(p_pState, 1.0, loci);
			}
			else if(p_uniform < m_vvMemWeigths[loci][3])
			{
				return new ChangeMemTransParam(p_pState, 1.0, loci);
			}
		}
		return new ChangeMemStatDist(p_pState, 1.0, 0);
	}

	TreeTime_Mcmcmc::Change* ProposeChange::getRateChangeModelChange(
			State* p_pState, double p_uniform) const
	{
		double uniform = RandomNumberGenerator::getRNG().getUniform();
		for(unsigned int i = 0; i < m_vRcmWeigths.size()-1 ;++i)
		{
			if(p_uniform < m_vRcmWeigths[i])
			{
				return p_pState->m_vpLocalRateChangeModel[i]->getChange(
						p_pState, i, uniform);
			}
		}
		return p_pState->m_pGlobalRateChangeModel->getChange(p_pState, -1,
				uniform);
	}

}//TreeTime_Core
