/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "RandomNumberGenerator.hpp"

namespace TreeTime_Core
{

	RandomNumberGenerator::RandomNumberGenerator()
	{
		m_pRNGTausworthe = gsl_rng_alloc(gsl_rng_taus);//Tausworthe generator
	}

	RandomNumberGenerator::~RandomNumberGenerator()
	{
		gsl_rng_free(m_pRNGTausworthe);//Free Space
	}

	RandomNumberGenerator& RandomNumberGenerator::getRNG()
	{
		static RandomNumberGenerator RNG;
		return RNG;
	}
	
	void RandomNumberGenerator::setSeed(int p_seed)
	{
		gsl_rng_set(m_pRNGTausworthe,p_seed);//Seed
	}

	double RandomNumberGenerator::getUniform() const
	{
		return gsl_rng_uniform_pos(m_pRNGTausworthe);
	}

	double RandomNumberGenerator::getUniform(double p_min, double p_max)
	{
		return p_min + (p_max-p_min)*gsl_rng_uniform_pos(m_pRNGTausworthe);
	}

	double RandomNumberGenerator::getUniform(double p_min, double p_max) const
	{
		return p_min + (p_max-p_min)*gsl_rng_uniform_pos(m_pRNGTausworthe);
	}

	int RandomNumberGenerator::getInteger(int p_span) const
	{
		return gsl_rng_uniform_int(m_pRNGTausworthe,p_span);
	}
	
	int RandomNumberGenerator::getInteger(int p_span, int p_minValue)
	{
		return p_minValue + gsl_rng_uniform_int(m_pRNGTausworthe,p_span);
	}
	
	double RandomNumberGenerator::getGaussian(double p_mean,
			double p_standardDeviation)
	{		
		double rgaus = p_mean;
		try
		{
			rgaus = p_mean + gsl_ran_gaussian(m_pRNGTausworthe, p_standardDeviation);
		}
		catch(...)
		{
			std::cout << "Exception occured in function: RandomNumberGenerator::getGaussian(...)" << " ";
			std::cout << p_mean << " " << p_standardDeviation << " " << rgaus <<std::endl;
		}
		return rgaus;
	}
	
	double RandomNumberGenerator::getGaussianDensity(double p_value,
			double p_standardDeviation)
	{
		double density = 1E-09;
		try
		{
			density = gsl_ran_gaussian_pdf(p_value , p_standardDeviation);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getGaussianDensity(...)" << " ";
			std::cout << p_value << " " << p_standardDeviation << " " << density << std::endl;
		}
		return density;
	}
	
	double RandomNumberGenerator::getGamma(double p_alpha, double p_beta)
	{
		double rgamma = p_alpha/p_beta;
		try
		{
			rgamma = gsl_ran_gamma(m_pRNGTausworthe, p_alpha, 1.0/p_beta);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getGamma(...)" << " ";
			std::cout << p_alpha << " " << p_beta << " " << rgamma << std::endl;
		}
		return rgamma;
	}
	
	double RandomNumberGenerator::getGammaDensity(double p_value,
			double p_alpha, double p_beta)
	{
		double dgamma = 1E-09;
		try
		{
			dgamma = gsl_ran_gamma_pdf(p_value, p_alpha, 1.0/p_beta);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getGammaDensity(...)" << " ";
			std::cout << p_alpha << " " << p_beta << " " << p_value << std::endl;
		}
		return dgamma;
	}
	
	unsigned int RandomNumberGenerator::getPoisson(double p_mean)
	{
		unsigned int rpois = 0;
		try
		{
			rpois = gsl_ran_poisson(m_pRNGTausworthe, p_mean);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getPoisson(...)" << " ";
			std::cout << p_mean << " " << rpois << std::endl;
		}
		return rpois;
	}

	double RandomNumberGenerator::getPoissonDensity(unsigned int p_value,
			double p_mean)
	{
		double dpois = 1E-09;
		try
		{
			dpois = gsl_ran_poisson_pdf(p_value, p_mean);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getPoissonDensity(...)" << " ";
			std::cout << p_mean << " " << p_value << " " << dpois << std::endl;
		}
		return dpois;
	}

	void RandomNumberGenerator::getDirichlet(const double p_param[],
					double p_target[]) const
	{
		try
		{
			gsl_ran_dirichlet(m_pRNGTausworthe, 4, p_param, p_target);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getDirichlet(...)" << " ";
		}
	}

	double RandomNumberGenerator::getDirichletDensity(const double p_param[],
					const double p_target[]) const
	{
		double ddirich = 1E-09;
		try
		{
			ddirich = gsl_ran_dirichlet_pdf(4, p_param, p_target);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getDirichletDensity(...)" << std::endl;
		}
		return ddirich;
	}

	void RandomNumberGenerator::getDirichlet(const double p_param[],
					double p_target[], int p_number) const
	{
		try
		{
			gsl_ran_dirichlet(m_pRNGTausworthe, p_number, p_param, p_target);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getDirichlet(..., int)" << std::endl;
		}
	}

	double RandomNumberGenerator::getLogDirichletDensity(const double p_param[],
					const double p_target[], int p_number) const
	{
		double logddirich = -10000.0;
		try
		{
			logddirich = gsl_ran_dirichlet_lnpdf(p_number, p_param, p_target);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getLogDirichletDensity(...)" << std::endl;
		}
		return logddirich;
	}

	double RandomNumberGenerator::getExponential(double p_meanValue) const
	{		
		double rexp = p_meanValue;
		try
		{
			rexp = gsl_ran_exponential(m_pRNGTausworthe, p_meanValue);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getExponential(...)" << " ";
			std::cout<< p_meanValue << std::endl;
		}
		return rexp;
	}

	double RandomNumberGenerator::getExponentialDensity(double p_value,
			double p_meanValue)const
	{
		double dexp = 1E-09;
		try
		{
			dexp = gsl_ran_exponential_pdf(p_value, p_meanValue);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getExponentialDensity(...)" << " ";
			std::cout<< p_value << " " << p_meanValue << std::endl;
		}
		return dexp;
	}

	double RandomNumberGenerator::getLognormal(double p_mu, double p_sigma)
			const
	{
		double rlognorm = p_mu;
		try
		{
			rlognorm = gsl_ran_lognormal(m_pRNGTausworthe, p_mu, p_sigma);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getLognormal(...)" << " ";
			std::cout<< p_mu << " " << p_sigma << std::endl;
		}
		return rlognorm;
	}

	double RandomNumberGenerator::getLognormalDensity(double p_value,
			double p_mu, double p_sigma) const
	{		
		double dlognorm = 1E-09;
		try
		{
			dlognorm = gsl_ran_lognormal_pdf(p_value, p_mu, p_sigma);
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: RandomNumberGenerator::getLognormalDensity(...)" << " ";
			std::cout<< p_value << " " << p_mu << " " << p_sigma << std::endl;
		}
		return dlognorm;
		
	}

}//TreeTime_Core
