/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "ChangeInitialRates.hpp"
#include "../RandomNumberGenerator.hpp"
#include "../LikelihoodCalculations.hpp"

namespace TreeTime_Core
{

	ChangeInitialRates::ChangeInitialRates(State* p_pState,
			double p_proposalRatio, double p_scaling, double p_calibration)
	: m_pState						(p_pState)
	, m_proposalRatio				(p_proposalRatio)
	, m_scaling						(p_scaling)
	, m_pProposalTree				(0)
	, m_vRateScaling				(p_pState->m_vRateScaling)
	, m_proposalTreeLogLikelihood	(0)
	, m_calibration					(p_calibration)
	, m_logPriorRatio				(0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	ChangeInitialRates::~ChangeInitialRates()
	{
	}

	void ChangeInitialRates::changeState()
	{
		m_pProposalTree = m_pState->m_pTree->copy();
		allocateSubtreeMemory(m_pProposalTree->getRoot());
		for(vector<double>::iterator I = m_vRateScaling.begin();
				I != m_vRateScaling.end(); ++I)
		{
			double old = (*I);
			(*I) += RandomNumberGenerator::getRNG().getUniform(
					-0.3*m_calibration*m_scaling, 0.3*m_calibration*m_scaling);
			if( (*I) < 0.0001*m_scaling)
			{
				(*I) = (0.0002*m_scaling)-(*I);
			}
			else if( (*I) > 100000.0*m_scaling)
			{
				(*I) = (200000.0*m_scaling)  - (*I);
			}
			// Prior Ratio: E(m_scaling) = t_PEM/t_REAL:
			m_logPriorRatio += (old-(*I))/m_scaling;
		}
		LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
				m_pProposalTree->getRoot(), m_pState->m_vpDnaEvolutionModel,
				m_vRateScaling);
		m_proposalTreeLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodTree(
				m_pProposalTree, m_pState->m_vpDnaEvolutionModel);
	}

	double ChangeInitialRates::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood - m_pState->m_treeLogLikelihood +
				m_logPriorRatio);
	}

	double ChangeInitialRates::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void ChangeInitialRates::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
		m_pState->m_pTree				= m_pProposalTree;
		m_pState->m_vRateScaling 		= m_vRateScaling;
		m_pState->m_totalLogLikelihood	+= (m_proposalTreeLogLikelihood -
				m_pState->m_treeLogLikelihood);
		m_pState->m_treeLogLikelihood	= m_proposalTreeLogLikelihood;
	}

	void ChangeInitialRates::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	void ChangeInitialRates::allocateSubtreeMemory(
			TreeTime_MolecularTree::iterator<NodeData> p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LikelihoodCalculations::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	int ChangeInitialRates::getId () const
	{
		return 4;
	}

}//TreeTime_Core
