/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "ChangeMemGammaRates.hpp"
#include "../RandomNumberGenerator.hpp"
#include "../LikelihoodCalculations.hpp"

namespace TreeTime_Core
{

	ChangeMemGammaRates::ChangeMemGammaRates(State* p_pState,
			double p_proposalRatio, int p_modelNr)
	: m_modelNr			 			(p_modelNr)
	, m_pState						(p_pState)
	, m_proposalRatio				(p_proposalRatio)
	, m_pDnaEvolutionModel			(0)
	, m_proposalTreeLogLikelihood	(0)
	, m_lIteratorOldProbs			()
	{
	}

	ChangeMemGammaRates::~ChangeMemGammaRates()
	{
	}

	void ChangeMemGammaRates::changeState()
	{
		vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>
				vpDnaEvolutionModel = m_pState->m_vpDnaEvolutionModel;
		if(m_pState->m_vpDnaEvolutionModel.size() > 1)
		{
			allocateAndCopySubtreeMemory(m_pState->m_pTree->getRoot());
		}
		else
		{
			allocateSubtreeMemory(m_pState->m_pTree->getRoot());
		}
		m_pDnaEvolutionModel = m_pState->m_vpDnaEvolutionModel[m_modelNr]->
				clone();
		vpDnaEvolutionModel[m_modelNr] = m_pDnaEvolutionModel;
		m_proposalRatio *= m_pDnaEvolutionModel->changeGammaRates();
		LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
				m_pState->m_pTree->getRoot(), vpDnaEvolutionModel,
				m_pState->m_vRateScaling, m_modelNr);
		m_proposalTreeLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodTree(
				m_pState->m_pTree, vpDnaEvolutionModel);
	}

	double ChangeMemGammaRates::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood - m_pState->m_treeLogLikelihood);
	}

	double ChangeMemGammaRates::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void ChangeMemGammaRates::accept()
	{
		for(list<pair<TreeTime_MolecularTree::iterator<NodeData>,realf****> >::iterator I =
				m_lIteratorOldProbs.begin(); I!=m_lIteratorOldProbs.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(I->second);
		}
		delete m_pState->m_vpDnaEvolutionModel[m_modelNr];
		m_pState->m_vpDnaEvolutionModel[m_modelNr] = m_pDnaEvolutionModel;
		m_pState->m_totalLogLikelihood += (m_proposalTreeLogLikelihood -
				m_pState->m_treeLogLikelihood);
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
	}

	void ChangeMemGammaRates::reject()
	{
		for(list<pair<TreeTime_MolecularTree::iterator<NodeData>,realf****> >::iterator I =
				m_lIteratorOldProbs.begin(); I!=m_lIteratorOldProbs.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(I->first);
			I->first->m_molecularData = I->second;
		}
		delete m_pDnaEvolutionModel;
	}

	void ChangeMemGammaRates::allocateSubtreeMemory(
			TreeTime_MolecularTree::iterator<NodeData> p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lIteratorOldProbs.push_back(pair<TreeTime_MolecularTree::iterator<NodeData>,
					realf****>(p_iterator, p_iterator->m_molecularData));
			LikelihoodCalculations::allocateMemory(p_iterator);
			// Recursion
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	void ChangeMemGammaRates::allocateAndCopySubtreeMemory(
			TreeTime_MolecularTree::iterator<NodeData> p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lIteratorOldProbs.push_back(pair<TreeTime_MolecularTree::iterator<NodeData>,
					realf****>(p_iterator, p_iterator->m_molecularData));
			LikelihoodCalculations::allocateAndCopyMemory(p_iterator);
			// Recursion
			allocateAndCopySubtreeMemory(p_iterator.getFirstChild());
			allocateAndCopySubtreeMemory(p_iterator.getLastChild());
		}
	}

	int ChangeMemGammaRates::getId () const
	{
		return 8;
	}

}//TreeTime_Core
