/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "ChangeMemStatDist.hpp"
#include "../RandomNumberGenerator.hpp"
#include "../LikelihoodCalculations.hpp"

namespace TreeTime_Core
{

	ChangeMemStatDist::ChangeMemStatDist(State* p_pState, 
			double p_proposalRatio, int p_modelNr)
	: m_modelNr			 			(p_modelNr)
	, m_pState						(p_pState)
	, m_proposalRatio				(p_proposalRatio)
	, m_pProposalTree	 			(0)
	, m_pDnaEvolutionModel			(0)
	, m_proposalTreeLogLikelihood	(0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	ChangeMemStatDist::~ChangeMemStatDist()
	{
	}

	void ChangeMemStatDist::changeState()
	{
		m_pProposalTree = m_pState->m_pTree->copy();
		vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>
				vpDnaEvolutionModel = m_pState->m_vpDnaEvolutionModel;
		if(m_pState->m_vpDnaEvolutionModel.size() > 1)
		{
			allocateAndCopySubtreeMemory(m_pProposalTree->getRoot());
		}
		else
		{
			allocateSubtreeMemory(m_pProposalTree->getRoot());
		}
		m_pDnaEvolutionModel = m_pState->m_vpDnaEvolutionModel[m_modelNr]->
				clone();
		vpDnaEvolutionModel[m_modelNr] = m_pDnaEvolutionModel;
		m_proposalRatio *= m_pDnaEvolutionModel->changeStationaryDistribution();
		LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
				m_pProposalTree->getRoot(), vpDnaEvolutionModel,
				m_pState->m_vRateScaling, m_modelNr);
		m_proposalTreeLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodTree(
				m_pProposalTree, vpDnaEvolutionModel);
	}

	double ChangeMemStatDist::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood - m_pState->m_treeLogLikelihood);
	}

	double ChangeMemStatDist::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void ChangeMemStatDist::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
		delete m_pState->m_vpDnaEvolutionModel[m_modelNr];
		m_pState->m_vpDnaEvolutionModel[m_modelNr] = m_pDnaEvolutionModel;
		m_pState->m_pTree				= m_pProposalTree;
		m_pState->m_totalLogLikelihood += (m_proposalTreeLogLikelihood -
				m_pState->m_treeLogLikelihood);
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
	}

	void ChangeMemStatDist::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
		delete m_pDnaEvolutionModel;
	}

	void ChangeMemStatDist::allocateSubtreeMemory(
			TreeTime_MolecularTree::iterator<NodeData> p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LikelihoodCalculations::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	void ChangeMemStatDist::allocateAndCopySubtreeMemory(
			TreeTime_MolecularTree::iterator<NodeData> p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LikelihoodCalculations::allocateAndCopyMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateAndCopySubtreeMemory(p_iterator.getFirstChild());
			allocateAndCopySubtreeMemory(p_iterator.getLastChild());
		}
	}

	int ChangeMemStatDist::getId () const
	{
		return 5;
	}

}//TreeTime_Core
