/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "ChangeNodeTimepoint.hpp"
#include "../RandomNumberGenerator.hpp"
#include "../LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"

namespace TreeTime_Core
{

	ChangeNodeTimepoint::ChangeNodeTimepoint(State* p_pState,
			double p_proposalRatio, double p_calibration)
	: m_pState							(p_pState)
	, m_proposalRatio					(p_proposalRatio)
	, m_pProposalTree					(0)
	, m_proposalTreeLogLikelihood		(0)
	, m_proposalGlobalRateLogLikelihood	(p_pState->m_globalRateLoglikelihood)
	, m_vProposalLocalRateLoglikelihood (p_pState->m_vLocalRateLoglikelihood)
	, m_proposalConstraintLogLikelihood	(0)
	, m_proposalTotalLogLikelihood		(0)
	, m_calibration						(p_calibration)
	, m_lpCurrentProbablities			()
	, m_lpProposalProbablities			()
	{
	}

	ChangeNodeTimepoint::~ChangeNodeTimepoint()
	{
	}

	void ChangeNodeTimepoint::changeState()
	{
		m_pProposalTree = m_pState->m_pTree->copy();
		list<TreeTime_MolecularTree::iterator<NodeData> > InnerNodes;
		TreeTime_MolecularTree::iterator<NodeData> I = m_pProposalTree->
				getRoot();
		stack<TreeTime_MolecularTree::iterator<NodeData> > sIterator;
		if(I.hasChildNodes())
		{
			if(I.getFirstChild().hasChildNodes())
				sIterator.push(I.getFirstChild());
			if(I.getLastChild().hasChildNodes())
				sIterator.push(I.getLastChild());
		}
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(I.hasChildNodes())
				InnerNodes.push_back(I);
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		int pos = RandomNumberGenerator::getRNG().getInteger(InnerNodes.size());
		list<TreeTime_MolecularTree::iterator<NodeData> >::iterator iNodes =
				InnerNodes.begin();
		for(int i = 0; i<pos; ++i)
			++iNodes;
		I = *iNodes;
		TreeTime_MolecularTree::iterator<NodeData> insertLocation = I;
		while(I)
		{
			m_lpCurrentProbablities.push_back(I->m_molecularData);
			LikelihoodCalculations::allocateMemory(I);
			m_lpProposalProbablities.push_back(I->m_molecularData);
			I = I.getFather();
		}
		double maxi = insertLocation.getFather()->m_timepoint;
		double mini = insertLocation.getFirstChild()->m_timepoint;
		mini = std::max(mini, insertLocation.getLastChild()->m_timepoint);
		double le = m_calibration*(maxi - mini)/4.0;
		double newTimepoint = insertLocation->m_timepoint +
				RandomNumberGenerator::getRNG().getUniform(-le,le);
		if(newTimepoint>maxi)
				newTimepoint = maxi + maxi - newTimepoint;
		if(newTimepoint<mini)
				newTimepoint = mini + mini - newTimepoint;
		insertLocation->m_timepoint = newTimepoint;
		scaleBranchLength(insertLocation,
				(maxi-newTimepoint)/(insertLocation->m_length));
		scaleBranchLength(insertLocation.getFirstChild(),
				(newTimepoint-insertLocation.getFirstChild()->m_timepoint)
				/(insertLocation.getFirstChild()->m_length));
		scaleBranchLength(insertLocation.getLastChild(),
				(newTimepoint-insertLocation.getLastChild()->m_timepoint)
				/(insertLocation.getLastChild()->m_length));
		LikelihoodCalculations::calcScaledUpwardsDnaProbabilities(
				insertLocation, m_pState->m_vpDnaEvolutionModel,
				m_pState->m_vRateScaling);
		m_proposalTreeLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodTree(
				m_pProposalTree, m_pState->m_vpDnaEvolutionModel);			
		m_proposalTotalLogLikelihood += m_proposalTreeLogLikelihood;
		TreeTime_MolecularTree::Tree<NodeData>*	pCurrentTree =m_pState->m_pTree;
		m_pState->m_pTree = m_pProposalTree;
		m_proposalGlobalRateLogLikelihood = m_pState->m_pGlobalRateChangeModel->
				getLogDensity(m_pState,-1);
		m_proposalTotalLogLikelihood += m_proposalGlobalRateLogLikelihood;
		for(unsigned int i=0; i < m_vProposalLocalRateLoglikelihood.size(); ++i)
		{
			m_vProposalLocalRateLoglikelihood[i] = m_pState->
					m_vpLocalRateChangeModel[i]->getLogDensity(m_pState,i);
			m_proposalTotalLogLikelihood +=m_vProposalLocalRateLoglikelihood[i];
		}
		m_pState->m_pTree = pCurrentTree;
		m_proposalConstraintLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodConstraints(
				m_pProposalTree, m_pState->s_vpConstraint);
		m_proposalTotalLogLikelihood += m_proposalConstraintLogLikelihood;
		/* todo: yule prior with correct intensity:
		double intensity = 1.0 / 50.0;
		double treeLength_current  = LikelihoodCalculations::getTreeLength(pCurrentTree);
		double treeLength_proposal = LikelihoodCalculations::getTreeLength(m_pProposalTree);
		m_proposalRatio *= exp(intensity*(treeLength_current-treeLength_proposal));
		*/
	}

	double ChangeNodeTimepoint::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTotalLogLikelihood-m_pState->m_totalLogLikelihood);
	}

	double ChangeNodeTimepoint::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void ChangeNodeTimepoint::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
		m_pState->m_pTree				= m_pProposalTree;
		m_pState->m_treeLogLikelihood	= m_proposalTreeLogLikelihood;
		m_pState->m_globalRateLoglikelihood = m_proposalGlobalRateLogLikelihood;
		m_pState->m_vLocalRateLoglikelihood = m_vProposalLocalRateLoglikelihood;
		m_pState->m_constraintLogLikelihood	= m_proposalConstraintLogLikelihood;
		m_pState->m_totalLogLikelihood	= m_proposalTotalLogLikelihood;
	}

	void ChangeNodeTimepoint::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	void ChangeNodeTimepoint::scaleBranchLength(
			TreeTime_MolecularTree::iterator<NodeData> p_I, double p_factor)
	{
		LikelihoodCalculations::scaleInnerBranch(p_I,p_factor);
	}

	int ChangeNodeTimepoint::getId () const
	{
		return 2;
	}

}//TreeTime_Core
