/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "ChangeRootTimepoint.hpp"
#include "../RandomNumberGenerator.hpp"
#include "../LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"

namespace TreeTime_Core
{

	ChangeRootTimepoint::ChangeRootTimepoint(State* p_pState,
			double p_proposalRatio, double p_expectedRootRealTime,
			double p_calibration)
	: m_pState							(p_pState)
	, m_proposalRatio					(p_proposalRatio)
	, m_expectedRootRealTime			(p_expectedRootRealTime)
	, m_calibration						(p_calibration)
	, m_logRootTimePriorRatio			(0)
	, m_pProposalTree					(0)
	, m_proposalTreeLogLikelihood		(0)
	, m_proposalGlobalRateLogLikelihood	(p_pState->m_globalRateLoglikelihood)
	, m_vProposalLocalRateLoglikelihood	(p_pState->m_vLocalRateLoglikelihood)
	, m_proposalConstraintLogLikelihood (0)
	, m_proposalTotalLogLikelihood		(0)
	, m_lpCurrentProbablities			()
	, m_lpProposalProbablities			()
	{
	}

	ChangeRootTimepoint::~ChangeRootTimepoint()
	{
	}

	void ChangeRootTimepoint::changeState()
	{
		m_pProposalTree = m_pState->m_pTree->copy();
		TreeTime_MolecularTree::iterator<NodeData> Root = 
				m_pProposalTree->getRoot();
		m_lpCurrentProbablities.push_back(Root->m_molecularData);
		LikelihoodCalculations::allocateMemory(Root);
		m_lpProposalProbablities.push_back(Root->m_molecularData);
		double mini = Root.getFirstChild()->m_timepoint;
		mini = std::max(mini, Root.getLastChild()->m_timepoint);
		double newTimepoint = Root->m_timepoint +
				RandomNumberGenerator::getRNG().getUniform(
				- 0.3 * m_calibration * m_expectedRootRealTime, 
				0.3 * m_calibration * m_expectedRootRealTime);
		if(newTimepoint < mini)
			newTimepoint = mini + mini - newTimepoint;
		// Exponential Prior for Root Timepoint:
		m_logRootTimePriorRatio = (Root->m_timepoint-newTimepoint)/
				m_expectedRootRealTime;
		Root->m_timepoint = newTimepoint;
		scaleBranchLength(Root.getFirstChild(), 
				(newTimepoint-Root.getFirstChild()->m_timepoint)
				/(Root.getFirstChild()->m_length));
		scaleBranchLength(Root.getLastChild(), 
				(newTimepoint-Root.getLastChild()->m_timepoint)
				/(Root.getLastChild()->m_length));
		LikelihoodCalculations::calcScaledUpwardsDnaProbabilities(
				Root, m_pState->m_vpDnaEvolutionModel,
				m_pState->m_vRateScaling);
		m_proposalTreeLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodTree(
				m_pProposalTree, m_pState->m_vpDnaEvolutionModel);
		m_proposalTotalLogLikelihood += m_proposalTreeLogLikelihood;
		TreeTime_MolecularTree::Tree<NodeData>*	pCurrentTree= m_pState->m_pTree;
		m_pState->m_pTree = m_pProposalTree;
		m_proposalGlobalRateLogLikelihood = m_pState->m_pGlobalRateChangeModel->
				getLogDensity(m_pState,-1);
		m_proposalTotalLogLikelihood += m_proposalGlobalRateLogLikelihood;
		for(unsigned int i=0; i < m_vProposalLocalRateLoglikelihood.size(); ++i)
		{
			m_vProposalLocalRateLoglikelihood[i] = m_pState->
					m_vpLocalRateChangeModel[i]->getLogDensity(m_pState,i);
			m_proposalTotalLogLikelihood +=m_vProposalLocalRateLoglikelihood[i];
		}
		m_pState->m_pTree = pCurrentTree;
		m_proposalConstraintLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodConstraints(
				m_pProposalTree, m_pState->s_vpConstraint);
		m_proposalTotalLogLikelihood += m_proposalConstraintLogLikelihood;
		/* todo: yule prior with correct intensity:
		double intensity = 1.0 / 50.0;
		double treeLength_current  = LikelihoodCalculations::getTreeLength(pCurrentTree);
		double treeLength_proposal = LikelihoodCalculations::getTreeLength(m_pProposalTree);
		m_proposalRatio *= exp(intensity*(treeLength_current-treeLength_proposal));
		*/
	}

	double ChangeRootTimepoint::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTotalLogLikelihood-m_pState->m_totalLogLikelihood
				+ m_logRootTimePriorRatio);
	}

	double ChangeRootTimepoint::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void ChangeRootTimepoint::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
		m_pState->m_pTree				= m_pProposalTree;
		m_pState->m_treeLogLikelihood	= m_proposalTreeLogLikelihood;
		m_pState->m_globalRateLoglikelihood = m_proposalGlobalRateLogLikelihood;
		m_pState->m_vLocalRateLoglikelihood = m_vProposalLocalRateLoglikelihood;
		m_pState->m_constraintLogLikelihood	= m_proposalConstraintLogLikelihood;
		m_pState->m_totalLogLikelihood	= m_proposalTotalLogLikelihood;

	}

	void ChangeRootTimepoint::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	void ChangeRootTimepoint::scaleBranchLength(
			TreeTime_MolecularTree::iterator<NodeData> p_I, double p_factor)
	{
		LikelihoodCalculations::scaleInnerBranch(p_I, p_factor);
	}

	int ChangeRootTimepoint::getId () const
	{
		return 3;
	}

}//TreeTime_Core
