/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "ChangeTopology.hpp"
#include "../RandomNumberGenerator.hpp"
#include "../LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"

namespace TreeTime_Core
{

	ChangeTopology::ChangeTopology(State* p_pState, double p_proposalRatio,
			bool p_hasOutgroup)
	: m_pState							(p_pState)
	, m_proposalRatio					(p_proposalRatio)
	, m_hasOutgroup						(p_hasOutgroup)
	, m_pProposalTree					(0)
	, m_proposalTreeLogLikelihood		(0)
	, m_proposalGlobalRateLogLikelihood	(p_pState->m_globalRateLoglikelihood)
	, m_vProposalLocalRateLoglikelihood	(p_pState->m_vLocalRateLoglikelihood)
	, m_proposalConstraintLogLikelihood	(0)
	, m_proposalTotalLogLikelihood		(0)
	, m_lpCurrentProbablities			()
	, m_lpProposalProbablities			()
	{
	}

	ChangeTopology::~ChangeTopology()
	{
	}

	void ChangeTopology::changeState()
	{
		m_pProposalTree = m_pState->m_pTree->copy();
		//todo: take this in the state, then you have to calculate it only once...
		set<TreeTime_MolecularTree::iterator<NodeData> > RotateableNodes =
				LikelihoodCalculations::getRotateableNodes(
				m_pProposalTree,m_pState->s_vpConstraint,m_hasOutgroup);
		int nRotateableNodes = RotateableNodes.size();
		if(nRotateableNodes<1)
		{
			m_proposalTotalLogLikelihood = -100.0 + m_pState->
					m_totalLogLikelihood;
			cout << "ChangeTopology: no Rotations to perfom" << endl;
			return;
		}
		int pos = RandomNumberGenerator::getRNG().getInteger(nRotateableNodes);
		set<TreeTime_MolecularTree::iterator<NodeData> >::iterator iNodes =
				RotateableNodes.begin();
		for(int i = 0; i < pos; ++i)
			++iNodes;
		TreeTime_MolecularTree::iterator<NodeData>	grandchild = *iNodes;
		TreeTime_MolecularTree::iterator<NodeData>	child  = 
				grandchild.getFather();
		TreeTime_MolecularTree::iterator<NodeData>	father = child.getFather();
		TreeTime_MolecularTree::iterator<NodeData>	uncle  = 
				child.getFather().getFirstChild();
		if(father.getFirstChild() == child)
		{
			uncle = child.getFather().getLastChild();
		}
		if(father.getFirstChild() == child)
			father.setLastChild(grandchild);
		else
			father.setFirstChild(grandchild);
		if(child.getFirstChild() == grandchild)
			child.setFirstChild(uncle);
		else
			child.setLastChild(uncle);
		grandchild.setFather(father);
		uncle.setFather(child);
		double scaling = father->m_timepoint/child->m_timepoint;
		LikelihoodCalculations::scaleSubTree(grandchild, scaling);
		LikelihoodCalculations::scaleSubTree(uncle, 1.0/scaling);
		list<int> lCppModels;
		for(map<int,list<pair<double,double> > >::iterator I = child->
				m_mlChangepoints.begin(); I != child->m_mlChangepoints.end();
				++I)
		{
			lCppModels.push_back(I->first);
		}
		for(list<int>::iterator I = lCppModels.begin(); I!=lCppModels.end();++I)
		{
			if(child->m_mlChangepoints[*I].size() > 2)
			{
				double initialRatio = grandchild->m_mlChangepoints[*I].begin()->
						second/uncle->m_mlChangepoints[*I].begin()->second;
				adjustSubtreeInitialRates(grandchild, *I, 1.0/initialRatio);
				adjustSubtreeInitialRates(uncle, *I, initialRatio);
			}
		}
		allocateSubtreeMemory(grandchild);
		allocateSubtreeMemory(uncle);
		TreeTime_MolecularTree::iterator<NodeData> I = child;
		while(I)
		{
			m_lpCurrentProbablities.push_back(I->m_molecularData);
			LikelihoodCalculations::allocateMemory(I);
			m_lpProposalProbablities.push_back(I->m_molecularData);
			I = I.getFather();
		}
		LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
				grandchild, m_pState->m_vpDnaEvolutionModel,
				m_pState->m_vRateScaling);	
		LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
				uncle, m_pState->m_vpDnaEvolutionModel,
				m_pState->m_vRateScaling);
		LikelihoodCalculations::calcScaledUpwardsDnaProbabilities(
				child, m_pState->m_vpDnaEvolutionModel,
				m_pState->m_vRateScaling);
		m_proposalTreeLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodTree(
				m_pProposalTree, m_pState->m_vpDnaEvolutionModel);
		m_proposalTotalLogLikelihood += m_proposalTreeLogLikelihood;
		TreeTime_MolecularTree::Tree<NodeData>*	pCurrentTree= m_pState->m_pTree;
		m_pState->m_pTree = m_pProposalTree;
		m_proposalGlobalRateLogLikelihood = m_pState->m_pGlobalRateChangeModel->
				getLogDensity(m_pState,-1);
		m_proposalTotalLogLikelihood += m_proposalGlobalRateLogLikelihood;
		for(unsigned int i=0; i < m_vProposalLocalRateLoglikelihood.size(); ++i)
		{
			m_vProposalLocalRateLoglikelihood[i] = m_pState->
					m_vpLocalRateChangeModel[i]->getLogDensity(m_pState,i);
			m_proposalTotalLogLikelihood +=m_vProposalLocalRateLoglikelihood[i];
		}
		m_pState->m_pTree = pCurrentTree;
		m_proposalConstraintLogLikelihood =
				LikelihoodCalculations::getLogLikelihoodConstraints(
				m_pProposalTree, m_pState->s_vpConstraint);
		m_proposalTotalLogLikelihood += m_proposalConstraintLogLikelihood;
		int nNewRotateableNodes = (LikelihoodCalculations::getRotateableNodes(
				m_pProposalTree,m_pState->s_vpConstraint,m_hasOutgroup)).size();
		m_proposalRatio *=  (static_cast<double>(nRotateableNodes)/
				static_cast<double>(nNewRotateableNodes));
		/* todo: yule prior with correct intensity:
		double intensity = 1.0 / 50.0;
		double treeLength_current  = LikelihoodCalculations::getTreeLength(pCurrentTree);
		double treeLength_proposal = LikelihoodCalculations::getTreeLength(m_pProposalTree);
		m_proposalRatio *= exp(intensity*(treeLength_current-treeLength_proposal));
		*/
	}
	double ChangeTopology::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTotalLogLikelihood-m_pState->m_totalLogLikelihood);
	}

	double ChangeTopology::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void ChangeTopology::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
		m_pState->m_pTree				= m_pProposalTree;
		m_pState->m_treeLogLikelihood	= m_proposalTreeLogLikelihood;
		m_pState->m_globalRateLoglikelihood = m_proposalGlobalRateLogLikelihood;
		m_pState->m_vLocalRateLoglikelihood = m_vProposalLocalRateLoglikelihood;
		m_pState->m_constraintLogLikelihood	= m_proposalConstraintLogLikelihood;
		m_pState->m_totalLogLikelihood	= m_proposalTotalLogLikelihood;
	}

	void ChangeTopology::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	void ChangeTopology::allocateSubtreeMemory(
			TreeTime_MolecularTree::iterator<NodeData> p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LikelihoodCalculations::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	void ChangeTopology::adjustSubtreeInitialRates(TreeTime_MolecularTree::
			iterator<NodeData> p_iterator, int p_loci, double p_multiplier)
	{
		p_iterator->m_mlChangepoints[p_loci].begin()->second *= p_multiplier;
		if(p_loci == -1)
		{
			p_iterator->m_globalRateMultiplier *= p_multiplier;
		}
		else
		{
			p_iterator->m_vLocalRateMultiplier[p_loci] *= p_multiplier;
		}
		if(p_iterator.hasChildNodes())
		{
			adjustSubtreeInitialRates(p_iterator.getFirstChild(),p_loci,
					p_multiplier);
			adjustSubtreeInitialRates(p_iterator.getLastChild(),p_loci,
					p_multiplier);
		}
	}

	int ChangeTopology::getId () const
	{
		return 1;
	}
	
}//TreeTime_Core
