/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "StateSampler.hpp"
#include "State.hpp"
#include "../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../TreeTime_RateChangeModel/CompoundPoissonProcess.hpp"
#include "../TreeTime_RateChangeModel/MolecularClock.hpp"

namespace TreeTime_Core
{
	StateSampler::StateSampler(map<int,string>& p_mIdToName, string p_outputDir,
			const int p_nChains)
	: m_mIdToName	(p_mIdToName)
	, m_outputDir	(p_outputDir)
	, m_nChains		(p_nChains)
	, m_countSwappings(p_nChains, vector<pair<int,int> >
			(p_nChains, pair<int,int>(0,0)))
	, m_countChanges(p_nChains, vector<pair<int,int> >
			(1+22, pair<int,int>(0,0)))
	, m_mode		(std::ios::out)
	, m_outTreesUM	((m_outputDir + "tt_treesUM.tre").c_str(),		m_mode)
	, m_outTreesUR	((m_outputDir + "tt_treesUR.tre").c_str(),		m_mode)
	, m_outState	((m_outputDir + "tt_stateProbs.txt").c_str(),	m_mode)
	, m_outRcms		((m_outputDir + "tt_rcms.txt").c_str(),			m_mode)
	, m_outMems		((m_outputDir + "tt_mems.txt").c_str(),			m_mode)
	//, m_outCp		((m_outputDir + "tt_changePoints.txt").c_str(),	m_mode)
	, m_outChanges	((m_outputDir + "tt_changes.txt").c_str(),		m_mode)
	, m_outSwaps	((m_outputDir + "tt_swaps.txt").c_str(),		m_mode)
	, m_nrOfSample  (0)
	{
		m_outTreesUM.close();
		m_outTreesUM.open((m_outputDir+"tt_treesUM.tre").c_str(),std::ios::app);
		m_outTreesUR.close();
		m_outTreesUR.open((m_outputDir+"tt_treesUR.tre").c_str(),std::ios::app);
		m_outState.close();
		m_outState.open((m_outputDir+"tt_stateProbs.txt").c_str(),std::ios::app);
		m_outRcms.close();
		m_outRcms.open((m_outputDir+"tt_rcms.txt").c_str(),std::ios::app);
		m_outMems.close();
		m_outMems.open((m_outputDir+"tt_mems.txt").c_str(),std::ios::app);
		//m_outCp.close();
		//m_outCp.open((m_outputDir+"tt_changePoints.txt").c_str(),std::ios::app);
		m_outChanges.close();
		m_outChanges.open((m_outputDir+"tt_changes.txt").c_str(),std::ios::app);
		m_outSwaps.close();
		m_outSwaps.open((m_outputDir+"tt_swaps.txt").c_str(),std::ios::app);
	}

	StateSampler::~StateSampler()
	{
	}

	void StateSampler::sample(const State* const p_pState)
	{
		m_nrOfSample++;
		cout << "Sampled state nr." << std::setw(4) << std::setfill('0') << m_nrOfSample << ", "
		     << "log posterior density : " << p_pState->m_totalLogLikelihood << endl; 
		
		m_outTreesUM << getTreeUM(p_pState->m_pTree->getRoot()) << ";" << endl;


		m_outTreesUR << getTreeUR(p_pState->m_pTree->getRoot(), -1) << ";" << endl;
		for(unsigned int i=0; i<p_pState->m_vLocalRateLoglikelihood.size();++i)
			m_outTreesUR << getTreeUR(p_pState->m_pTree->getRoot(), i) << ";"
					<< endl;

		m_outState	<< p_pState->m_totalLogLikelihood		<< "\t"
					<< p_pState->m_treeLogLikelihood		<< "\t"
					<< p_pState->m_constraintLogLikelihood	<< "\t"
					<< p_pState->m_globalRateLoglikelihood	<< "\t";
		for(unsigned int i=0; i<p_pState->m_vLocalRateLoglikelihood.size();++i)
			m_outState << p_pState->m_vLocalRateLoglikelihood[i] << "\t";
		m_outState	<< endl;

		m_outRcms << *p_pState->m_pGlobalRateChangeModel	<< "\t";
		for(unsigned int i=0; i<p_pState->m_vpLocalRateChangeModel.size();++i)
			m_outRcms << *p_pState->m_vpLocalRateChangeModel[i] << "\t";
		m_outRcms << endl;

		for(unsigned int i=0; i<p_pState->m_vpDnaEvolutionModel.size(); ++i)
		{
			m_outMems	<< *p_pState->m_vpDnaEvolutionModel[i]	<< "\t"
						<< p_pState->m_vRateScaling[i]			<< "\t";
		}
		m_outMems << endl;
	}

	string StateSampler::getTreeUM(TreeTime_MolecularTree::iterator<NodeData>
			p_iterator)
	{
		stringstream ss;
		if(p_iterator.hasChildNodes())
		{
			ss	<< "(" << getTreeUM(p_iterator.getFirstChild())
				<< "," << getTreeUM(p_iterator.getLastChild())
				<< "):"<< p_iterator->m_length;
		}
		else
		{
			ss	<< m_mIdToName[p_iterator->m_id] << ":"
				<< p_iterator->m_length;
		}
		return ss.str();
	}

	string StateSampler::getTreeUR(TreeTime_MolecularTree::iterator<NodeData>
			p_iterator, int p_loci)
	{
		stringstream ss;
		if(p_iterator.hasChildNodes())
		{
			ss	<< "(" << getTreeUR(p_iterator.getFirstChild(), p_loci)
				<< "," << getTreeUR(p_iterator.getLastChild(), p_loci)
				<< "):"<< getScaledLength(p_iterator, p_loci);
		}
		else
		{
			ss	<< m_mIdToName[p_iterator->m_id] << ":"
				<< getScaledLength(p_iterator, p_loci);
		}
		return ss.str();
	}

	double StateSampler::getScaledLength(TreeTime_MolecularTree::iterator
			<NodeData> p_I, int p_loci)
	{
		if(p_loci == -1)
			return p_I->m_length * p_I->m_globalRateMultiplier;
		else
			return p_I->m_length * p_I->m_globalRateMultiplier *
					p_I->m_vLocalRateMultiplier[p_loci];
	}

	void StateSampler::monitorChange(const int p_chain, const int p_changeId,
			const bool p_isAccepted)
	{
		if(p_isAccepted)
			++ m_countChanges[p_chain][p_changeId].first;
		else
			++ m_countChanges[p_chain][p_changeId].second;
	}

	void StateSampler::monitorSwap(const int p_chain, const int p_chain2,
			const bool p_isAccepted)
	{
		if(p_isAccepted)
			++ m_countSwappings[p_chain][p_chain2].first;
		else
			++ m_countSwappings[p_chain][p_chain2].second;
	}

	void StateSampler::finishOutput()
	{
		for(unsigned int i = 0; i < m_countChanges.size(); ++i)
		{
			for(unsigned int j = 1; j < m_countChanges[0].size(); ++j)
			{
				m_outChanges	<< j << "\t" << m_countChanges[i][j].first
								<< "\t" << m_countChanges[i][j].second << endl;
			}
		}
		for(int i = 0; i < m_nChains; ++i)
		{
			for(int j = 0; j < m_nChains; ++j)
			{
				if(m_countSwappings[i][j].first == 0)
					m_outSwaps	<< 0 << "\t";
				else
					m_outSwaps << static_cast<double>(m_countSwappings[i][j]
					.first)/(m_countSwappings[i][j].first+m_countSwappings[i][j]
					.second) <<"\t";
			}
			m_outSwaps << endl;
		}
	}

	void StateSampler::getChangePointValues(
			TreeTime_MolecularTree::iterator<NodeData> p_I,
			list<double>& p_lChangePoints)
	{
		stack<TreeTime_MolecularTree::iterator<NodeData> > sIterator;
		sIterator.push(p_I);
		while(!sIterator.empty())
		{
			TreeTime_MolecularTree::iterator<NodeData> I = sIterator.top();
			sIterator.pop();
//todo
//			if(I->m_lChangepoints.size()>2)
//			{
//				list<pair<double,double> >::iterator lI =
//						++I->m_lChangepoints.begin();
//				list<pair<double,double> >::iterator lE =
//						--I->m_lChangepoints.end();
//				for(; lI != lE; ++lI)
//				{
//					p_lChangePoints.push_back(lI->second);
//				}
//			}
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
	}

}//TreeTime_Core
