/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "TreeTime.hpp"
#include "RandomNumberGenerator.hpp"
#include "StateManager.hpp"
#include "State.hpp"
#include "StateSampler.hpp"
#include "ProposeChange.hpp"
#include "NodeData.hpp"
#include "LikelihoodCalculations.hpp"
#include "../TreeTime_Mcmcmc/McmcmcData.hpp"
#include "../TreeTime_Mcmcmc/Mcmcmc.hpp"
#include "../TreeTime_NexusProcessing/NexusParser.hpp"
#include "../TreeTime_NexusProcessing/Data.hpp"
#include "../TreeTime_MolecularTree/BinaryTree.hpp"
#include "../TreeTime_MolecularTree/iterator.hpp"
#include "../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../TreeTime_RateChangeModel/CompoundPoissonProcess.hpp"
#include <gsl/gsl_errno.h>

namespace TreeTime_Core
{
	void TreeTime_error_handler(const char *reason, const char *file, int line, int gslErrorId)
	{
		cout << "GSL exception: " << reason << " in " <<file << ", l. " << line << endl;
		// todo: Disable this for macOs, you can Enable this WindowsXP, Linux, etc.
		//throw(gslErrorId);
	}

	void TreeTime::run(string p_inputFilePath)
	{
		// Set TreeTime Exception Handler for Gsl exceptions
		gsl_set_error_handler (&TreeTime_error_handler);
		// Don't turn the errors and warninhgs off:
		// gsl_set_error_handler_off();
		// Using directives
		using TreeTime_NexusProcessing::Data;
		using TreeTime_NexusProcessing::NexusParser;

		cout <<                                                                       endl;
		cout << "                    TreeTime v1.0                               " << endl;
		cout <<                                                                       endl;
		cout << "           Bayesian Estimation of Phylogenies                   " << endl;
		cout << "      with real Branching Times from molecular Data             " << endl;
		cout <<                                                                       endl;
		cout << "                          by                                    " << endl;
		cout << "           Lin M. Himmelmann and Dirk Metzler                   " << endl;
		cout << "                    25. April 2009                              " << endl;
		cout <<                                                                       endl;
		cout << "             Scientific Software Development                    " << endl;
		cout << "            www.linhi.com, treetime@linhi.com                   " << endl;
		cout <<                                                                       endl;
		cout << "   Ludwig-Maximilians-Universitaet Muenchen, BioCenter          " << endl;
		cout << "            metzler@biologie.uni-muenchen.de                    " << endl;
		cout <<                                                                       endl;
		cout << " www.zi.biologie.uni-muenchen.de/evol/statgen/software/treetime " << endl;
		cout << "      Distributed under the GNU General Public License          " << endl;
		cout <<                                                                       endl;
		cout <<                                                                       endl;
		cout << "--------------------------------------------------------------------" << endl;
		cout << "Parsing nexus-file:"                                              << endl;
		cout << "--------------------------------------------------------------------" << endl;
		// init Data
		Data* 			pData	= new Data();
		NexusParser* 	pParser	= new NexusParser(pData);
		pParser->parseNexusFile(p_inputFilePath);
		delete pParser;
		cout << "Parsing nexus-file finished."                                     << endl;
		cout <<                                                                       endl;
		cout << "--------------------------------------------------------------------" << endl;
		cout << "Setup analysis:"                                                  << endl;
		cout << "--------------------------------------------------------------------" << endl;
		// Initilise static Elements
		RandomNumberGenerator* pRng = &RandomNumberGenerator::getRNG();
		pRng->setSeed(pData->getSeed());
		LikelihoodCalculations::init(*pData);
		// initialise the initial State from the Data
		State* pState = new State();
		State::s_vpConstraint = pData->getConstraints();
		pState->m_pTree = new TreeTime_MolecularTree::BinaryTree<NodeData>(
				pData->getInitialTree(), pData->getSpecies2Id());
		LikelihoodCalculations::initTimepoints(pState->m_pTree->getRoot());
		bool hasOutgroup = false;
		if(pData->getOutgroup() >= 0)
		{
			LikelihoodCalculations::rootWithOutgroup(pState->m_pTree,
					pData->getOutgroup());
			hasOutgroup = true;
			cout << " outgroup:" << (pData->getId2Species()[pData->getOutgroup()]) << endl;
		}
		else
		{
			cout << " no outgroup specified."                                      << endl;
		}
		double scaling = pData->initScaling(pState->m_pTree);
		LikelihoodCalculations::scaleSubTree(pState->m_pTree->getRoot(),
				scaling);
		double pemTime = pData->estimatePEMTime();
		double rootTime= pState->m_pTree->getRoot()->m_timepoint;
		cout << " initial timepoint of root: "  << std::setprecision(3) <<rootTime << endl;
		cout << " estimated PEM-time of root: " << std::setprecision(3) << pemTime << endl;
		for(unsigned int i = 0; i<pData->getInitialMolecularModels().size();++i)
		{
			pState->m_vRateScaling.push_back(pemTime/rootTime);
		}
		pState->m_vpDnaEvolutionModel	= pData->getInitialMolecularModels();
		cout << " initialise molecular evolution model...";
		for(unsigned int i = 0; i<pData->getInitialMolecularModels().size();++i)
		{
			pState->m_vpDnaEvolutionModel[i] =
					(pData->getInitialMolecularModels())[i]->clone();
		}
		cout << "done." << endl;
		TreeTime_MolecularEvolution::DnaEvolutionModel::setRandom(pRng);
		//initRateChangeModels.
		cout << " initialise rate change model...";
		pState->m_pGlobalRateChangeModel = TreeTime_RateChangeModel::
				RateChangeModel<State>::newRateChangeModel(
				pData->getInitialGlobalRateChangeModel());
		pState->m_pGlobalRateChangeModel->initRateChanges(
				pState, -1);
		cout << "done." << endl;
		pState->m_vpLocalRateChangeModel.clear();
		for(unsigned int i = 0; i<pData->getInitialMolecularModels().size();++i)
		{
			pState->m_vpLocalRateChangeModel.push_back(
					TreeTime_RateChangeModel::
					RateChangeModel<State>::newRateChangeModel(
					(pData->getInitialRateChangeModels())[i]));
			((pState->m_vpLocalRateChangeModel)[i])->initRateChanges(pState,i);
		}
		// Calculate Initial Probabilities:
		cout << " allocate and initialise memory...";
		LikelihoodCalculations::assignLeafMemory(pState->m_pTree);
		LikelihoodCalculations::allocateSubtreeMemory(
				pState->m_pTree->getRoot());
		cout << "done." << endl;
		cout << " calculate initial prob. densities: ";
		cout << "tree..." ;
		LikelihoodCalculations::initRateMultipliers(pState->m_pTree->getRoot(),
				pData->getInitialMolecularModels().size());
		LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
				pState->m_pTree->getRoot(), pState->m_vpDnaEvolutionModel,
				pState->m_vRateScaling);
		pState->m_treeLogLikelihood			=
				LikelihoodCalculations::getLogLikelihoodTree
				(pState->m_pTree,pState->m_vpDnaEvolutionModel);
		cout << "constraints..." ;
		pState->m_constraintLogLikelihood	= 0.0;
		if(!State::s_vpConstraint.empty())
		{
			pState->m_constraintLogLikelihood = 
					LikelihoodCalculations::getLogLikelihoodConstraints(
					pState->m_pTree, State::s_vpConstraint);
		}
		cout << "evolution..." ;
		pState->m_globalRateLoglikelihood = pState->m_pGlobalRateChangeModel->
				getLogDensity(pState,-1);
		for(unsigned int i = 0; i<pData->getInitialMolecularModels().size();++i)
		{
			pState->m_vLocalRateLoglikelihood.push_back(
			((pState->m_vpLocalRateChangeModel)[i])->getLogDensity(pState,i)
			);
		}
		pState->m_totalLogLikelihood = pState->m_treeLogLikelihood +
				pState->m_constraintLogLikelihood +
				pState->m_globalRateLoglikelihood;
		for(unsigned int i = 0; i<pData->getInitialMolecularModels().size();++i)
		{
			pState->m_totalLogLikelihood +=
					(pState->m_vLocalRateLoglikelihood)[i];
		}
		cout << "done."                  << endl;
		// initialise one StateManager
		cout << " initialise state...";
		StateManager*	pStateManager	= new StateManager(pState);
		cout << "done."                  << endl;
		// set Changes
		cout << " initialise proposal moves...";
		if(!pData->getMemChanges())
			 pData->setWeightEvolution(0.0);
		if(!pData->getRateChanges())
			pData->setWeightRatechange(0.0);
		ProposeChange*	pProposeChange	= new ProposeChange(pemTime,
				pData->hasTimeConstraints(), rootTime, pData->getNTaxa(),
				hasOutgroup,
				pData->getWeightTopology(), pData->getWeightNodes(),
				pData->getWeightEvolution(), pData->getWeightRatechange(),
				pData->getWeightTimeconversion(), 
				pData->getNodeTime(),
				pData->getTimeSynchronisation(),
				pData->getRootTime(),
				pState);
		StateManager::setProposeChange(pProposeChange);
		cout << "done."                  << endl;
		cout << " initialise state sampler...";
		// set Sampler
		StateSampler*	pStateSampler	= new StateSampler(
				pData->getId2Species(), pData->getOutputPath(),
				pData->getNChains());
		StateManager::setStateSampler(pStateSampler);
		cout << "done."                  << endl;
		// delete initialising State elements
		delete pState->m_pTree;
		for(unsigned int i = 0; i<pData->getInitialMolecularModels().size();++i)
			delete pState->m_vpDnaEvolutionModel[i];
		delete pState->m_pGlobalRateChangeModel;
		for(unsigned int i = 0; i<pData->getInitialMolecularModels().size();++i)
			delete pState->m_vpLocalRateChangeModel[i];
		delete pState;
		// initialise Mcmcmc Application
		vector<TreeTime_Mcmcmc::StateManager*> vpStateManager;
		vpStateManager.push_back(pStateManager);
		for(int i = 1; i<pData->getNChains(); ++i)
		{
			StateManager* tempStateManager = new StateManager(*pStateManager);
			vpStateManager.push_back(tempStateManager);
			LikelihoodCalculations::allocateSubtreeMemory(
					tempStateManager->getState()->m_pTree->getRoot());
			LikelihoodCalculations::calcScaledSubtreeDnaProbabilities(
					tempStateManager->getState()->m_pTree->getRoot(),
					tempStateManager->getState()->m_vpDnaEvolutionModel,
					tempStateManager->getState()->m_vRateScaling);
		}
		TreeTime_Mcmcmc::Mcmcmc Application(*pData,vpStateManager,pStateSampler);
		TreeTime_Mcmcmc::Mcmcmc::setRandom(pRng);
		cout << "Setup analysis finished."                                         << endl;
		cout <<                                                                       endl;
		cout << "--------------------------------------------------------------------" << endl;
		cout << "MCMC-sampling:"                                                   << endl;
		cout << "--------------------------------------------------------------------" << endl;
		Application.startMcmcmc();
		delete pProposeChange;
		delete pStateSampler;
		delete pData;
		//Finish
		cout << "MCMC-sampling finished."                                          << endl;
		cout <<                                                                       endl;
		cout << "--------------------------------------------------------------------" << endl;
		cout << "TreeTime finished."                                               << endl;
		cout << "--------------------------------------------------------------------" << endl;
		cout << "See output-files for the result of the analysis."                  << endl;
	}

}//TreeTime_Core
