/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "Mcmcmc.hpp"
#include "McmcmcData.hpp"
#include "StateManager.hpp"
#include "Change.hpp"
#include "Sampler.hpp"
#include "Random.hpp"

namespace TreeTime_Mcmcmc
{

	Random*			Mcmcmc::s_pRandom;


	Mcmcmc::Mcmcmc(const McmcmcData& p_McmcmcData,
					vector<StateManager*>& p_vpStateManager,
					Sampler* p_pSampler)
	: m_nSamples			(p_McmcmcData.getNSamples())
	, m_burnIn				(p_McmcmcData.getBurnIn())
	, m_sampleFrequency		(p_McmcmcData.getSampleFrequency())
	, m_nChains				(p_McmcmcData.getNChains())
	, m_heatingCoefficent	(p_McmcmcData.getHeatingCoefficent())
	, m_SwapFrequency		(p_McmcmcData.getSwapFrequency())
	, m_vpStateManager		(p_vpStateManager)
	, m_pSampler			(p_pSampler)
	{
	}

	Mcmcmc::~Mcmcmc()
	{
	}

	void Mcmcmc::startMcmcmc()
	{
		int swapCounter = 1;
		for(int i=0; i<m_burnIn; ++i)
		{
			changeStateModels();
			++swapCounter;
			if(swapCounter > m_SwapFrequency)
			{
				swapCounter = 1;
				swapChains();
			}
		}
		for(int i=0; i<m_nSamples-1; ++i)
		{
			sampleState();
			for(int j=0; j<m_sampleFrequency; ++j)
			{
				changeStateModels();
				++swapCounter;
				if(swapCounter > m_SwapFrequency)
				{
					swapCounter = 1;
					swapChains();
				}
			}
		}
		sampleState();
		m_pSampler->finishOutput();
	}

	void Mcmcmc::changeStateModels()
	{
		for(int chain=0; chain < m_nChains; chain++)
		{
			m_vpStateManager[chain]->proposeNewChange();
			m_vpStateManager[chain]->getChange()->changeState();
			double acceptProb =
					m_vpStateManager[chain]->getChange()->getProposalRatio() *
					pow(m_vpStateManager[chain]->getChange()
							->getStationaryDistributionRatio(),
					1.0/(1.0+(m_heatingCoefficent*static_cast<double>(chain))));
			if(acceptProb > s_pRandom->getUniform())
			{
				m_vpStateManager[chain]->getChange()->accept();
				m_pSampler->monitorChange(chain,m_vpStateManager[chain]->
						getChange()->getId(),true);
			}
			else
			{
				m_vpStateManager[chain]->getChange()->reject();
				m_pSampler->monitorChange(chain,m_vpStateManager[chain]->
						getChange()->getId(),false);
			}
			delete m_vpStateManager[chain]->getChange();
		}
	}

	void Mcmcmc::sampleState()
	{
		m_vpStateManager[0]->sample();
	}

	void Mcmcmc::swapChains()
	{
		if(m_nChains>1)//nothing to be done for only one chain
		{
			int nr1 = s_pRandom->getInteger(m_nChains);
			int nr2 = s_pRandom->getInteger(m_nChains);
			if(nr1 != nr2)//nothing to be done if they are equal
			{
				double acceptProb = pow(
					exp(m_vpStateManager[nr1]->getLogStationaryDistribution()-
					m_vpStateManager[nr2]->getLogStationaryDistribution()),
					(1.0/(1.0+(m_heatingCoefficent*static_cast<double>(nr2)))) -
					(1.0/(1.0+(m_heatingCoefficent*static_cast<double>(nr1)))));
				if(acceptProb > s_pRandom->getUniform())
				{ 
					StateManager* pTempState = m_vpStateManager[nr1];
					m_vpStateManager[nr1]    = m_vpStateManager[nr2];
					m_vpStateManager[nr2]    = pTempState;
					m_pSampler->monitorSwap(nr1,nr2,true);
				}
				else
				{
					m_pSampler->monitorSwap(nr1,nr2,false);
				}
			}
		}
	}

	void Mcmcmc::insertStateManager(StateManager* p_pStateManager, int p_pos)
	{
		m_vpStateManager[p_pos] = p_pStateManager;
	}

	void Mcmcmc::setRandom(Random* p_pRandom)
	{
		s_pRandom = p_pRandom;
	}

}//TreeTime_Mcmcmc
