/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef MCMCMC_HPP_
#define MCMCMC_HPP_

#include "Settings.hpp"

namespace TreeTime_Mcmcmc
{

	class McmcmcData;//forward declaration
	class StateManager;//forward declaration
	class Random;//forward declaration
	class Sampler;//forward declaration

	/**
	 * Defines the Controller Class of the Mcmcmc Algorithmn
	 */
	class Mcmcmc
	{

		public:
			/**
			 * The Constructor initialises the Mcmcmc with the appropriate Data
			 * given by the Class McmcmcData
			 */
			Mcmcmc(const McmcmcData& p_McmcmcData,
					vector<StateManager*>& p_vpStateManager,
					Sampler* p_pSampler);

			/**
			 * Destructor of Mcmcmc
			 */
			~Mcmcmc();

			/**
			 * starts the Mcmcmc Algorithmn 
			 */
			void	startMcmcmc();

			/**
			 * Sets the State Models in m_pStateModel
			 */
			void	insertStateManager(StateManager* p_pStateManager,
					int p_pos);

			/**
			 * Sets the Pointer to the Random Generator
			 */
			static void setRandom(Random* p_pRandom);

		private:

			/**
			 * Changes the current State in the StateModels
			 * see J. Parallel Distrib. Comput. 63 (2003) 707718
			 */
			void	changeStateModels();
			
			/**
			 * Swaps the States in the StateModels (heated Chains)
			 * according to Metropolis Coupled MCMC
			 */
			void	swapChains();
			
			/**
			 * Samples the current State in the StateModels
			 */
			void	sampleState();

			/**
			 * Forbidden Copying, not implemented
			 */
			Mcmcmc(const Mcmcmc&);
			
			/**
			 * Forbidden Assignment, not implemented
			 */
			Mcmcmc& operator=(const Mcmcmc&);

		private:

			/** Number of States to be sampled */
			int						m_nSamples;
			/** Number of Steps for the Burn In */
			int						m_burnIn;
			/** Frequency Number of Steps for Sampling */
			int						m_sampleFrequency;
			/** Number of diffrent Chains */
			int						m_nChains;
			/** The heating parameter for the Chains */
			double					m_heatingCoefficent;
			/** Frequency Number of Steps to swap the MetropolisCoupled Chains*/
			int						m_SwapFrequency;
			/** Vector of Pointers to the State Models*/
			vector<StateManager*>	m_vpStateManager;
			/** Sampler for Output */
			Sampler*				m_pSampler;
			/** Random Generator */
			static Random*			s_pRandom;

	};//Mcmcmc

}//TreeTime_Mcmcmc

#endif /*MCMCMC_HPP_*/
