/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "DnaEvolutionModel.hpp"
#include "Helper/GammaRates.hpp"
#include "DnaModels/JukesCantorModel.hpp"
#include "DnaModels/GtrModel.hpp"

namespace TreeTime_MolecularEvolution
{

	Random* DnaEvolutionModel::s_pRandom;


	DnaEvolutionModel::DnaEvolutionModel(bool p_hasInvSites,
			double p_propOfInvSites, int p_nGammaRates,
			double p_gammaShapeParameter, int p_nIndependentStatDistrParameters,
			int p_nIndependentTransModelParameters)
	: m_hasInvariantSites				(p_hasInvSites)
	, m_proportionOfInvariantSites		(p_propOfInvSites)
	, m_nGammaRates						(p_nGammaRates)
	, m_gammaShapeParameter				(p_gammaShapeParameter)
	, m_nIndependentStatDistrParameters	(p_nIndependentStatDistrParameters)
	, m_nIndependentTransModelParameters(p_nIndependentTransModelParameters)
	, m_stationaryDistribution			(new double[4])
	, m_transitionMatrix				(allocate4x4Matrix())
	, m_rates							(initRates())
	{
	}

	DnaEvolutionModel::DnaEvolutionModel(const DnaEvolutionModel& p_Rhs)
	: m_hasInvariantSites				(p_Rhs.m_hasInvariantSites)
	, m_proportionOfInvariantSites		(p_Rhs.m_proportionOfInvariantSites)
	, m_nGammaRates						(p_Rhs.m_nGammaRates)
	, m_gammaShapeParameter				(p_Rhs.m_gammaShapeParameter)
	, m_nIndependentStatDistrParameters
			(p_Rhs.m_nIndependentStatDistrParameters)
	, m_nIndependentTransModelParameters
			(p_Rhs.m_nIndependentTransModelParameters)
	, m_stationaryDistribution
			(cloneArray(4,p_Rhs.m_stationaryDistribution))
	, m_transitionMatrix				(allocate4x4Matrix())
	, m_rates							
				(cloneArray(p_Rhs.m_nGammaRates, p_Rhs.m_rates))
	{
	}

	DnaEvolutionModel::~DnaEvolutionModel()
	{
		delete[] m_stationaryDistribution;
		for(int i=0; i<4; ++i)
		{
			delete[] m_transitionMatrix[i];
		}
		delete[] m_transitionMatrix;
		delete[] m_rates;
	}

	DnaEvolutionModel* DnaEvolutionModel::newDnaEvolutionModel(
			string p_stringRepresentation)
	{
		string model = p_stringRepresentation.substr(0,
				p_stringRepresentation.find_first_of("+("));
		bool hasInvariantSites = false;
		bool hasGammaRates = false;
		int nGammaRates = 1;
		vector<double> par;
		p_stringRepresentation = p_stringRepresentation.substr(
				p_stringRepresentation.find_first_of("+("));
		while(p_stringRepresentation.at(0)=='+')
		{
			p_stringRepresentation = p_stringRepresentation.substr(1);
			if(p_stringRepresentation.at(0)=='I')
			{
				hasInvariantSites = true;
				p_stringRepresentation = p_stringRepresentation.substr(1);
			}
			else if(p_stringRepresentation.at(0)=='G')
			{
				hasGammaRates = true;
				p_stringRepresentation = p_stringRepresentation.substr(1);
				int nextIndex = p_stringRepresentation.find_first_of("+(");
				nGammaRates = atoi
						((p_stringRepresentation.substr(0,nextIndex)).c_str());
				p_stringRepresentation =
						p_stringRepresentation.substr(nextIndex);
			}
		}
		p_stringRepresentation =
				p_stringRepresentation.substr(1,p_stringRepresentation.size()-2);
		if(p_stringRepresentation.size()>0)
			p_stringRepresentation.append(",");//append comma
		while(p_stringRepresentation!="")
		{
			int nextComma = p_stringRepresentation.find_first_of(",");
			par.push_back(
					atof((p_stringRepresentation.substr(0,nextComma)).c_str()));
			p_stringRepresentation = p_stringRepresentation.substr(nextComma+1);
		}
		double gammaShapeParameter  = 1.0;
		double propOfInvariantSites = 0.0;
		if(hasGammaRates)
		{
			gammaShapeParameter = par.back();
			par.pop_back();
		}
		if(hasInvariantSites)
		{
			propOfInvariantSites = par.back();
			par.pop_back();
		}
		if(model=="JC")
		{
			return new JukesCantorModel(hasInvariantSites, propOfInvariantSites,
			nGammaRates, gammaShapeParameter);
		}
		else if(model=="GTR")
		{
			return new GtrModel(par[0],par[1],par[2],par[3], //stat distribution
					par[4],par[5],par[6],par[7],par[8],par[9],//other parameters
					hasInvariantSites,propOfInvariantSites,nGammaRates,
					gammaShapeParameter);
		}
		else
		{
			return new JukesCantorModel(false,0.0,1,1.0);
		}
	}

	void DnaEvolutionModel::operator<<(std::ostream& p_out) const
	{
		p_out << "EM ";
	}

	std::ostream& operator<<(std::ostream& p_out,
			const DnaEvolutionModel& p_DnaMem)
	{
		p_DnaMem.operator<<(p_out);
		return p_out;
	}

	double* const DnaEvolutionModel::initRates()
	{
		double* const pRates = new double[m_nGammaRates];
		vector<double> Rates =
				GammaRates::getGammaRates(m_nGammaRates,m_gammaShapeParameter);
		for(int i=0; i<m_nGammaRates; ++i)
			pRates[i] = Rates[i];
		return pRates;
	}

	double** const DnaEvolutionModel::allocate4x4Matrix()
	{
		double** const matrix = new double*[4];
		for(int i=0; i<4; ++i)
		{
			matrix[i] = new double[4];
		}
		return matrix;
	}

	double** const DnaEvolutionModel::clone4x4Matrix(
			double** const p_matrix)
	{
	 	double** const matrix = new double*[4];
		for(int i=0; i<4; ++i)
		{
			matrix[i] = new double[4];
			for(int j=0; j<4; ++j)
			{
				matrix[i][j] = p_matrix[i][j];
			}
		}
		return matrix;
	}

	double* const DnaEvolutionModel::cloneArray(int p_length,
			const double* const p_array)
	{
		double* const array = new double[p_length];
		for(int i=0; i<p_length; ++i)
			array[i] = p_array[i];
		return array;
	}

	void DnaEvolutionModel::setTransitionMatrix(double** const p_rateMatrix,
			double p_time)
	{
		double denominator = pow(2.0,s_transitionMatrixApproxSteps);
		for(int i=0; i<4; ++i)
			for(int j=0; j<4; ++j)
				m_transitionMatrix[i][j] = p_rateMatrix[i][j];
		for(int i=0; i<4; ++i)
		{
			for(int j=0; j<4; ++j)
			{
				m_transitionMatrix[i][j] *= (p_time/denominator);
				if(i==j)
					m_transitionMatrix[i][j] += 1;
			}
		}
		for(int i=0; i<s_transitionMatrixApproxSteps; ++i)
			squareMatrix(m_transitionMatrix);
	}

	void DnaEvolutionModel::squareMatrix(double** __restrict const p_matrix)
	{
		const double tempMatrix[4][4] = {
			{p_matrix[0][0], p_matrix[0][1], p_matrix[0][2], p_matrix[0][3]},
			{p_matrix[1][0], p_matrix[1][1], p_matrix[1][2], p_matrix[1][3]},
			{p_matrix[2][0], p_matrix[2][1], p_matrix[2][2], p_matrix[2][3]},
			{p_matrix[3][0], p_matrix[3][1], p_matrix[3][2], p_matrix[3][3]}};
		for(int i=0; i<4; ++i)
		{
				p_matrix[i][0] = (
					tempMatrix[i][0]*tempMatrix[0][0] +
					tempMatrix[i][1]*tempMatrix[1][0] +
					tempMatrix[i][2]*tempMatrix[2][0] +
					tempMatrix[i][3]*tempMatrix[3][0] );
				p_matrix[i][1] = (
					tempMatrix[i][0]*tempMatrix[0][1] +
					tempMatrix[i][1]*tempMatrix[1][1] +
					tempMatrix[i][2]*tempMatrix[2][1] +
					tempMatrix[i][3]*tempMatrix[3][1] );
				p_matrix[i][2] = (
					tempMatrix[i][0]*tempMatrix[0][2] +
					tempMatrix[i][1]*tempMatrix[1][2] +
					tempMatrix[i][2]*tempMatrix[2][2] +
					tempMatrix[i][3]*tempMatrix[3][2] );
				p_matrix[i][3] = (
					tempMatrix[i][0]*tempMatrix[0][3] +
					tempMatrix[i][1]*tempMatrix[1][3] +
					tempMatrix[i][2]*tempMatrix[2][3] +
					tempMatrix[i][3]*tempMatrix[3][3] );
		}
	}

	void DnaEvolutionModel::setGammaShapeParameter(double p_shape)
	{
		m_gammaShapeParameter = p_shape;
		vector<double> rates =
				GammaRates::getGammaRates(m_nGammaRates,m_gammaShapeParameter);
		for(int i=0; i<m_nGammaRates; ++i)
			m_rates[i] = rates[i];
	}

	double DnaEvolutionModel::changeInvariantSites()
	{
		const double delta = 0.1;
		m_proportionOfInvariantSites = fabs( m_proportionOfInvariantSites +
			s_pRandom->getUniform(-delta,delta));
		if(m_proportionOfInvariantSites > 1.0)
			m_proportionOfInvariantSites = 2.0 - m_proportionOfInvariantSites;
		return 1.0;
	}
	
	double DnaEvolutionModel::changeGammaRates()
	{
		const double delta = 1.0;
	const double oldValue = m_gammaShapeParameter;
		m_gammaShapeParameter =  m_gammaShapeParameter +
			s_pRandom->getUniform(-delta,delta);
		if(m_gammaShapeParameter<0.1)
			m_gammaShapeParameter = 0.2 - m_gammaShapeParameter;
		else if(m_gammaShapeParameter>20.0)
			m_gammaShapeParameter = 40.0 - m_gammaShapeParameter;
		vector<double> rates =
				GammaRates::getGammaRates(m_nGammaRates,m_gammaShapeParameter);
		for(int i=0; i<m_nGammaRates; ++i)
			m_rates[i] = rates[i];
		//return 1.0;
	return exp((oldValue - m_gammaShapeParameter)/2.0);
	}


}//TreeTime_MolecularEvolution
