/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef DNAEVOLUTIONMODEL_HPP_
#define DNAEVOLUTIONMODEL_HPP_

#include "Settings.hpp"
#include "Random.hpp"

namespace TreeTime_MolecularEvolution
{

	class DnaEvolutionModel;//Forward Declaration

	/**
	 * PrintStream for Output of DnaEvolutionModedl,
	 * implemented as a namespace level function.
	 * @param ostream&
	 * @param const DnaEvolutionModel&
	 * @return ostream&
	 */
	ostream& operator<<(ostream& , const DnaEvolutionModel&);

	/// Represents the abstract Interface for a Dna Evolution Model
	/**
	 * The Evolution Model consists of a Stationary Distribution
	 * and a Rate Change Model, from which a Substitution(Transition)-Matrix
	 * can be computed, for any positive Distance.
	 * The stat. Distribution and Rate Change Model may be parametrisised.
	 * There may be as well a fixed proportion of invariant Sites,
	 * and a Gamma Shaped Approximation to the Rate Changes.
	 * The Model can be represented by a string in the following way:
	 * for Example: GTR+I+G4(0.35,0.25,0.2,0.2,1,1.2,1.5,2,1.1,1,0.3,1.0)
	 * <ModelIdentifier>[+I<if there are >]
	 * for Example: JK+I+G4(0.35,0.25,0.2,0.2,1,1.2,1.5,2,1.1,1,0.3,1)
	 */
	class DnaEvolutionModel
	{
		/**
		 * Declaration of friend ouput operator
		 */
		friend ostream& operator<<(ostream& , const DnaEvolutionModel&);

		public:

			/// Static Factory Method for generating a DnaEvolutionModel
			/**
			 * This Factory Method expects a String Representation of the Dna
			 * Evolution Model, and returns a Pointer to a new allocated Dna
			 * Evolution Model
			 * @param string
			 * @return DnaEvolutionModel* 
			 */
			static DnaEvolutionModel* newDnaEvolutionModel(string
					p_stringRepresentation);

			/**
			 * Destructor of DnaEvolutionModel
			 */
			virtual ~DnaEvolutionModel();

			/**
			 * Returns a Clone of the Dna Evolution Model
			 * @return DnaEvolutionModel* 
			 */
			virtual DnaEvolutionModel* clone() const = 0;

			/**
			 * Returns the Stationary Distribution of the Model
			 * @return double*
			 */
			inline const double* const getStationaryDistribution() const;

			/**
			 * Returns the Transition Matrix to a given Time Distance
			 * @param double
			 * @return double**
			 */
			virtual double** const getTransitionMatrix(double p_time) = 0;

			/**
			 * Returns if the Model has a Gamma Rate Approximation
			 * @return bool
			 */
			inline bool			hasGammaRates() const;

			/**
			 * Returns the Number of Rates used in the Gamma Rate
			 * Approximation
			 * @return int
			 */
			inline int			getNGammaRates() const;

			/**
			 * Returns an Array with the Rate Values
			 * @return const double*
			 */
			inline const double* const getGammaRates() const;
			
			/**
			 * Return the Shape Parameter of the Gamma Rate Approximation.
			 * @return double
			 */
			inline double		getGammaShapeParameter() const;
			
			/**
			 * Sets the Shape Parameter of the Gamma Rate Approximation.
			 * and refreshes the GammaRates Array
			 * @param double
			 */
			void			setGammaShapeParameter(double p_shape);

			/**
			 * Returns if the Model has Invariant Sites
			 * @return bool
			 */
			inline bool			hasInvariantSites() const;

			/**
			 * Returns the Proportion of Invariant Sites
			 * @return double
			 */
			inline double		getProportionOfInvariantSites() const;

			/**
			 * Sets the Proportion of Invariant Sites
			 * @param double
			 */
			inline void			setProportionOfInvariantSites(double);

			/**
			 * Returns the Number of Independent Parameters in the Model,
			 * excluding inv Sites and Rate Changes.
			 * @return int
			 */
			inline int			getNIndependentStatDistrParameters() const;

			/**
			 * Returns the Number of Independent Parameters in the Model,
			 * excluding inv Sites and Rate Changes.
			 * @return int
			 */
			inline int			getNIndependentTransModelParameters() const;

			/**
			 * Changes Proportion of invariant Sites
			 * @return double
			 */
			double		changeInvariantSites();

			/**
			 * Changes the Gamma Rate Approximation
			 * @return double
			 */
			double		changeGammaRates();

			/**
			 * Changes the Model, specified by ModelChange
			 * @return double
			 */
			virtual double		changeStationaryDistribution() = 0;

			/**
			 * Changes the Model, specified by ModelChange
			 * @return double
			 */
			virtual double		changeTransitionParameters() = 0;
			
			/**
			 * Sets the Random Generator for changing the Model Parametrers
			 */
			static inline void setRandom(Random* p_pRandom);

		protected:

			/**
			 * Constructor of DnaEvolutionModel
			 */
			DnaEvolutionModel(bool p_hasInvSites, double p_propOfInvSites,
					int p_nGammaRates, double p_gammaShapeParameter,
					int p_nIndependentStatDistrParameters,
					int p_nIndependentTransModelParameters);

			/**
			 * Copy Constructor, clones all array data
			 */
			DnaEvolutionModel(const DnaEvolutionModel& p_Rhs);

			/**
			 * Allocates new Memory for a 4x4 Matrix
			 */
			double** const	allocate4x4Matrix();

			/**
			 * Allocatesnew Memory for a 4x4 Matrix, and initialsies with the
			 * Matrix given in the Argument.
			 */
			double** const	clone4x4Matrix(double** const p_matrix);

			/**
			 * Allocate new Memory for a 4x4 Matrix, and initialsies with the
			 * Matrix given in the Argument.
			 */
			double* const	cloneArray(int p_length,
					const double* const p_array);
			
			/**
			 * Calculate and sets the Transition Matrix from the Rate Matrix,
			 * for a given time distance.
			 */
			void setTransitionMatrix(double** const p_rateMatrix,double p_time);

		private:

			/**
			 * Private Copy Constructor, not implemented
			 */
			DnaEvolutionModel& operator=(const DnaEvolutionModel& rhs);

			/**
			 * Output Operator, which is called by the namespace level output
			 * operator, and is deduced to the concrete classes.
			 */
			virtual void operator<<(ostream&) const = 0;

			/**
			 * Initialises new Memory for the discrete Rate Approximation
			 */
			double* const	initRates();
			
			/**
			 * Squares the Matrix given in the Argument
			 */
			void squareMatrix(double** __restrict const p_matrix);

		protected:

			/** Approximation Steps in Calculating the Transition Matrix*/
			static const int	s_transitionMatrixApproxSteps = 5;
			/** Boolean Flag if the Model has Ivariant Sites */
			const bool			m_hasInvariantSites;
			/** Value in [0,1], giving the Proportion of invariant Sites */
			double				m_proportionOfInvariantSites;
			/** Number of discrete Rates in the Gamma Rate Approximation */
			const int			m_nGammaRates;
			/** Parameter to control the Distribution of Rate Approximation */
			double				m_gammaShapeParameter;
			/** Number of independent Parameters controlling the stationary
			 * Distribution */
			const int			m_nIndependentStatDistrParameters;
			/** Number of independent Parameters controlling the Rate Matrix,
			 * not counting the Parameters for the Stationary Distribution */
			const int			m_nIndependentTransModelParameters;
			/** Pointer to the stationary Distribution */
			double* const		m_stationaryDistribution;
			/** Pointer to the actual Transition Matrix */
			double** const		m_transitionMatrix;
			/** Pointer to the Rates */
			double*	const		m_rates;
			/** Pointer to Random Generator and Distribution Density Interface */
			static Random*		s_pRandom;

	};//DnaEvolutionModel

}//TreeTime_MolecularEvolution

#include "DnaEvolutionModel.cxx"

#endif /*DNAEVOLUTIONMODEL_HPP_*/
