/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "GtrModel.hpp"
#include "../Helper/GammaRates.hpp"

namespace TreeTime_MolecularEvolution
{

	GtrModel::GtrModel(double p_piA, double p_piC, double p_piG, double p_piT,
			double p_x1, double p_x2, double p_x3, double p_x4, double p_x5,
			double p_x6, bool p_hasInvSites, double p_propOfInvSites,
			int p_nGammaRates, double p_gammaShapeParameter)
	: DnaEvolutionModel(p_hasInvSites,p_propOfInvSites,p_nGammaRates,
			p_gammaShapeParameter, 3, 5)
	, m_parameters(new double[6])
	, m_rateMatrix(allocate4x4Matrix())
	{
		p_piA = fabs(p_piA); p_piC = fabs(p_piC);
		p_piG = fabs(p_piG); p_piT = fabs(p_piT);
		p_x1=fabs(p_x1); p_x2=fabs(p_x2); p_x3=fabs(p_x3);
		p_x4=fabs(p_x4); p_x5=fabs(p_x5); p_x6=fabs(p_x6);
		m_stationaryDistribution[0] = (p_piA/(p_piA+p_piC+p_piG+p_piT));
		m_stationaryDistribution[1] = (p_piC/(p_piA+p_piC+p_piG+p_piT));
		m_stationaryDistribution[2] = (p_piG/(p_piA+p_piC+p_piG+p_piT));
		m_stationaryDistribution[3] = (p_piT/(p_piA+p_piC+p_piG+p_piT));
		m_parameters[0] = p_x1;
		m_parameters[1] = p_x2;
		m_parameters[2] = p_x3;
		m_parameters[3] = p_x4;
		m_parameters[4] = p_x5;
		m_parameters[5] = p_x6;
		updateParametersTo1Pem();
		updateRateMatrix();
	}

	GtrModel::GtrModel(const GtrModel& p_Rhs)
	: DnaEvolutionModel(p_Rhs)
	, m_parameters(cloneArray(6,p_Rhs.m_parameters))
	, m_rateMatrix(clone4x4Matrix(p_Rhs.m_rateMatrix))
	{
	}

	GtrModel::~GtrModel()
	{		
		for(int i=0; i<4; ++i)
		{
			delete[] m_rateMatrix[i];
		}
		delete[] m_rateMatrix;
		delete[] m_parameters;
	}

	GtrModel* GtrModel::clone() const
	{
		GtrModel* newGtr = new GtrModel(*this);
		return newGtr;
	}

	double** const GtrModel::getTransitionMatrix(double p_time)
	{
		setTransitionMatrix(m_rateMatrix, p_time);
		return m_transitionMatrix;
	}

	double GtrModel::changeStationaryDistribution()
	{
		const double multi = 800.0;// Huelsenbeck et al. took multi(100), 1000 has less Var
		double parameter[4];
		for(int i=0; i<4; i++)
			parameter[i] = multi * m_stationaryDistribution[i];
		double newStatDistribution[4];
		s_pRandom->getDirichlet(parameter, newStatDistribution);
		double oldToNew = s_pRandom->getDirichletDensity(parameter, newStatDistribution);
		for(int i=0; i<4; i++)
			parameter[i] = multi * newStatDistribution[i];
		double newToOld = s_pRandom->getDirichletDensity(parameter,
				m_stationaryDistribution);
		for(int i=0; i<4; ++i)
			m_stationaryDistribution[i] = newStatDistribution[i];
		updateParametersTo1Pem();
		updateRateMatrix();
		return (newToOld/oldToNew);
	}

	double GtrModel::changeTransitionParameters()
	{
		const double delta = 0.2;
		for(int i=0; i<5; ++i)
		{
			m_parameters[i] = fabs((m_parameters[i]/m_parameters[5]) +
					s_pRandom->getUniform(-delta,delta));
		}
		m_parameters[5] = 1.0;
		updateParametersTo1Pem();
		updateRateMatrix();
		return 1.0;
	}

	void GtrModel::updateParametersTo1Pem()
	{
		double normPem = 200.0*(
				(m_stationaryDistribution[0] *
				(m_parameters[0] + m_parameters[1] + m_parameters[2])) +
				(m_stationaryDistribution[1] *
				(m_parameters[3] + m_parameters[4])) +
				(m_stationaryDistribution[2] *
				(m_parameters[5])) );
		for(int i=0 ; i<6; i++)
			m_parameters[i] /= normPem;
	}

	void GtrModel::updateRateMatrix()
	{
		m_rateMatrix[0][0]= -(m_parameters[0]+m_parameters[1]+m_parameters[2]);
		m_rateMatrix[0][1]=   m_parameters[0];
		m_rateMatrix[0][2]=   m_parameters[1];
		m_rateMatrix[0][3]=   m_parameters[2];
		m_rateMatrix[1][0]=   m_parameters[0]*m_stationaryDistribution[0]/
				m_stationaryDistribution[1];
		m_rateMatrix[1][1]= -(m_parameters[0]*m_stationaryDistribution[0]/
				m_stationaryDistribution[1]+m_parameters[3]+m_parameters[4]);
		m_rateMatrix[1][2]=   m_parameters[3];
		m_rateMatrix[1][3]=   m_parameters[4];
		m_rateMatrix[2][0]=   m_parameters[1]*m_stationaryDistribution[0]/
				m_stationaryDistribution[2];
		m_rateMatrix[2][1]=   m_parameters[3]*m_stationaryDistribution[1]/
				m_stationaryDistribution[2];
		m_rateMatrix[2][2]= -(m_parameters[1]*m_stationaryDistribution[0]/
				m_stationaryDistribution[2] + 
				m_parameters[3]*m_stationaryDistribution[1]/
				m_stationaryDistribution[2] + 
				m_parameters[5]);
		m_rateMatrix[2][3]=   m_parameters[5];
		m_rateMatrix[3][0]=   m_parameters[2]*m_stationaryDistribution[0]/
				m_stationaryDistribution[3];
		m_rateMatrix[3][1]=   m_parameters[4]*m_stationaryDistribution[1]/
				m_stationaryDistribution[3];
		m_rateMatrix[3][2]=   m_parameters[5]*m_stationaryDistribution[2]/
				m_stationaryDistribution[3];
		m_rateMatrix[3][3]= -(m_parameters[2]*m_stationaryDistribution[0]/
				m_stationaryDistribution[3] + 
				m_parameters[4]*m_stationaryDistribution[1]/
				m_stationaryDistribution[3] + 
				m_parameters[5]*m_stationaryDistribution[2]/
				m_stationaryDistribution[3]);
	}

	void GtrModel::operator<<(std::ostream& p_out) const
	{
		p_out << "GTR\t";
		p_out << m_nGammaRates					<< "\t";
		p_out << m_gammaShapeParameter			<< "\t";
		p_out << m_hasInvariantSites			<< "\t";
		p_out << m_proportionOfInvariantSites	<< "\t";
		for(unsigned int i=0; i<4; ++i)
			p_out << m_stationaryDistribution[i]<< "\t";
		for(unsigned int i=0; i<5; ++i)
			p_out << m_parameters[i]<< "\t";
		p_out << m_parameters[5];
	}

}//TreeTime_MolecularEvolution
