/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef BINARYTREE_HPP_
#define BINARYTREE_HPP_

#include "Settings.hpp"
#include "iterator.hpp"
#include "BinaryNode.hpp"
#include "Tree.hpp"

namespace TreeTime_MolecularTree
{

	/**
	* @brief BinaryTree is a Tree Container with Elements NODEDATA;
	* NODEDATA can be accessed via the Iterator iterator;
	* MolecularData are automatically managed by the MemoyAllocator.
	* 
	* The Elements of BinaryTree are are organized in a Tree Structure
	* with Nodes defined in BinaryNode.
	* BinaryNode holds a Pointer to NodeData and MolecularData.
	* NodeData is generic, and must be given as the Template NODEDATA.
	* MolecularData is administrated automatically by the MemoryController.
	* MolecularData for the Leafs were stored in the MemoryAllocator,
	* and point for different BinaryTrees to the same Location.
	* The Pointer to MolecularData at Inner Nodes
	* can share the same Location in Memory.
	*/
	template<class NODEDATA>
	class BinaryTree: public Tree<NODEDATA>
	{

		public:
			/**
			 * The Constructor of BinaryTree is build from the Newick String 
			 * given in the the first Argument,
			 * which contains topological and branchlength Indformation.
			 * The Names in the Newick String corresponding to a Nodes,
			 * were mapped to a unique id, given by the the Map in the 
			 * second Argument, and assigned to the Nodes attribute.
			 * @param Newick String
			 * @param Map LeafIdentifier String to unique Id
			 * @param Reference to MemoryAllocator 
			 */
			BinaryTree(const string& p_newickString,
					map<string,int>& p_mSpeciesToId);
			/**
			 * The Copy Constructor of BinaryTree clones the whole
			 * TreeStructure. The Data at the Nodes were copyed 
			 * by the Standard CopyConstructor of NodeData.
			 */
			BinaryTree(const BinaryTree& p_Tree);
			/**
			 * The Destructor for BinaryTree recursivly destructs all
			 * NodeElements.
			 */
			virtual ~BinaryTree();
			/**
			 * Returns an iterator to the first Node (Root) in the BinaryTree.
			 * @return Iterator to the Root Node
			 */
			virtual iterator<NODEDATA> getRoot() const;
			/**
			 * Sets the Root Node of the Tree to the Node in the Argument
			 * Iterator.
			 * @param Iterator to the new Root Node
			 */
			virtual void setRoot(iterator<NODEDATA> p_rootIterator);
			/**
			 * Clones the whole TreeStructure,
			 * The Data at the Nodes were copyed by the Standard
			 * Copy Constructor of NodeData.
			 */
			virtual Tree<NODEDATA>* copy();
			
		private:
			/**
			 * The Assignment Operator is ambivalent,
			 * so declared private and not implemented. 
			 */
			BinaryTree& operator=(const BinaryTree& p_Tree);
			/**
			 * Helper Function for recursive copying and cloning the Tree
			 * Structure 
			 */
			void copy(BinaryNode<NODEDATA>* p_pMaster, BinaryNode<NODEDATA>* p_pReplica);
			/**
			 * Helper Function for recursive deleting the Tree Structure 
			 */
			void remove(BinaryNode<NODEDATA>* p_pNode);

		private:
			iterator<NODEDATA>	m_iRoot;/**<Iterator to the Root*/

	};//BinaryTree

}//TreeTime_MolecularTree

#include "BinaryTree.cxx"

#endif /*BINARYTREE_HPP_*/
