/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef ITERATOR_HPP_
#define ITERATOR_HPP_

#include "BinaryNode.hpp"
//#include "BinaryTree.hpp"
#include "NodeData.hpp"

namespace TreeTime_MolecularTree
{
	// Foreward declaration
	template<class NODEDATA>
	class BinaryTree;

//	template<class NODEDATA>
//	class BinaryNode;
	
	/// Iterator
	/**
	 * Iterator which walks through the Nodes of the Tree. 
	 * The dereferenced Iterator points to the NodeData.
	 */
	template<class NODEDATA>
	class iterator
	{

		friend class BinaryTree<NODEDATA>;

		public:
			/**
			 * The Copy Constructs copys the Pointer to a Node
			 * @param Pointer to Node
			 */
			iterator(const iterator& p_iterator);
			/**
			 * The Assignment Operator copys the Pointer to Node from the
			 * Argument
			 * @param iterator to copy
			 */
			iterator& operator=(const iterator& p_iterator);
			/**
			 * The Destructor does nothing else.
			 */
			~iterator();
			/**
			 * The operator returns the dereferenced Pointer to Node.
			 * @return Reference to NodeData
			 */
			inline NODEDATA& operator*();
			/**
			 * The operator returns the Pointer to Node.
			 * @return Pointer to NodeData
			 */
			inline NODEDATA* operator->();
			/**
			 * False, if the Pointer to Node points to 0, else True
			 * @return bool
			 */
			inline operator bool();
			/**
			 * True, if the argument iterator points to the same location
			 * @return bool
			 */
			inline bool operator ==(const iterator& p_iterator) const;
			/**
			 * Compares the Pointer Locations
			 * @return bool
			 */
			inline bool operator <(const iterator& p_iterator) const;
			/**
			 * Returns an Iterator to the Father Node
			 * @return iterator
			 */
			inline iterator<NODEDATA> getFather();
			/**
			 * Returns an Iterator to the First Child Node
			 * @return iterator
			 */
			inline iterator<NODEDATA> getFirstChild();
			/**
			 * Returns an Iterator to the Next Child Node
			 * @return iterator
			 */
			inline iterator<NODEDATA> getNextChild();
			/**
			 * Returns an Iterator to the Last Child Node
			 * @return iterator
			 */
			inline iterator<NODEDATA> getLastChild();
			/**
			 * Sets the Father of the Iterator, to the Iterator given in the
			 * Argument
			 * @param iterator
			 */
			inline void setFather(const iterator<NODEDATA>& p_iterator);
			/**
			 * Sets the First Child of the Iterator, to the Iterator given in
			 * the Argument
			 * @param iterator&
			 */
			inline void setFirstChild(const iterator<NODEDATA>& p_iterator);
			/**
			 * Sets the Last Child of the Iterator, to the Iterator given in
			 * the Argument
			 * @param iterator&
			 */
			inline void setLastChild(const iterator<NODEDATA>& p_iterator);
			/**
			 * Appends the Iterator given in the Argument, as a Child Node
			 * if the Pointer to the Left Child is 0, the Iterator is inserted
			 * as the Left Child
			 * else if the Pointer to the Right Child is 0, the Iterator is
			 * inserted as the Right Child
			 * else nothing happens. 
			 * @param iterator&
			 */
			inline void appendChild(const iterator<NODEDATA>& p_iterator);
			/**
			 * Removes the Child Node given in the Argument from the Child Nodes
			 * @param iterator&
			 */
			inline void removeChild(const iterator<NODEDATA>& p_iterator); 
			/**
			 * False, if the Pointer to Father points to 0, else True
			 * @return bool
			 */
			inline bool hasFather();
			/**
			 * False, if the Pointer to Right and Left Child points to 0,
			 * else True
			 * @return bool
			 */
			inline bool hasChildNodes();
			/**
			 * Creates a new allocated Node, with the Standard Constructor of
			 * Node
			 * @return iterator which points to the new allocated Node
			 */
			inline iterator<NODEDATA> createNewNode();
			/**
			 * Creates a new allocated Node, with the Copy Constructor of Node,
			 * which takes the Node where p_iterator points to as the prototype
			 * @return iterator which points to the new allocated Node
			 */
			inline iterator<NODEDATA> createNewNode(iterator<NODEDATA> p_iterator);
			/**
			 * Deletes a new allocated Node and frees the Memory via the Node
			 * Destructor
			 */
			inline void  deleteNode();

		private:
			/**
			 * The Constructor expects a Pointer to a Node
			 * @param Pointer to Node
			 */
			iterator(BinaryNode<NODEDATA>* p_pNode = 0);

		private:
			BinaryNode<NODEDATA>* m_pNode;

	};//iterator

}//TreeTime_MolecularTree

#include "iterator.cxx"

#endif /*ITERATOR_HPP_*/
