/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "Data.hpp"
#include "SymDistMatrix.hpp"
#include "Upgma.hpp"
#include "../TreeTime_Core/LikelihoodCalculations.hpp"

namespace TreeTime_NexusProcessing
{

	Data::Data()
	: m_withRateChanges			(false)
	, m_withMemChanges			(true)
	, m_vInitialRateChangeModels(vector<string>(5,""))
	, m_InitialGlobalRateChangeModel("")
	, m_outgroup				(-1)
	, m_heatingCoefficient		(0.3)
	, m_swapFrequency			(1)
	, m_weightTopology			(0.2)
	, m_weightNodes				(0.2)
	, m_weightEvolution			(0.28)
	, m_weightRatechange		(0.3)
	, m_weightTimeconversion	(0.02)
	, m_cNodeTime				(1.0)
	, m_cTimeSynchronisation	(1.0)
	, m_cRootTime				(1.0)
	{
		m_randomSeed = time(NULL);
	}

	Data::~Data()
	{
	}

	int Data::getNSamples() const
	{
		return m_nSamples;
	}

	int Data::getBurnIn() const
	{
		return m_burnIn;
	}

	int Data::getSampleFrequency() const
	{
		return m_sampleFrequency;
	}

	int Data::getNChains() const
	{
		return m_nChains;
	}

	double Data::getHeatingCoefficent() const
	{
		return m_heatingCoefficient;
	}

	int Data::getSwapFrequency() const
	{
		return m_swapFrequency;
	}

	void Data::insertSpeciesData(string p_name ,string p_characters)
	{
		if(m_mSpeciesToId.find(p_name) != m_mSpeciesToId.end())//gefunden
		{
			int currentId			= m_mSpeciesToId[p_name];
			m_mIdToDna[currentId]	= (m_mIdToDna[currentId]+p_characters);
		}
		else//nicht gefunden
		{
			int newID				= m_mSpeciesToId.size();
			m_mSpeciesToId[p_name]	= newID;
			m_mIdToSpecies[newID]	= p_name;
			m_mIdToDna[newID]		= p_characters;
		}
	}

//	void Data::insertConstraint(set<int> p_ids1, set<int> p_ids2,
//					double p_mean, double p_sd)
//	{
//		m_vpConstraints.push_back(new TreeTime_Core::Constraint(p_ids1,p_ids2,p_mean,p_sd));
//	}

	void Data::insertTopologicalConstraint(set<int> p_ids1, set<int> p_ids2)
	{
		TreeTime_Constraint::TopologicalConstraint* pC = new 
				TreeTime_Constraint::TopologicalConstraint(p_ids1,p_ids2);
		m_vpConstraints.push_back(pC);
		m_vpTopConstraints.push_back(pC);
	}
	void Data::insertGaussianConstraint(set<int> p_ids1, set<int> p_ids2,
			double p_mean, double p_sd)
	{
		TreeTime_Constraint::GaussianConstraint* pC = new
				TreeTime_Constraint::GaussianConstraint(p_ids1,p_ids2,p_mean,p_sd);
		m_vpConstraints.push_back(pC);
		m_vpGauConstraints.push_back(pC);
	}
	void Data::insertGammaConstraint(set<int> p_ids1, set<int> p_ids2,
			double p_min, double p_mean, double p_alpha, double p_threshold)
	{
		TreeTime_Constraint::GammaConstraint* pC = new
				TreeTime_Constraint::GammaConstraint(p_ids1,p_ids2,p_min,p_mean,p_alpha,p_threshold);
		m_vpConstraints.push_back(pC);
		m_vpGamConstraints.push_back(pC);
	}

	void Data::insertDnaEvolutionModel(string p_memRepresentation)
	{
		m_vpInitialDnaModels.push_back(TreeTime_MolecularEvolution::DnaEvolutionModel::newDnaEvolutionModel(p_memRepresentation));
	}

	void Data::printMaps()
	{
	}

	void Data::printData()
	{
	}

	void Data::stringTo(int id)
	{
		string a = m_mIdToDna[id];
		std::cout<<a<<std::endl; 
	}
	
	void Data::preprocessDnaData()
	{
		map<int,string> DnaCompressed;
		for(int i=0;i<m_nTaxa;++i)
		{
			DnaCompressed[i]="";
		}
		vector<vector<char> > Dna_Site_Species(m_nChar, vector<char>(m_nTaxa,'x'));
		for(int site=0;site<m_nChar;++site)
			for(int species=0;species<m_nTaxa;++species)
				Dna_Site_Species[site][species]=m_mIdToDna[species].at(site);
		vector<int> partition(m_vBeginningFragments);
		partition.push_back(m_nChar);
		for(unsigned int loci=0;loci<partition.size()-1;++loci)
		{
			map<vector<char>,int> DnaMap_site_mult;
			map<vector<char>,int>::iterator IF;
			for(int i=partition[loci];i<partition[loci+1];++i)
			{
				vector<char>& neu = Dna_Site_Species[i];
				IF = DnaMap_site_mult.find(neu);
				if(IF != DnaMap_site_mult.end())//gefunden
					++(IF->second);
				else//neu anlegen
					DnaMap_site_mult[neu] = 1;
			}
			vector<int> multi_loci;
			for(IF = DnaMap_site_mult.begin(); IF != DnaMap_site_mult.end(); IF++)
			{
				for(int site_=0;site_<m_nTaxa;++site_)
				{
					DnaCompressed[site_] += (IF->first)[site_];
				}
				multi_loci.push_back(IF->second);
			}
			m_vvDnaMultipliers.push_back(multi_loci);
		}
		// m_Id2DnaData (map<int,vector4d>) aus DnaCompressed (map<int,string>) aufbauen:
		// Zuhilfe kommt m_DnaMultipliers (vector<vector<int> >) dim1 = #loci, dim2 = #sites im loci, daten sind multiplizitaeten
		m_mIdToCompressedDna.clear();
		int temp_site = 0;
		m_vBeginningAndEndLoci.clear();
		m_vBeginningAndEndLoci.push_back(temp_site);
		for(unsigned int i=0; i<m_vvDnaMultipliers.size(); ++i)
		{
			temp_site += m_vvDnaMultipliers[i].size();
			m_vBeginningAndEndLoci.push_back(temp_site);
		}
		for(map<int,string>::iterator I = DnaCompressed.begin(); I!=DnaCompressed.end(); ++I)
		{
			vector4d oneSpeciesDna;
			for(unsigned int loci=0; loci<m_vBeginningAndEndLoci.size()-1; ++loci)
			{//[loci][site][nuc][rate]
				int sites = m_vBeginningAndEndLoci[loci+1]-m_vBeginningAndEndLoci[loci];
				int nucs  = 4;
				int rates = 1;//da leaf!
				oneSpeciesDna.push_back(vector<vector<vector<double> > >(sites,vector<vector<double> >(nucs,vector<double>(rates,0.0))));
			}
			// Jetzt Id suchen und Dna initialisieren:
			string data = I->second;
			for(unsigned int loci=0; loci<m_vBeginningAndEndLoci.size()-1; ++loci)
			{
				for(int site = m_vBeginningAndEndLoci[loci];site<m_vBeginningAndEndLoci[loci+1];++site)
				{//A,C,G,T
					if(data.at(site)=='A' || data.at(site)=='a')
					{
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][0][0] = 1.0;
					}
					else if(data.at(site)=='C' || data.at(site)=='c')
					{
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][1][0] = 1.0;
					}
					else if(data.at(site)=='G' || data.at(site)=='g')
					{
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][2][0] = 1.0;
					}
					else if(data.at(site)=='T' || data.at(site)=='t')
					{
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][3][0] = 1.0;
					}
					else if(data.at(site)==m_missing || data.at(site)==m_gap)
					{
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][0][0] = 1.0;
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][1][0] = 1.0;
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][2][0] = 1.0;
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][3][0] = 1.0;
					}
					else if(data.at(site)==m_purin)
					{
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][0][0] = 1.0;
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][2][0] = 1.0;
					}
					else if(data.at(site)==m_pyrimidin)
					{
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][1][0] = 1.0;
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][3][0] = 1.0;
					}
					else
					{
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][0][0] = 1.0;
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][1][0] = 1.0;
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][2][0] = 1.0;
						oneSpeciesDna[loci][site-m_vBeginningAndEndLoci[loci]][3][0] = 1.0;
					}
				}
			}
			m_mIdToCompressedDna.insert(m_mIdToCompressedDna.end(),map<int,vector4d>::value_type(I->first,oneSpeciesDna));
		}
		//m_RateDimensions initialisieren:
		m_vRateDimensions.clear();
		for(vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>::iterator I = m_vpInitialDnaModels.begin(); I != m_vpInitialDnaModels.end(); ++I)
		{
			int rates = (*I)->getNGammaRates();
			//todo check this...
//			if( (*I)->getProportionOfInvariantSites() > 0 )
//				rates++;
			m_vRateDimensions.push_back(rates);
		}
		// initialise m_vvInvariantNucleotides
		for(unsigned int loci = 0; loci < m_vvDnaMultipliers.size(); ++loci)
		{
			m_vvInvariantNucleotides.push_back(vector<int>(4,0));
			for(unsigned int nuc = 0; nuc < 4; ++nuc)
			{
				for(unsigned int s = 0; s < (m_vvDnaMultipliers[loci]).size(); ++s)
				{
					bool isInvariant = true;
					for(map<int,vector4d>::iterator I =	m_mIdToCompressedDna.begin();
							I != m_mIdToCompressedDna.end(); ++I)
					{
							if((I->second)[loci][s][nuc][0] < 0.0000001)
							{
								isInvariant = false;
								break;
							}
					}
					if(isInvariant)
						m_vvInvariantNucleotides[loci][nuc] += m_vvDnaMultipliers[loci][s];
				}
			}
		}
		// initialise m_vvvInvariantNucleotides
		m_vvvInvariantNucleotides.clear();
		for(unsigned int loci = 0; loci < m_vvDnaMultipliers.size(); ++loci)
		{
			m_vvvInvariantNucleotides.push_back(vector<vector<int> >
					(m_vvDnaMultipliers[loci].size(),vector<int>(4,0)));
			for(unsigned int nuc = 0; nuc < 4; ++nuc)
			{
				for(unsigned int s = 0; s < (m_vvDnaMultipliers[loci]).size(); ++s)
				{
					bool isInvariant = true;
					for(map<int,vector4d>::iterator I =	m_mIdToCompressedDna.begin();
							I != m_mIdToCompressedDna.end(); ++I)
					{
							if((I->second)[loci][s][nuc][0] < 0.0000001)
							{
								isInvariant = false;
								break;
							}
					}
					if(isInvariant)
						m_vvvInvariantNucleotides[loci][s][nuc] = 1;
				}
			}
		}
	}

	void Data::checkInitialTree()
	{
		if(m_initialTree == "" || m_initialTree == ";")//baue eigenen Tree auf
		{
			vector<int>		vIds;
			vIds.reserve(m_mIdToDna.size());
			vector<string>	vSpeciesname;
			vSpeciesname.reserve(m_mIdToDna.size());
			map<int,string>::iterator I = m_mIdToDna.begin();
			for(unsigned int i = 0; i<m_mIdToDna.size(); ++i)
			{
				vIds.push_back(I->first);
				vSpeciesname.push_back(m_mIdToSpecies[I->first]);
				++I;
			}
			SymDistMatrix D(m_mIdToDna.size());
			double removeZeroLength = 1.0E-7;
			for(unsigned int i=0;i<m_mIdToDna.size();++i)
				for(unsigned int j=i+1;j<m_mIdToDna.size();++j)
				{
					D(i,j) = calcJukeCantorDistance(m_mIdToDna[vIds[i]],
							m_mIdToDna[vIds[j]]);
					if(D(i,j) < 1.0E-7)
					{
						D(i,j) = removeZeroLength;
						removeZeroLength += 1.0E-7;
					}
				}
			Upgma U(&D);
			m_initialTree = U.calcNewickString(&vSpeciesname);
		}
	}
	
	double Data::calcJukeCantorDistance(const string& p_seq1, const string& p_seq2)
	{
		// x -> -(3/4)*ln(1-(4*x/3)), x:= fraction diffrent positions / all positions
		int diff = 0;
		for(unsigned int i=0; i<p_seq1.size(); ++i)
		{
			if(p_seq1[i]!=p_seq2[i])
				diff++;
		}
		if(4*diff > static_cast<int>(3*p_seq1.size()))
		{
			return 1.0E2;
		}
		return -3.0/4.0*log(1.0-((4.0/3.0)*(static_cast<double>(diff)/static_cast<double>(p_seq1.size()))));
	}


	bool Data::hasTimeConstraints() const
	{
		if( (m_vpGauConstraints.size() + m_vpGamConstraints.size()) > 0 )
				return true;
		return false;
	}

	double Data::estimatePEMTime()//is approxmately given by estimating 1 PAM Time 
	{
		string* out = &m_mIdToDna[m_outgroup];
		vector<string*> x;
		x.reserve(m_nTaxa-1);
		for(vector<int>::const_iterator I = m_vActiveSpeciesId.begin(); I!=m_vActiveSpeciesId.end(); ++I)
		{
			if(m_outgroup == *I)
				continue;
			x.push_back(&m_mIdToDna[*I]);
		}
		int diff = 0;
		for(int pos=0;pos<m_nChar;++pos)
		{
			char cOut = (*out)[pos];
			for(vector<string*>::iterator I = x.begin(); I!=x.end(); ++I)
			{
				if(cOut != (**I)[pos])
					diff++;
			}
		}
		return 50.0*static_cast<double>(diff)/static_cast<double>(m_nChar*(m_nTaxa-1));
	}

	double Data::initScaling(TreeTime_MolecularTree::Tree<TreeTime_Core::NodeData>* p_pTree)
	{
		if(hasTimeConstraints())
		{
			double weightedSum = 0.0;
			if(!m_vpGauConstraints.empty())
			{
				list<vector<double> > con;
				for(vector<TreeTime_Constraint::GaussianConstraint*>::iterator I = getGaussianConstraints().begin();
						I != getGaussianConstraints().end(); ++I)
				{
					double mean      = (*I)->getMean();	
					vector<double> vec(3,mean);
					vec[1] = pow((*I)->getDeviation(),2);
					set<int> Ids  = (*I)->getSpeciesIds1();
					Ids.insert((*I)->getSpeciesIds2().begin(),(*I)->getSpeciesIds2().end());
					TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> treeIterator =
							TreeTime_Core::LikelihoodCalculations::getMrca(p_pTree, Ids);
					vec[2] = treeIterator->m_timepoint;
					con.push_back(vec);
				}
				weightedSum = 0.0;
				double sumVariance = 0.0;
				for(list<vector<double> >::iterator I = con.begin(); I!=con.end(); ++I)
				{
					sumVariance += (1.0/(*I)[1]);
					weightedSum += (*I)[0]/((*I)[1]*(*I)[2]);
				}
				weightedSum /= sumVariance;
			}
			double max = 0.0;
			if(!m_vpGamConstraints.empty())
			{
				for(vector<TreeTime_Constraint::GammaConstraint*>::iterator I = getGammaConstraints().begin();
						I != getGammaConstraints().end(); ++I)
				{
					set<int> Ids  = (*I)->getSpeciesIds1();
					Ids.insert((*I)->getSpeciesIds2().begin(),(*I)->getSpeciesIds2().end());
					TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> treeIterator =
							TreeTime_Core::LikelihoodCalculations::getMrca(p_pTree, Ids);
					double time = (*I)->getMinimum()/treeIterator->m_timepoint;
					max = std::max(max,time);
				}
			}
			return(std::max(weightedSum,max*1.02));
		}
		else
		{
			return (1.0/p_pTree->getRoot()->m_timepoint);
		}
	}

}//TreeTime_NexusProcessing
