/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef DATA_HPP_
#define DATA_HPP_

#include "Settings.hpp"
#include "../TreeTime_Mcmcmc/McmcmcData.hpp"
#include "../TreeTime_MolecularEvolution/DnaEvolutionModel.hpp"
#include "../TreeTime_Constraint/TopologicalConstraint.hpp"
#include "../TreeTime_Constraint/GaussianConstraint.hpp"
#include "../TreeTime_Constraint/GammaConstraint.hpp"
#include "../TreeTime_Core/NodeData.hpp"
#include "../TreeTime_MolecularTree/Tree.hpp"
#include "../TreeTime_MolecularTree/iterator.hpp"
#include "../TreeTime_MolecularTree/NodeData.hpp"

namespace TreeTime_NexusProcessing
{

	class Data : public TreeTime_Mcmcmc::McmcmcData
	{

		public:

			// Constructor,Destructur
			Data();
			virtual ~Data();
			// setting
			inline int					getSeed() 				const;
			inline void					setSeed(int);
			inline bool					getRateChanges() 		const;
			inline void					setRateChanges(bool);
			inline bool					getMemChanges() 		const;
			inline void					setMemChanges(bool);
			inline string				getOutputPath() 		const;
			inline void					setOutputPath(string);
			// MCMCMC
			virtual int					getNSamples()			const;
			virtual int					getBurnIn()				const;
			virtual int					getSampleFrequency()	const;
			virtual int					getNChains() 			const;
			virtual int					getSwapFrequency()		const;
			inline void					setBurnIn(int);			
			inline void					setSampleFrequency(int);
			inline void					setNSamples(int);			
			inline void					setNChains(int);
			// Parameter
			virtual double				getHeatingCoefficent()	const;
			inline void					setHeatingCoefficent(double);
			inline void					setNodeTime(double p_);
			inline void					setTimeSynchronisation(double p_);
			inline void					setRootTime(double p_);
			inline double				getNodeTime();
			inline double				getTimeSynchronisation();
			inline double				getRootTime();
			// Weighting
			inline double				getWeightTopology() const;
			inline double				getWeightNodes() const;
			inline double				getWeightEvolution() const;
			inline double				getWeightRatechange() const;
			inline double				getWeightTimeconversion() const;
			inline void					setWeightTopology(double);
			inline void					setWeightNodes(double);
			inline void					setWeightEvolution(double);
			inline void					setWeightRatechange(double);
			inline void					setWeightTimeconversion(double);
			// Dna
			inline int					getNChar()				const;
			inline void					setNChar(int);
			inline int					getNTaxa()				const;
			inline void					setNTaxa(int);
			inline map<string,int>&		getSpecies2Id();
			inline map<int,string>&		getId2Species();
			inline void					setActiveSpecies(vector<int>);
			inline map<int,string>&		getDna();
			inline int					getOutgroup()			const;
			inline void					setOutgroup(int);
			inline void					setMissing(char);
			inline char					getMissing();
			inline void					setGap(char);
			inline char					getGap();
			inline void					setPurin(char);
			inline char					getPurin();
			inline void					setPyrimidin(char);
			inline char					getPyrimidin();
			inline void					setMatching(char);
			inline string&				getInitialTree();
			inline void					setInitialTree(string);
			inline vector<int>&			getBeginningFragments();
			
			inline vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>&	getInitialMolecularModels();
			inline vector<string>&		getInitialRateChangeModels();
			inline string&				getInitialGlobalRateChangeModel();
			
			inline map<int,vector4d>&	getId2DnaData();//[loci][site][nuc][rate]
			inline vector<vector<int> >& getDnaMultipliers();
			inline vector<int>&			getBeginningAndEndLoci();
			inline vector<int>&			getRateDimensions();
			inline vector<vector<vector<int> > >& getInvariantNucleotides();
			// Constraints
			inline vector<TreeTime_Constraint::TopologicalConstraint*>&	getConstraints();
			inline vector<TreeTime_Constraint::TopologicalConstraint*>&	getTopologicalConstraints();
			inline vector<TreeTime_Constraint::GaussianConstraint*>&	getGaussianConstraints();
			inline vector<TreeTime_Constraint::GammaConstraint*>&		getGammaConstraints();

		public:

			void	insertTopologicalConstraint(set<int> p_ids1, set<int> p_ids2);
			void	insertGaussianConstraint(set<int> p_ids1, set<int> p_ids2,
					double p_mean, double p_sd);
			void	insertGammaConstraint(set<int> p_ids1, set<int> p_ids2,
					double p_min, double p_mean, double p_alpha, double p_threshold);
			void	insertDnaEvolutionModel(string p_memRepresentation);

			void	insertSpeciesData(string,string);
			void	printMaps();
			void	printData();
			void	stringTo(int id);
			void	preprocessDnaData();
			void	checkInitialTree();
			double	calcJukeCantorDistance(const string&, const string&);
			bool	hasTimeConstraints() const;
			double	estimatePEMTime();
			double	initScaling(TreeTime_MolecularTree::Tree<TreeTime_Core::NodeData>* p_pTree);

		private:
			// General Settings
			/** Initial Seed for the Random Generator */
			int							m_randomSeed;
			/** Sample Rate Changes */
			bool						m_withRateChanges;
			/** Sample Molecular Evolution Models */
			bool						m_withMemChanges;
			/** Output Path */
			string						m_outputPath;
			// Raw Dna Data
			/** Number of all Dna Sites */
			int							m_nChar;
			/** Number of Taxa with Outgroup */
			int							m_nTaxa;
			/** Map from Species String to unique Id Identifier */
			map<string,int>				m_mSpeciesToId;
			/** Map from unique Id Identifier to Species String */
			map<int,string>				m_mIdToSpecies;
			/** Vector of active Species Ids */
			vector<int>					m_vActiveSpeciesId;
			/** Map from unique Id Identifier to Dna String */
			map<int,string>				m_mIdToDna;
			/** Character Symbol for a Missing Dna */
			char						m_missing;
			/** Character Symbol for a Gap */
			char						m_gap;
			/** Character Symbol for a Purin (A,G) */
			char						m_purin;
			/** Character Symbol for a Pyrimidin (C,T) */
			char						m_pyrimidin;
			/** Character Symbol for a Missingatching Character */
			char						m_matching;
			/** Initial Tree String in newick format */
			string						m_initialTree;
			/** Vector of Beginning Fragments */
			vector<int>					m_vBeginningFragments;
			/** Vector of initial Molecular Evolution Models */
			vector<TreeTime_MolecularEvolution::DnaEvolutionModel*>	m_vpInitialDnaModels;
			/** Vector of initial Rate Change Models */
			vector<string> 				m_vInitialRateChangeModels;
			/** Vector of initial Rate Change Models */
			string		 				m_InitialGlobalRateChangeModel;
			/** Unique Id for the Outgroup Species*/
			int							m_outgroup;
			// Preprocessed Dna Data
			/** Map from Species Id to 4 dim Vector of Compressed Dna Data,
			 * where each nucleotide Pattern for one Site and Loci appears once.
			 * The 4 dimensions are [loci][site][nuc][rate] */
			map<int,vector4d>			m_mIdToCompressedDna;
			/** 2 dim Vector where each entry stands for the Multiplier of 
			 * the corresponding nucleotide Pattern for one Site and Loci  in
			 * m_mIdToCompressedDna, the 2 dim are [loci][site] */
			vector<vector<int> >		m_vvDnaMultipliers;
			/** Vector of Beginning Fragments and Last Loci */
			vector<int>					m_vBeginningAndEndLoci;
			/** Vector of Rate Dimensions for the diffrent Loci */
			vector<int>					m_vRateDimensions;
			/** Counts the unchanged Nucleotides for every loci and nuc [loci][nuc] */
			vector<vector<int> >		m_vvInvariantNucleotides;
			/** Counts the unchanged Nucleotides for every loci, site and nuc
			 * [loci][site][nuc] */
			vector<vector<vector<int> > > m_vvvInvariantNucleotides;
			// MCMCMC Data
			/** Number of Burn In Steps */
			int							m_burnIn;
			/** Number of Sample Frequency */
			int							m_sampleFrequency;
			/** Number of Samples */
			int							m_nSamples;
			/** Number of Chains */
			int							m_nChains;
			/** Number of Temperature Coefficient */
			double						m_heatingCoefficient;
			/** Frequency of Swapping the Heated Chains */
			int							m_swapFrequency;
			/** Vector of Constraints */
			//vector<TreeTime_Core::Constraint*> m_vpConstraints;
			vector<TreeTime_Constraint::TopologicalConstraint*>	m_vpConstraints;
			vector<TreeTime_Constraint::TopologicalConstraint*>	m_vpTopConstraints;
			vector<TreeTime_Constraint::GaussianConstraint*>	m_vpGauConstraints;
			vector<TreeTime_Constraint::GammaConstraint*>		m_vpGamConstraints;
			// Weighting:
			double						m_weightTopology;
			double						m_weightNodes;
			double						m_weightEvolution;
			double						m_weightRatechange;
			double						m_weightTimeconversion;

			/** Change Calibration */
			double						m_cNodeTime;
			double						m_cTimeSynchronisation;
			double						m_cRootTime;

	};//Data

}//TreeTime_NexusProcessing

#include "Data.cxx"

#endif /*DATA_HPP_*/
