/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "SymDistMatrix.hpp"

namespace TreeTime_NexusProcessing
{
	
	SymDistMatrix::SymDistMatrix(unsigned int p_dim, double p_init)
	: m_dim(p_dim)
	{
		m_data = new double*[m_dim];
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			m_data[dim1] = new double[dim1+1];
			//initialise:
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				m_data[dim1][dim2] = p_init; 
			}
		}
	}
	
	SymDistMatrix::SymDistMatrix(const SymDistMatrix& p_rhs)
	: m_dim(p_rhs.getDimension())
	{
		m_data = new double*[m_dim];
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			m_data[dim1] = new double[dim1+1];
			//assign:
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				m_data[dim1][dim2] = p_rhs(dim1,dim2);
			}
		}
	}
	
	SymDistMatrix& SymDistMatrix::operator=(const SymDistMatrix& p_rhs)
	{
		if(this != &p_rhs)
		{
			for(unsigned int dim1=0; dim1<m_dim; ++dim1)
			{
				for(unsigned int dim2=0; dim2<=dim1; ++dim2)
				{
					m_data[dim1][dim2] = p_rhs(dim1,dim2);
				}
			}
		}
		return *this;
	}
	
	SymDistMatrix::~SymDistMatrix()
	{
		for(unsigned int dim=0; dim<m_dim; ++dim)
		{
			delete [] m_data[dim]; 
		}
		delete [] m_data;
	}
	
	
	SymDistMatrix& SymDistMatrix::operator +=(const SymDistMatrix& p_rhs)
	{
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				m_data[dim1][dim2] += p_rhs(dim1,dim2);
			}
		}
		return *this;
	}
	
	SymDistMatrix& SymDistMatrix::operator -=(const SymDistMatrix& p_rhs)
	{
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				m_data[dim1][dim2] -= p_rhs(dim1,dim2);
			}
		}
		return *this;
	}
	
	SymDistMatrix& SymDistMatrix::operator *=(double p_multi)
	{
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				m_data[dim1][dim2] *= p_multi;
			}
		}
		return *this;
	}
	
	SymDistMatrix& SymDistMatrix::operator /=(double p_divisor)
	{
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				m_data[dim1][dim2] /= p_divisor;
			}
		}
		return *this;
	}
	
	const SymDistMatrix SymDistMatrix::operator +(const SymDistMatrix& p_rhs)
	{
		return SymDistMatrix(*this) += p_rhs;
	}
	
	const SymDistMatrix SymDistMatrix::operator -(const SymDistMatrix& p_rhs)
	{
		return SymDistMatrix(*this) -= p_rhs;
	}
	
	
	double& SymDistMatrix::operator()(unsigned int p_dim1, unsigned int p_dim2)
	{
		if(p_dim1>p_dim2)
			return m_data[p_dim1][p_dim2];
		return m_data[p_dim2][p_dim1];
	}
	
	const double& SymDistMatrix::operator()(unsigned int p_dim1, unsigned int p_dim2) const
	{
		if(p_dim1>p_dim2)
			return m_data[p_dim1][p_dim2];
		return m_data[p_dim2][p_dim1];
	}
	
	double SymDistMatrix::calcEuclideanDistance(const SymDistMatrix& p_rhs) const
	{
		double temp=0;
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<dim1; ++dim2)
			{
				temp += pow(p_rhs(dim1,dim2)-this->operator ()(dim1,dim2),2.0); 
			}
		}
		temp *= 2.0;
		for(unsigned int dim=0; dim<m_dim; ++dim)
		{
			temp += pow(p_rhs(dim,dim)-this->operator ()(dim,dim),2.0);
		}
		return pow(temp,0.5);
	}
	
	unsigned int  SymDistMatrix::getDimension() const
	{
		return m_dim;
	}
	
	void SymDistMatrix::print()
	{
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				cout << m_data[dim1][dim2]<<" "; 
			}cout << endl;
		}
	}
	
	void SymDistMatrix::print() const
	{
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				cout << m_data[dim1][dim2]<<" "; 
			}cout << endl;
		}
	}
	
	pair<int,int> SymDistMatrix::getMinimum() const
	{
		unsigned int i = 0;
		unsigned int j = 0;
		double min = std::numeric_limits<double>::max();
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				if(m_data[dim1][dim2] < min)
				{
					min = m_data[dim1][dim2];
					i = dim1;
					j = dim2;
				}
			}
		}
		return pair<int,int>(i,j);
	}
	
	pair<int,int> SymDistMatrix::getMaximum() const
	{
		unsigned int i = 0;
		unsigned int j = 0;
		double max = std::numeric_limits<double>::min();
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				if(m_data[dim1][dim2] > max)
				{
					max = m_data[dim1][dim2];
					i = dim1;
					j = dim2;
				}
			}
		}
		return pair<int,int>(i,j);
	}
	
	double SymDistMatrix::getMin() const
	{
		double min = std::numeric_limits<double>::max();
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				if(m_data[dim1][dim2] < min)
				{
					min = m_data[dim1][dim2];
				}
			}
		}
		return min;
	}
	
	double SymDistMatrix::getMax() const
	{
		double max = std::numeric_limits<double>::min();
		for(unsigned int dim1=0; dim1<m_dim; ++dim1)
		{
			for(unsigned int dim2=0; dim2<=dim1; ++dim2)
			{
				if(m_data[dim1][dim2] > max)
				{
					max = m_data[dim1][dim2];
				}
			}
		}
		return max;
	}

	
}//TreeTime_NexusProcessing
