/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "Upgma.hpp"
#include "SymDistMatrix.hpp"

namespace TreeTime_NexusProcessing
{

	Upgma::Upgma(const SymDistMatrix* const p_pDM)
	: m_pDistanceMatrix(p_pDM)
	{
	}

	Upgma::~Upgma()
	{
	}

	void Upgma::setDistanceMatrix(const SymDistMatrix* const p_pDM)
	{
		m_pDistanceMatrix = p_pDM;
	}
	
	string Upgma::calcNewickString(vector<string>* p_pvNames) const
	{
		UpgmaCluster* pFirstCluster = startUpgma();
		if(pFirstCluster->getChildCluster().size()<2)
			return "0:0;";
		string newick("(");
		list<const UpgmaCluster*>::const_iterator I =
				pFirstCluster->getChildCluster().begin();
		calcNewickString(*I,newick,pFirstCluster->getHeight(),p_pvNames);
		for(++I;I !=pFirstCluster->getChildCluster().end() ;++I)
		{
			newick += ",";
			calcNewickString(*I,newick,pFirstCluster->getHeight(),p_pvNames);			
		}
		newick += "):0;";
		delete pFirstCluster;
		return newick;
	}

	void Upgma::calcNewickString(const UpgmaCluster* p_pCluster,
			string& p_newick, double p_fatherTimepoint,
			vector<string>* p_pvNames) const
	{
		if(p_pCluster->getChildCluster().begin() == 
				p_pCluster->getChildCluster().end())
		{
			std::stringstream ss;
			if(p_pvNames)
				ss << (*p_pvNames)[*(p_pCluster->getIds().begin())]
						<< ":" << p_fatherTimepoint - p_pCluster->getHeight();
			else
				ss << *(p_pCluster->getIds().begin()) << ":"
						<< p_fatherTimepoint - p_pCluster->getHeight();
			p_newick += ss.str();
		}
		else
		{
			list<const UpgmaCluster*>::const_iterator I =
					p_pCluster->getChildCluster().begin();
			p_newick += "(";
			calcNewickString(*I,p_newick,p_pCluster->getHeight(),p_pvNames);
			for(++I;I !=p_pCluster->getChildCluster().end() ;++I)
			{
				p_newick += ",";
				calcNewickString(*I,p_newick,p_pCluster->getHeight(),p_pvNames);			
			}	
			p_newick += "):";
			std::stringstream ss;
			ss << p_fatherTimepoint - p_pCluster->getHeight();
			p_newick += ss.str();
		}
	}
	
	
	
	Upgma::UpgmaCluster* Upgma::startUpgma() const
	{
		SymDistMatrix DM(2*m_pDistanceMatrix->getDimension()-1,0.0);
		for(unsigned int dim1=0; dim1<m_pDistanceMatrix->getDimension(); ++dim1)
			for(unsigned int dim2=0; dim2<dim1; ++dim2)
				DM(dim1,dim2) = (*m_pDistanceMatrix)(dim1,dim2);
		// Init Algorithm
		map<int,UpgmaCluster*> AllCluster;
		for(unsigned int i=0; i<m_pDistanceMatrix->getDimension(); ++i)
		{
			AllCluster.insert(AllCluster.end(),pair<int,UpgmaCluster*>
					(i,new UpgmaCluster(0.0,i)));
		}
		int nextClusterId = m_pDistanceMatrix->getDimension();
		// Perform Algorithm
		while(AllCluster.size()>1)
		{
			pair<int,int> point = getMin(DM,AllCluster);
			UpgmaCluster* pCluster = mergeCluster(AllCluster[point.first],
					AllCluster[point.second], DM(point.first,point.second)/2.0);
			AllCluster.erase(point.first);
			AllCluster.erase(point.second);
			updateDistances(DM,AllCluster,pCluster,nextClusterId);
			AllCluster.insert(AllCluster.end(),
					pair<int,UpgmaCluster*>(nextClusterId,pCluster));
			++nextClusterId;
		}
		return AllCluster.begin()->second;
	}
	
	Upgma::UpgmaCluster* Upgma::mergeCluster(const UpgmaCluster* p_lhs, 
			const UpgmaCluster* p_rhs, double p_height) const
	{
		UpgmaCluster* pCluster = new UpgmaCluster(p_height);
		pCluster->insertChildCluster(p_lhs);
		pCluster->insertChildCluster(p_rhs);
		pCluster->insertIds(p_lhs->getIds());
		pCluster->insertIds(p_rhs->getIds());
		return pCluster;
	}
	
	pair<int,int> Upgma::getMin(const SymDistMatrix& p_DM, const map<int,
			UpgmaCluster*>& p_mIndex) const
	{
		int i=0;
		int j=0;
		double min = std::numeric_limits<double>::max();
		for(map<int,UpgmaCluster*>::const_iterator I = p_mIndex.begin();
				I!=p_mIndex.end();++I)
		{
			map<int,UpgmaCluster*>::const_iterator J = I;
			for(++J; J != p_mIndex.end(); ++J)
			{
				if(p_DM(I->first,J->first) < min)
				{
					min = p_DM(I->first,J->first);
					i = I->first;
					j = J->first;
				}
			}
		}
		return pair<int,int>(i,j);
	}
	
	void Upgma::updateDistances(SymDistMatrix& p_DM,
			map<int,UpgmaCluster*>& p_mI2C, const UpgmaCluster* p_pNewCluster,
			int p_newIndex) const
	{
		for(map<int,UpgmaCluster*>::iterator I = p_mI2C.begin();
				I != p_mI2C.end(); ++I)
			p_DM(I->first,p_newIndex) = p_pNewCluster->calcDistance(I->second,
					p_DM);
	}
	
	
//------------------------------------------------------------------------------
//---------------------UggmaCluster---------------------------------------------
//------------------------------------------------------------------------------

	Upgma::UpgmaCluster::UpgmaCluster(double p_height)
	: m_height(p_height)
	, m_sIds()
	, m_lChildCluster()
	{
	}
	
	Upgma::UpgmaCluster::UpgmaCluster(double p_height, int p_id)
	: m_height(p_height)
	, m_sIds()
	, m_lChildCluster()
	{
		m_sIds.insert(p_id);
	}
	
	Upgma::UpgmaCluster::~UpgmaCluster()
	{
		for(list<const UpgmaCluster*>::iterator I = m_lChildCluster.begin();
				I!=m_lChildCluster.end();++I)
			delete *I;
	}
	
	
	double Upgma::UpgmaCluster::getHeight() const
	{
		return m_height;
	}
	
	void Upgma::UpgmaCluster::insertChildCluster(const UpgmaCluster*
			p_pChildCluster)
	{
		m_lChildCluster.push_back(p_pChildCluster);
	}
	
	const set<int>& Upgma::UpgmaCluster::getIds() const
	{
		return m_sIds;
	}
	
	void Upgma::UpgmaCluster::insertIds(const set<int>& p_sIds)
	{
		m_sIds.insert(p_sIds.begin(),p_sIds.end());
	}
	
	const list<const Upgma::UpgmaCluster*>&
			Upgma::UpgmaCluster::getChildCluster() const
	{
		return m_lChildCluster;
	}
	
	
	double Upgma::UpgmaCluster::calcDistance(const UpgmaCluster* p_rhs,
			const SymDistMatrix& p_DM) const
	{
		int		count = 0;
		double	sum   = 0;
		for(set<int>::iterator I = m_sIds.begin(); I != m_sIds.end(); ++I)
		{
			for(set<int>::iterator J = p_rhs->getIds().begin();
					J != p_rhs->getIds().end(); ++J)
			{
			 	sum += p_DM(*I,*J);
			 	++count;
			}
		}
		return sum/static_cast<double>(count);
	}
	


}//TreeTime_NexusProcessing
