/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "CppChangeBirth.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/CompoundPoissonProcess.hpp"

namespace TreeTime_RateChangeModel
{

	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;

	CppChangeBirth::CppChangeBirth(State* p_pState, double p_proposalRatio,
			int p_loci)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_pProposalTree				(0)
	, m_proposalTreeLogLikelihood	(0)
	, m_proposalRateLogLikelihood	(0)
	, m_currentRateLogLikelihood	(0)
	, m_pProposalCpp				(0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	CppChangeBirth::~CppChangeBirth()
	{
	}

	void CppChangeBirth::changeState()
	{
		if(m_loci == -1)//global RateChangeModel
		{
			m_pProposalCpp = dynamic_cast<CompoundPoissonProcess<State>*>
					(m_pState->m_pGlobalRateChangeModel);
			m_currentRateLogLikelihood = m_pState->m_globalRateLoglikelihood;
		}
		else//local RateChangeModel
		{
			m_pProposalCpp = dynamic_cast<CompoundPoissonProcess<State>*>
					(m_pState->m_vpLocalRateChangeModel[m_loci]);
			m_currentRateLogLikelihood = m_pState->m_vLocalRateLoglikelihood
					[m_loci];
		}
		m_pProposalTree = m_pState->m_pTree->copy();
		double modulationValue = RNG::getRNG().getGamma(
				m_pProposalCpp->getAlpha(), m_pProposalCpp->getBeta());
		double valueDensity = RNG::getRNG().getGammaDensity(modulationValue,
				m_pProposalCpp->getAlpha(), m_pProposalCpp->getBeta());
		double treeLength = LC::getTreeLength(m_pState->m_pTree);
		double insertPlace = treeLength * RNG::getRNG().getUniform();
		Iterator I = m_pProposalTree->getRoot();
		stack<Iterator> sIterator;
		sIterator.push(I);
		double tempLength = 0;
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			tempLength += I->m_length;
			if(tempLength>=insertPlace)
			{
				tempLength -= insertPlace;
				break;
			}
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		list<pair<double,double> >& lChangePoints = I->m_mlChangepoints[m_loci];
		for(list<pair<double,double> >::iterator lI = lChangePoints.begin()
				;lI != lChangePoints.end(); ++lI)
		{
			if(lI->first > tempLength)
			{
				lChangePoints.insert(lI,pair<double,double>(tempLength,
						modulationValue));
				break;
			}
		}
		Iterator insertLocation = I;
		// allocate Memory
		allocateSubtreeMemory(I);
		while(I.getFather())
		{
			I = I.getFather();
			m_lpCurrentProbablities.push_back(I->m_molecularData);
			LC::allocateMemory(I);
			m_lpProposalProbablities.push_back(I->m_molecularData);
		}
		//scale Subtree
		if(insertLocation.hasChildNodes())
		{
			adjustSubtreeInitialRates(insertLocation.getFirstChild(), m_loci,
					modulationValue);
			adjustSubtreeInitialRates(insertLocation.getLastChild(), m_loci,
					modulationValue);
		}
		calcRateMultiplier(insertLocation, m_loci);
		// todo auch in UnterBumen die RatenMultiplier neu berechnen!!! nein
		// todo: optimize: you have not to calculate all loci..., but than copy!
		LC::calcScaledSubtreeDnaProbabilities(insertLocation,
				m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		insertLocation = insertLocation.getFather();
		if(insertLocation)
		{
			LC::calcScaledUpwardsDnaProbabilities(insertLocation,
					m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		}
		m_proposalTreeLogLikelihood = LC::getLogLikelihoodTree(m_pProposalTree,
				m_pState->m_vpDnaEvolutionModel);
		m_proposalRateLogLikelihood = m_currentRateLogLikelihood + 
				log( m_pProposalCpp->getLambda() * valueDensity /
				static_cast<double>(m_pProposalCpp->getNrChangePoints()+1));
		m_proposalRatio *= ((LC::getTreeLength(m_pState->m_pTree)/
				m_pState->m_pTree->getRoot()->m_timepoint)/valueDensity)/
				(m_pProposalCpp->getNrChangePoints()+1);
		// Proposed in Huelsenbeck 2000, with ordering of map
//		m_proposalRatio *= ((m_pState->m_treeLength/
//				m_pState->m_pTree->getRoot()->m_timepoint)/valueDensity);
	}

	double CppChangeBirth::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood -
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood) );
	}

	double CppChangeBirth::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void CppChangeBirth::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
		m_pState->m_pTree = m_pProposalTree;
		m_pProposalCpp->setNrChangePoints(m_pProposalCpp->getNrChangePoints()
				+1);
		m_pState->m_totalLogLikelihood += (
				(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood)-
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood));
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
		if(m_loci == -1)
		{
			m_pState->m_globalRateLoglikelihood = m_proposalRateLogLikelihood;
		}
		else
		{
			m_pState->m_vLocalRateLoglikelihood[m_loci] =
					m_proposalRateLogLikelihood;
		}
	}

	void CppChangeBirth::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			TreeTime_Core::LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	void CppChangeBirth::allocateSubtreeMemory(Iterator p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LC::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	void CppChangeBirth::adjustSubtreeInitialRates(Iterator p_iterator,
			int p_loci, double p_multiplier)
	{
		p_iterator->m_mlChangepoints[p_loci].begin()->second *= p_multiplier;
		if(p_loci == -1)
		{
			p_iterator->m_globalRateMultiplier *= p_multiplier;
		}
		else
		{
			p_iterator->m_vLocalRateMultiplier[p_loci] *= p_multiplier;
		}
		if(p_iterator.hasChildNodes())
		{
			adjustSubtreeInitialRates(p_iterator.getFirstChild(), p_loci,
					p_multiplier);
			adjustSubtreeInitialRates(p_iterator.getLastChild(), p_loci,
					p_multiplier);
		}
	}

	void CppChangeBirth::calcRateMultiplier(Iterator p_iterator, int p_loci)
	{
		list<pair<double,double> >::iterator I = p_iterator->
				m_mlChangepoints[p_loci].begin();
		double length	= 0;
		double position	= I->first;
		double rate		= I->second;
		for(++I; I != p_iterator->m_mlChangepoints[p_loci].end(); ++I)
		{
			length		+= ((I->first-position)*rate);
			position	 = I->first;
			rate		*= I->second;
		}
		if(p_loci == -1)
		{
			p_iterator->m_globalRateMultiplier = length/p_iterator->m_length;
		}
		else
		{
			p_iterator->m_vLocalRateMultiplier[p_loci] = length/p_iterator->
					m_length;
		}
	}

	int CppChangeBirth::getId () const
	{
		return 9;
	}

}//TreeTime_RateChangeModel
