/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "CppChangeDeath.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/CompoundPoissonProcess.hpp"

namespace TreeTime_RateChangeModel
{
	
	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;
	
	CppChangeDeath::CppChangeDeath(State* p_pState, double p_proposalRatio,
			int p_loci)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_pProposalTree				(0)
	, m_proposalTreeLogLikelihood	(0)
	, m_proposalRateLogLikelihood	(0)
	, m_currentRateLogLikelihood	(0)
	, m_pProposalCpp				(0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	CppChangeDeath::~CppChangeDeath()
	{
	}

	void CppChangeDeath::changeState()
	{
		if(m_loci == -1)//global RateChangeModel
		{
			m_pProposalCpp = dynamic_cast<CompoundPoissonProcess<State>*>
					(m_pState->m_pGlobalRateChangeModel);
			m_currentRateLogLikelihood = m_pState->m_globalRateLoglikelihood;
		}
		else//local RateChangeModel
		{
			m_pProposalCpp = dynamic_cast<CompoundPoissonProcess<State>*>
					(m_pState->m_vpLocalRateChangeModel[m_loci]);
			m_currentRateLogLikelihood = m_pState->m_vLocalRateLoglikelihood
					[m_loci];
		}
		m_pProposalTree		= m_pState->m_pTree->copy();
		int deletePosition	= RNG::getRNG().getInteger(m_pProposalCpp->
				getNrChangePoints());
		double modulationValue = 1.0;
		Iterator I = m_pProposalTree->getRoot();
		stack<Iterator > sIterator;
		sIterator.push(I);
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			deletePosition -= (I->m_mlChangepoints[m_loci].size()-2);
			if(deletePosition < 0)
			{
				list<pair<double,double> >::iterator lp =
						I->m_mlChangepoints[m_loci].begin();
				while(deletePosition<0)
				{
					++lp;
					++deletePosition;
				}
				modulationValue = lp->second;
				I->m_mlChangepoints[m_loci].erase(lp);
				break;
			}
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		double valueDensity = RNG::getRNG().getGammaDensity(modulationValue,
				m_pProposalCpp->getAlpha(), m_pProposalCpp->getBeta());
		Iterator deleteLocation = I;
		allocateSubtreeMemory(I);
		while(I.getFather())
		{
			I = I.getFather();
			m_lpCurrentProbablities.push_back(I->m_molecularData);
			LC::allocateMemory(I);
			m_lpProposalProbablities.push_back(I->m_molecularData);
		}
		//scale Subtree
		if(deleteLocation.hasChildNodes())
		{
			adjustSubtreeInitialRates(deleteLocation.getFirstChild(), m_loci,
					1.0/modulationValue);
			adjustSubtreeInitialRates(deleteLocation.getLastChild(), m_loci,
					1.0/modulationValue);
		}
		calcRateMultiplier(deleteLocation, m_loci);
		// todo: optimize: you have not to calculate all loci..., but than copy!
		LC::calcScaledSubtreeDnaProbabilities(deleteLocation,
				m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		deleteLocation = deleteLocation.getFather();
		if(deleteLocation)
		{
			LC::calcScaledUpwardsDnaProbabilities(deleteLocation,
					m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		}
		m_proposalTreeLogLikelihood = LC::getLogLikelihoodTree(
				m_pProposalTree, m_pState->m_vpDnaEvolutionModel);
		m_proposalRateLogLikelihood = m_currentRateLogLikelihood + 
				log(static_cast<double>(m_pProposalCpp->getNrChangePoints())/
				(m_pProposalCpp->getLambda()*valueDensity));
		m_proposalRatio *= ((valueDensity * m_pProposalCpp->getNrChangePoints())/
				(LC::getTreeLength(m_pState->m_pTree)/
				m_pState->m_pTree->getRoot()->m_timepoint));
		// Proposed in Huelsenbeck 2000, with ordering of map
//		m_proposalRatio *= (valueDensity/(m_pState->m_treeLength/
//				m_pState->m_pTree->getRoot()->m_timepoint));
	}

	double CppChangeDeath::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood -
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood));
	}

	double CppChangeDeath::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void CppChangeDeath::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
		
		
		m_pState->m_pTree = m_pProposalTree;
		m_pProposalCpp->setNrChangePoints(m_pProposalCpp->getNrChangePoints()
				-1);
		m_pState->m_totalLogLikelihood += (
				(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood)-
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood));
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
		if(m_loci == -1)
		{
			m_pState->m_globalRateLoglikelihood = m_proposalRateLogLikelihood;
		}
		else
		{
			m_pState->m_vLocalRateLoglikelihood[m_loci] =
					m_proposalRateLogLikelihood;
		}
	}

	void CppChangeDeath::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
	}

	void CppChangeDeath::allocateSubtreeMemory(
			Iterator p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LC::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	void CppChangeDeath::adjustSubtreeInitialRates(Iterator p_iterator,
			int p_loci, double p_multiplier)
	{
		p_iterator->m_mlChangepoints[p_loci].begin()->second *= p_multiplier;
		if(p_loci == -1)
		{
			p_iterator->m_globalRateMultiplier *= p_multiplier;
		}
		else
		{
			p_iterator->m_vLocalRateMultiplier[p_loci] *= p_multiplier;
		}
		if(p_iterator.hasChildNodes())
		{
			adjustSubtreeInitialRates(p_iterator.getFirstChild(), p_loci,
					p_multiplier);
			adjustSubtreeInitialRates(p_iterator.getLastChild(), p_loci,
					p_multiplier);
		}
	}

	void CppChangeDeath::calcRateMultiplier(Iterator p_iterator, int p_loci)
	{
		list<pair<double,double> >::iterator I = p_iterator->
				m_mlChangepoints[p_loci].begin();
		double length	= 0;
		double position	= I->first;
		double rate		= I->second;
		for(++I; I != p_iterator->m_mlChangepoints[p_loci].end(); ++I)
		{
			length		+= ((I->first-position)*rate);
			position	 = I->first;
			rate		*= I->second;
		}
		if(p_loci == -1)
		{
			p_iterator->m_globalRateMultiplier = length/p_iterator->m_length;
		}
		else
		{
			p_iterator->m_vLocalRateMultiplier[p_loci] = length/p_iterator->
					m_length;
		}
	}

	int CppChangeDeath::getId () const
	{
		return 10;
	}

}//TreeTime_RateChangeModel
