/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "CppChangeRateDistribution.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/CompoundPoissonProcess.hpp"

namespace TreeTime_RateChangeModel
{

	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;

	CppChangeRateDistribution::CppChangeRateDistribution(State* p_pState,
			double p_proposalRatio, int p_loci, double p_calibration)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_pCurrentCpp					(0)
	, m_pProposalCpp				(0)
	, m_proposalRateLogLikelihood	(0)
	, m_currentRateLogLikelihood	(0)
	, m_calibration					(p_calibration)
	{
	}

	CppChangeRateDistribution::~CppChangeRateDistribution()
	{
	}

	void CppChangeRateDistribution::changeState()
	{
		if(m_loci == -1)//global RateChangeModel
		{
			m_pCurrentCpp = dynamic_cast<CompoundPoissonProcess<State>*>
					(m_pState->m_pGlobalRateChangeModel);
			m_currentRateLogLikelihood = m_pState->m_globalRateLoglikelihood;
		}
		else//local RateChangeModel
		{
			m_pCurrentCpp = dynamic_cast<CompoundPoissonProcess<State>*>
					(m_pState->m_vpLocalRateChangeModel[m_loci]);
			m_currentRateLogLikelihood = m_pState->m_vLocalRateLoglikelihood
					[m_loci];
		}
		m_pProposalCpp = m_pCurrentCpp->clone();
		double alpha = m_pProposalCpp->getAlpha();
		alpha += RNG::getRNG().getUniform(
				-1.0 * m_calibration, 1.0 * m_calibration);
		// alpha bigger => less variance
		if(alpha < 0.2)
		{
			alpha = 0.4 - alpha;
		}
		else if(alpha > 20.0)
		{
			alpha = 40.0 - alpha;
		}
		m_pProposalCpp->setAlpha(alpha);
		m_proposalRateLogLikelihood = m_pProposalCpp->getLogDensity(m_pState,
				m_loci);
	}

	double CppChangeRateDistribution::getStationaryDistributionRatio() const
	{
		return exp(m_proposalRateLogLikelihood - m_currentRateLogLikelihood);
	}

	double CppChangeRateDistribution::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void CppChangeRateDistribution::accept()
	{
		delete m_pCurrentCpp;
		if(m_loci == -1)//global RateChangeModel
		{
			m_pState->m_pGlobalRateChangeModel  = m_pProposalCpp;
			m_pState->m_globalRateLoglikelihood = m_proposalRateLogLikelihood;
		}
		else//local RateChangeModel
		{
			m_pState->m_vpLocalRateChangeModel[m_loci]  = m_pProposalCpp;
			m_pState->m_vLocalRateLoglikelihood[m_loci] = 
					m_proposalRateLogLikelihood;
		}
		m_pState->m_totalLogLikelihood	+= (m_proposalRateLogLikelihood -
				m_currentRateLogLikelihood);
	}

	void CppChangeRateDistribution::reject()
	{
		delete m_pProposalCpp;
	}

	int CppChangeRateDistribution::getId () const
	{
		return 14;
	}

}//TreeTime_Core
