/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "CppChangeRateValue.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/CompoundPoissonProcess.hpp"

namespace TreeTime_RateChangeModel
{

	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;

	CppChangeRateValue::CppChangeRateValue(State* p_pState,
			double p_proposalRatio, int p_loci)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_pProposalTree				(0)
	, m_proposalTreeLogLikelihood	(0)
	, m_tempRatioMultiplier			(1.0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	CppChangeRateValue::~CppChangeRateValue()
	{
	}

	void CppChangeRateValue::changeState()
	{
		CompoundPoissonProcess<State>* pProposalCpp;
		if(m_loci == -1)//global RateChangeModel
		{
			pProposalCpp = dynamic_cast<CompoundPoissonProcess<State>*>
					(m_pState->m_pGlobalRateChangeModel);
		}
		else//local RateChangeModel
		{
			pProposalCpp = dynamic_cast<CompoundPoissonProcess<State>*>
					(m_pState->m_vpLocalRateChangeModel[m_loci]);
		}
		m_pProposalTree = m_pState->m_pTree->copy();
		int changePosition = RNG::getRNG().getInteger(pProposalCpp->
				getNrChangePoints());
		Iterator I = m_pProposalTree->getRoot();
		stack<Iterator> sIterator;
		sIterator.push(I);
		double modulationRatio = 1.0;
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			changePosition -= (I->m_mlChangepoints[m_loci].size()-2);
			if(changePosition < 0)
			{
				list<pair<double,double> >::iterator lp =
						I->m_mlChangepoints[m_loci].begin();
				while(changePosition < 0)
				{
					++lp;
					++changePosition;
				}
				modulationRatio = 1.0/lp->second;
				m_tempRatioMultiplier = 1.0/
						RNG::getRNG().getGammaDensity(
						lp->second,
						pProposalCpp->getAlpha(),
						pProposalCpp->getBeta());
				double temp = RNG::getRNG().getGamma(
						pProposalCpp->getAlpha(),
						pProposalCpp->getBeta());			
				if(temp < 0.1)
				{
					temp = 0.2 - temp;
				}
				else if(temp > 10.0)
				{
					temp = 20.0 - temp;
				}	
				lp->second = temp;
				modulationRatio *= lp->second;
				m_tempRatioMultiplier *= 
						RNG::getRNG().getGammaDensity(
						lp->second,
						pProposalCpp->getAlpha(),
						pProposalCpp->getBeta());
				break;
			}
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		Iterator insertLocation = I;
		// allocate Memory
		allocateSubtreeMemory(I);
		while(I.getFather())
		{
			I = I.getFather();
			m_lpCurrentProbablities.push_back(I->m_molecularData);
			LC::allocateMemory(I);
			m_lpProposalProbablities.push_back(I->m_molecularData);
		}
		//scale Subtree
		if(insertLocation.hasChildNodes())
		{
			adjustSubtreeInitialRates(insertLocation.getFirstChild(), m_loci,
					modulationRatio);
			adjustSubtreeInitialRates(insertLocation.getLastChild(), m_loci,
					modulationRatio);
		}
		calcRateMultiplier(insertLocation, m_loci);
		LC::calcScaledSubtreeDnaProbabilities(insertLocation,
				m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		insertLocation = insertLocation.getFather();
		if(insertLocation)
		{
			LC::calcScaledUpwardsDnaProbabilities(insertLocation,
					m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		}
		m_proposalTreeLogLikelihood = LC::getLogLikelihoodTree(m_pProposalTree,
				m_pState->m_vpDnaEvolutionModel);
		m_proposalRatio /= m_tempRatioMultiplier;
	}

	double CppChangeRateValue::getStationaryDistributionRatio() const
	{
		return (exp(m_proposalTreeLogLikelihood - m_pState->m_treeLogLikelihood)
				* m_tempRatioMultiplier);
	}

	double CppChangeRateValue::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void CppChangeRateValue::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
		m_pState->m_pTree = m_pProposalTree;
		if(m_loci == -1)//global RateChangeModel
		{
			m_pState->m_globalRateLoglikelihood += log(m_tempRatioMultiplier);
		}
		else//local RateChangeModel
		{
			m_pState->m_vLocalRateLoglikelihood[m_loci] +=
					log(m_tempRatioMultiplier);
		}
		m_pState->m_totalLogLikelihood += (log(m_tempRatioMultiplier) +
				m_proposalTreeLogLikelihood - m_pState->m_treeLogLikelihood);
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
	}

	void CppChangeRateValue::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
	}

	void CppChangeRateValue::allocateSubtreeMemory(Iterator p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LC::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	void CppChangeRateValue::adjustSubtreeInitialRates(Iterator p_iterator,
			int p_loci, double p_multiplier)
	{
		p_iterator->m_mlChangepoints[p_loci].begin()->second *= p_multiplier;
		if(p_loci == -1)
		{
			p_iterator->m_globalRateMultiplier *= p_multiplier;
		}
		else
		{
			p_iterator->m_vLocalRateMultiplier[p_loci] *= p_multiplier;
		}
		if(p_iterator.hasChildNodes())
		{
			adjustSubtreeInitialRates(p_iterator.getFirstChild(), p_loci,
					p_multiplier);
			adjustSubtreeInitialRates(p_iterator.getLastChild(), p_loci,
					p_multiplier);
		}
	}

	void CppChangeRateValue::calcRateMultiplier(Iterator p_iterator, int p_loci)
	{
		list<pair<double,double> >::iterator I = p_iterator->
				m_mlChangepoints[p_loci].begin();
		double length	= 0;
		double position	= I->first;
		double rate		= I->second;
		for(++I; I != p_iterator->m_mlChangepoints[p_loci].end(); ++I)
		{
			length		+= ((I->first-position)*rate);
			position	 = I->first;
			rate		*= I->second;
		}
		if(p_loci == -1)
		{
			p_iterator->m_globalRateMultiplier = length/p_iterator->m_length;
		}
		else
		{
			p_iterator->m_vLocalRateMultiplier[p_loci] = length/p_iterator->
					m_length;
		}
	}

	int CppChangeRateValue::getId () const
	{
		return 12;
	}

}//TreeTime_RateChangeModel
