/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "DmChangeAllRates.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/DirichletModel.hpp"

namespace TreeTime_RateChangeModel
{

	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;

	DmChangeAllRates::DmChangeAllRates(State* p_pState, double p_proposalRatio,
			int p_loci, int p_nRates, double p_calibration)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_nRates						(p_nRates)
	, m_approxVariance				(0.0002*p_calibration)
	, m_pProposalTree				(0)
	, m_proposalTreeLogLikelihood	(0.0)
	, m_proposalRateLogLikelihood	(0.0)
	, m_currentRateLogLikelihood	(0.0)
	, m_pProposalDm					(0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	DmChangeAllRates::~DmChangeAllRates()
	{
	}

	void DmChangeAllRates::changeState()
	{
		if(m_loci == -1)//global RateChangeModel
		{
			m_pProposalDm = dynamic_cast<DirichletModel<State>*>(m_pState->
					m_pGlobalRateChangeModel);
			m_currentRateLogLikelihood = m_pState->m_globalRateLoglikelihood;
		}
		else//local RateChangeModel
		{
			m_pProposalDm = dynamic_cast<DirichletModel<State>*>(m_pState->
					m_vpLocalRateChangeModel[m_loci]);
			m_currentRateLogLikelihood = m_pState->m_vLocalRateLoglikelihood
					[m_loci];
		}
		m_pProposalTree = m_pState->m_pTree->copy();
		double* currentValues       = new double[m_nRates];
		double* paramCurrentValues  = new double[m_nRates];
		double* proposalValues      = new double[m_nRates];
		double* paramProposalValues = new double[m_nRates];
		double** pProposalValues    = new double*[m_nRates];
		int index = 0;
		Iterator I = m_pProposalTree->getRoot();
		stack<Iterator> sIterator;
		sIterator.push(I.getFirstChild());
		sIterator.push(I.getLastChild());
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(m_loci == -1)
			{
				currentValues[index] = I->m_globalRateMultiplier/
						static_cast<double>(m_nRates);
				pProposalValues[index] = &(I->m_globalRateMultiplier);
			}
			else
			{
				currentValues[index] = I->m_vLocalRateMultiplier[m_loci]/
						static_cast<double>(m_nRates);
				pProposalValues[index] = &(I->m_vLocalRateMultiplier[m_loci]);
			}
			++index;
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		for(int i = 0; i<m_nRates; ++i)
		{
			paramCurrentValues[i] = currentValues[i]/m_approxVariance;
		}
		RNG::getRNG().getDirichlet(paramCurrentValues,proposalValues, m_nRates);
		for(int i = 0; i<m_nRates; ++i)
		{
			paramProposalValues[i] = proposalValues[i]/m_approxVariance;
			*(pProposalValues[i])  = proposalValues[i] * m_nRates;
		}
		m_proposalRatio *= exp ( RNG::getRNG().getLogDirichletDensity(
				paramProposalValues, currentValues, m_nRates) -
				RNG::getRNG().getLogDirichletDensity(paramCurrentValues,
				proposalValues, m_nRates) );
		// allocate Memory
		allocateSubtreeMemory(m_pProposalTree->getRoot());
		LC::calcScaledSubtreeDnaProbabilities(m_pProposalTree->getRoot(),
				m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		m_proposalTreeLogLikelihood = LC::getLogLikelihoodTree(m_pProposalTree,
				m_pState->m_vpDnaEvolutionModel);
		Phylogeny* tempPhylo = m_pState->m_pTree;
		m_pState->m_pTree = m_pProposalTree;
		m_proposalRateLogLikelihood = m_pProposalDm->getLogDensity(m_pState,
				m_loci);
		m_pState->m_pTree = tempPhylo;
		delete[] currentValues;
		delete[] paramCurrentValues;
		delete[] proposalValues;
		delete[] paramProposalValues;
		delete[] pProposalValues;
	}

	double DmChangeAllRates::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood -
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood) );
	}

	double DmChangeAllRates::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void DmChangeAllRates::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
		m_pState->m_pTree = m_pProposalTree;
		m_pState->m_totalLogLikelihood += (
				(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood)-
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood));
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
		if(m_loci == -1)
		{
			m_pState->m_globalRateLoglikelihood = m_proposalRateLogLikelihood;
		}
		else
		{
			m_pState->m_vLocalRateLoglikelihood[m_loci] =
					m_proposalRateLogLikelihood;
		}
	}

	void DmChangeAllRates::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			TreeTime_Core::LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	void DmChangeAllRates::allocateSubtreeMemory(Iterator p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LC::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	int DmChangeAllRates::getId () const
	{
		return 16;
	}

}//TreeTime_RateChangeModel
