/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "DmChangeTwoRates.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/DirichletModel.hpp"

namespace TreeTime_RateChangeModel
{

	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;

	DmChangeTwoRates::DmChangeTwoRates(State* p_pState, double p_proposalRatio,
			int p_loci, double p_calibration)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_calibration					(p_calibration)
	, m_pProposalTree				(0)
	, m_proposalTreeLogLikelihood	(0.0)
	, m_proposalRateLogLikelihood	(0.0)
	, m_currentRateLogLikelihood	(0.0)
	, m_pProposalDm					(0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	DmChangeTwoRates::~DmChangeTwoRates()
	{
	}

	void DmChangeTwoRates::changeState()
	{
		if(m_loci == -1)//global RateChangeModel
		{
			m_pProposalDm = dynamic_cast<DirichletModel<State>*>(m_pState->
					m_pGlobalRateChangeModel);
			m_currentRateLogLikelihood = m_pState->m_globalRateLoglikelihood;
		}
		else//local RateChangeModel
		{
			m_pProposalDm = dynamic_cast<DirichletModel<State>*>(m_pState->
					m_vpLocalRateChangeModel[m_loci]);
			m_currentRateLogLikelihood = m_pState->m_vLocalRateLoglikelihood
					[m_loci];
		}
		m_pProposalTree = m_pState->m_pTree->copy();
		int pos1=RNG::getRNG().getInteger(m_pProposalDm->getNRateMultipliers());
		int pos2=RNG::getRNG().getInteger(m_pProposalDm->getNRateMultipliers());
		while(pos1 == pos2)
		{
			pos2=RNG::getRNG().getInteger(m_pProposalDm->getNRateMultipliers());
		}
		int counter = 0;
		double* pRate1 = 0;
		double* pRate2 = 0;
		Iterator I = m_pProposalTree->getRoot();
		stack<Iterator> sIterator;
		sIterator.push(I.getFirstChild());
		sIterator.push(I.getLastChild());
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(counter == pos1)
			{
				if(m_loci == -1)
					pRate1 = &(I->m_globalRateMultiplier);
				else
					pRate1 = &(I->m_vLocalRateMultiplier[m_loci]);
			}
			else if(counter == pos2)
			{
				if(m_loci == -1)
					pRate2 = &(I->m_globalRateMultiplier);
				else
					pRate2 = &(I->m_vLocalRateMultiplier[m_loci]);
			}
			++counter;
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		// Change RateMultipliers
		double uniform = RNG::getRNG().getUniform(0, 0.2 * m_calibration);
		if((*pRate1>uniform) || (*pRate2>uniform))
		{
			double sum = *pRate1 + *pRate2;
			*pRate1 = fabs(*pRate1-uniform);
			*pRate2 = sum - *pRate1;
		}
		// allocate Memory
		allocateSubtreeMemory(m_pProposalTree->getRoot());
		LC::calcScaledSubtreeDnaProbabilities(m_pProposalTree->getRoot(),
				m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		m_proposalTreeLogLikelihood = LC::getLogLikelihoodTree(m_pProposalTree,
				m_pState->m_vpDnaEvolutionModel);
		Phylogeny* tempPhylo = m_pState->m_pTree;
		m_pState->m_pTree = m_pProposalTree;
		m_proposalRateLogLikelihood = m_pProposalDm->getLogDensity(m_pState,
				m_loci);
		m_pState->m_pTree = tempPhylo;
	}

	double DmChangeTwoRates::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood -
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood) );
	}

	double DmChangeTwoRates::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void DmChangeTwoRates::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
		m_pState->m_pTree = m_pProposalTree;
		m_pState->m_totalLogLikelihood += (
				(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood)-
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood));
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
		if(m_loci == -1)
		{
			m_pState->m_globalRateLoglikelihood = m_proposalRateLogLikelihood;
		}
		else
		{
			m_pState->m_vLocalRateLoglikelihood[m_loci] =
					m_proposalRateLogLikelihood;
		}
	}

	void DmChangeTwoRates::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			TreeTime_Core::LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	void DmChangeTwoRates::allocateSubtreeMemory(Iterator p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LC::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	int DmChangeTwoRates::getId () const
	{
		return 17;
	}

}//TreeTime_RateChangeModel
