/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "DmChangeVariance.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/DirichletModel.hpp"

namespace TreeTime_RateChangeModel
{

	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;

	DmChangeVariance::DmChangeVariance(State* p_pState, double p_proposalRatio,
			int p_loci, double p_nRates, double p_expectedVariance,
			double p_calibration)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_nRates						(p_nRates)
	, m_expectedVariance			(p_expectedVariance)
	, m_calibration					(p_calibration)
	, m_currentVariance				(0.0)
	, m_proposalVariance			(0.0)
	, m_proposalRateLogLikelihood	(0.0)
	, m_currentRateLogLikelihood	(0.0)
	, m_ratioMultiplier				(1.0)
	, m_pProposalDm					(0)
	{
	}

	DmChangeVariance::~DmChangeVariance()
	{
	}

	void DmChangeVariance::changeState()
	{
		if(m_loci == -1)//global RateChangeModel
		{
			m_pProposalDm = dynamic_cast<DirichletModel<State>*>(m_pState->
					m_pGlobalRateChangeModel);
			m_currentRateLogLikelihood = m_pState->m_globalRateLoglikelihood;
		}
		else//local RateChangeModel
		{
			m_pProposalDm = dynamic_cast<DirichletModel<State>*>(m_pState->
					m_vpLocalRateChangeModel[m_loci]);
			m_currentRateLogLikelihood = m_pState->m_vLocalRateLoglikelihood
					[m_loci];
		}
		m_currentVariance  = m_pProposalDm->getVariance();
		m_proposalVariance = m_currentVariance + RNG::getRNG().getUniform(
				-m_calibration * m_expectedVariance,
				m_calibration * m_expectedVariance);
		m_proposalVariance = std::min(m_proposalVariance,m_nRates-1.0);
		if(m_proposalVariance < 0.001)
		{
			m_proposalVariance = 0.002 - m_proposalVariance;
		}
		else if(m_proposalVariance>(m_nRates-1.0))
		{
			m_proposalVariance = (2.0*(m_nRates-1.0))-m_proposalVariance;
		}
		//else if(m_proposalVariance>(50.0*m_expectedVariance))
		//{
		//	m_proposalVariance = (100.0*m_expectedVariance)-m_proposalVariance;
		//}
		m_pProposalDm->setVariance(m_proposalVariance);
		m_proposalRateLogLikelihood = m_pProposalDm->getLogDensity(m_pState,
				m_loci);
		m_ratioMultiplier = exp(-(m_proposalVariance-m_currentVariance)
				/m_expectedVariance);
		m_pProposalDm->setVariance(m_currentVariance);
	}

	double DmChangeVariance::getStationaryDistributionRatio() const
	{
		return exp(m_proposalRateLogLikelihood - m_currentRateLogLikelihood)
				* m_ratioMultiplier;
	}

	double DmChangeVariance::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void DmChangeVariance::accept()
	{
		m_pState->m_totalLogLikelihood += (m_proposalRateLogLikelihood -
				m_currentRateLogLikelihood);
		if(m_loci == -1)
		{
			m_pState->m_globalRateLoglikelihood = m_proposalRateLogLikelihood;
		}
		else
		{
			m_pState->m_vLocalRateLoglikelihood[m_loci] =
					m_proposalRateLogLikelihood;
		}
		m_pProposalDm->setVariance(m_proposalVariance);
	}

	void DmChangeVariance::reject()
	{
	}

	int DmChangeVariance::getId () const
	{
		return 15;
	}

}//TreeTime_RateChangeModel
