/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "UlnChangeAllRates.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/UncorrelatedLognormal.hpp"

namespace TreeTime_RateChangeModel
{

	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;

	UlnChangeAllRates::UlnChangeAllRates(State* p_pState, double p_proposalRatio,
			int p_loci, double p_changeCalibrationAll)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_changeCalibrationAll		(p_changeCalibrationAll)
	, m_pProposalTree				(0)
	, m_proposalTreeLogLikelihood	(0.0)
	, m_proposalRateLogLikelihood	(0.0)
	, m_currentRateLogLikelihood	(0.0)
	, m_pProposalUln				(0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	UlnChangeAllRates::~UlnChangeAllRates()
	{
	}

	void UlnChangeAllRates::changeState()
	{
		if(m_loci == -1)//global RateChangeModel
		{
			m_pProposalUln = dynamic_cast<UncorrelatedLognormal<State>*>(
					m_pState->m_pGlobalRateChangeModel);
			m_currentRateLogLikelihood = m_pState->m_globalRateLoglikelihood;
		}
		else//local RateChangeModel
		{
			m_pProposalUln = dynamic_cast<UncorrelatedLognormal<State>*>(
					m_pState->m_vpLocalRateChangeModel[m_loci]);
			m_currentRateLogLikelihood = m_pState->m_vLocalRateLoglikelihood
					[m_loci];
		}
		m_pProposalTree = m_pState->m_pTree->copy();
		Iterator I = m_pProposalTree->getRoot();
		stack<Iterator> sIterator;
		sIterator.push(I.getFirstChild());
		sIterator.push(I.getLastChild());
		double* pValue = 0;
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(m_loci == -1)
			{
				pValue = &(I->m_globalRateMultiplier);
			}
			else
			{
				pValue = &(I->m_vLocalRateMultiplier[m_loci]);
			}
			double uniform = RNG::getRNG().getUniform(-0.1, 0.1);
			uniform *= m_changeCalibrationAll;
			*pValue += uniform;
			if(*pValue < 0.1)
			{
				*pValue = 0.2 - *pValue;
			}
			else if(*pValue > 10.0)
			{
				*pValue = 20.0 - *pValue;
			}
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		// allocate Memory
		allocateSubtreeMemory(m_pProposalTree->getRoot());
		LC::calcScaledSubtreeDnaProbabilities(m_pProposalTree->getRoot(),
				m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		m_proposalTreeLogLikelihood = LC::getLogLikelihoodTree(m_pProposalTree,
				m_pState->m_vpDnaEvolutionModel);
		Phylogeny* tempPhylo = m_pState->m_pTree;
		m_pState->m_pTree = m_pProposalTree;
		m_proposalRateLogLikelihood = m_pProposalUln->getLogDensity(m_pState,
				m_loci);
		m_pState->m_pTree = tempPhylo;
	}

	double UlnChangeAllRates::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood -
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood) );
	}

	double UlnChangeAllRates::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void UlnChangeAllRates::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
		m_pState->m_pTree = m_pProposalTree;
		m_pState->m_totalLogLikelihood += (
				(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood)-
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood));
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
		if(m_loci == -1)
		{
			m_pState->m_globalRateLoglikelihood = m_proposalRateLogLikelihood;
		}
		else
		{
			m_pState->m_vLocalRateLoglikelihood[m_loci] =
					m_proposalRateLogLikelihood;
		}
	}

	void UlnChangeAllRates::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			TreeTime_Core::LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	void UlnChangeAllRates::allocateSubtreeMemory(Iterator p_iterator)
	{
		if(p_iterator.hasChildNodes())
		{
			m_lpCurrentProbablities.push_back(p_iterator->m_molecularData);
			LC::allocateMemory(p_iterator);
			m_lpProposalProbablities.push_back(p_iterator->m_molecularData);
			allocateSubtreeMemory(p_iterator.getFirstChild());
			allocateSubtreeMemory(p_iterator.getLastChild());
		}
	}

	int UlnChangeAllRates::getId () const
	{
		return 21;
	}

}//TreeTime_RateChangeModel
