/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "UlnChangeSingleRate.hpp"
#include "../../TreeTime_Core/RandomNumberGenerator.hpp"
#include "../../TreeTime_Core/LikelihoodCalculations.hpp"
#include "../../TreeTime_RateChangeModel/RateChangeModel.hpp"
#include "../../TreeTime_RateChangeModel/UncorrelatedLognormal.hpp"

namespace TreeTime_RateChangeModel
{

	typedef TreeTime_Core::RandomNumberGenerator RNG;
	typedef TreeTime_Core::LikelihoodCalculations LC;

	UlnChangeSingleRate::UlnChangeSingleRate(State* p_pState, double
			p_proposalRatio, int p_loci, int p_nRates, double p_calibration)
	: m_pState						(p_pState)
	, m_loci						(p_loci)
	, m_proposalRatio				(p_proposalRatio)
	, m_calibration					(p_calibration)
	, m_nRates						(p_nRates)
	, m_pProposalTree				(0)
	, m_proposalTreeLogLikelihood	(0.0)
	, m_proposalRateLogLikelihood	(0.0)
	, m_currentRateLogLikelihood	(0.0)
	, m_pProposalUln				(0)
	, m_lpCurrentProbablities		()
	, m_lpProposalProbablities		()
	{
	}

	UlnChangeSingleRate::~UlnChangeSingleRate()
	{
	}

	void UlnChangeSingleRate::changeState()
	{		
		if(m_loci == -1)//global RateChangeModel
		{
			m_pProposalUln = dynamic_cast<UncorrelatedLognormal<State>*>(
					m_pState->m_pGlobalRateChangeModel);
			m_currentRateLogLikelihood = m_pState->m_globalRateLoglikelihood;
		}
		else//local RateChangeModel
		{
			m_pProposalUln = dynamic_cast<UncorrelatedLognormal<State>*>(
					m_pState->m_vpLocalRateChangeModel[m_loci]);
			m_currentRateLogLikelihood = m_pState->m_vLocalRateLoglikelihood
					[m_loci];
		}
		m_pProposalTree = m_pState->m_pTree->copy();
		Iterator I = m_pProposalTree->getRoot();
		stack<Iterator> sIterator;
		sIterator.push(I.getFirstChild());
		sIterator.push(I.getLastChild());
		int counter = RNG::getRNG().getInteger(m_nRates);
		double* pValue = 0;
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(counter == 0)
			{
				if(m_loci == -1)
				{
					pValue = &(I->m_globalRateMultiplier);
				}
				else
				{
					pValue = &(I->m_vLocalRateMultiplier[m_loci]);
				}
				break;
			}
			--counter;
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		m_proposalRateLogLikelihood = m_currentRateLogLikelihood;
		m_proposalRateLogLikelihood -= log(RNG::getRNG().getLognormalDensity(
				*pValue, m_pProposalUln->getNormalMean(),
				m_pProposalUln->getNormalDeviation()));
		double uniform = RNG::getRNG().getUniform(-0.5, 0.5);
		uniform *= m_calibration;
		*pValue += uniform;
		if(*pValue < 0.1)
		{
			*pValue = 0.2 - *pValue;
		}
		else if(*pValue > 10.0)
		{
			*pValue = 20.0 - *pValue;
		}
		m_proposalRateLogLikelihood += log(RNG::getRNG().getLognormalDensity(
				*pValue, m_pProposalUln->getNormalMean(),
				m_pProposalUln->getNormalDeviation()));
		// allocate Memory
		Iterator insertLocation = I.getFather();
		I = I.getFather();
		while(I)
		{
			m_lpCurrentProbablities.push_back(I->m_molecularData);
			LC::allocateMemory(I);
			m_lpProposalProbablities.push_back(I->m_molecularData);
			I = I.getFather();
		}
		// todo make more efficient if multiple loci, but then copy the dna data
		LC::calcScaledUpwardsDnaProbabilities(insertLocation,
				m_pState->m_vpDnaEvolutionModel, m_pState->m_vRateScaling);
		m_proposalTreeLogLikelihood = LC::getLogLikelihoodTree(m_pProposalTree,
				m_pState->m_vpDnaEvolutionModel);
	}

	double UlnChangeSingleRate::getStationaryDistributionRatio() const
	{
		return exp(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood -
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood) );
	}

	double UlnChangeSingleRate::getProposalRatio() const
	{
		return m_proposalRatio;
	}

	void UlnChangeSingleRate::accept()
	{
		delete m_pState->m_pTree;
		for(list<realf****>::iterator I = m_lpCurrentProbablities.begin();
				I != m_lpCurrentProbablities.end(); ++I)
		{
			LC::deallocateMemory(*I);
		}
		m_pState->m_pTree = m_pProposalTree;
		m_pState->m_totalLogLikelihood += (
				(m_proposalTreeLogLikelihood + m_proposalRateLogLikelihood)-
				(m_pState->m_treeLogLikelihood + m_currentRateLogLikelihood));
		m_pState->m_treeLogLikelihood = m_proposalTreeLogLikelihood;
		if(m_loci == -1)
		{
			m_pState->m_globalRateLoglikelihood = m_proposalRateLogLikelihood;
		}
		else
		{
			m_pState->m_vLocalRateLoglikelihood[m_loci] =
					m_proposalRateLogLikelihood;
		}
	}

	void UlnChangeSingleRate::reject()
	{
		delete m_pProposalTree;
		for(list<realf****>::iterator I = m_lpProposalProbablities.begin();
				I != m_lpProposalProbablities.end(); ++I)
		{
			TreeTime_Core::LikelihoodCalculations::deallocateMemory(*I);
		}
	}

	int UlnChangeSingleRate::getId () const
	{
		return 22;
	}

}//TreeTime_RateChangeModel
