/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <gsl/gsl_sf_gamma.h>
#include "Changes/CppChangeBirth.hpp"
#include "Changes/CppChangeDeath.hpp"
#include "Changes/CppChangeRateDistribution.hpp"
#include "Changes/CppChangeRateIntensity.hpp"
#include "Changes/CppChangeRateLocation.hpp"
#include "Changes/CppChangeRateValue.hpp"
#include "../TreeTime_MolecularTree/Tree.hpp"
#include "../TreeTime_MolecularTree/iterator.hpp"
#include "../TreeTime_Core/RandomNumberGenerator.hpp"

namespace TreeTime_RateChangeModel
{
	template<class STATE>
	inline double CompoundPoissonProcess<STATE>::getLambda() const
	{
		return m_lambda;
	}

	template<class STATE>
	inline void CompoundPoissonProcess<STATE>::setLambda(double p_lambda)
	{
		m_lambda = p_lambda;
	}

	template<class STATE>
	inline double CompoundPoissonProcess<STATE>::getAlpha() const
	{
		return m_alpha;
	}

	template<class STATE>
	inline void CompoundPoissonProcess<STATE>::setAlpha(double p_alpha)
	{
		m_alpha = p_alpha;
		setBeta(p_alpha);
	}

	template<class STATE>
	inline double CompoundPoissonProcess<STATE>::getBeta() const
	{
		return m_beta;
	}

	template<class STATE>
	inline int CompoundPoissonProcess<STATE>::getNrChangePoints() const
	{
		return m_nChangePoints;
	}

	template<class STATE>
	inline void CompoundPoissonProcess<STATE>::setNrChangePoints(
			int p_nChangePoints)
	{
		m_nChangePoints = p_nChangePoints;
	}

	template<class STATE>
	std::ostream& operator<<(std::ostream& p_out,
			const CompoundPoissonProcess<STATE>& p_RCM)
	{
		p_RCM.operator<<(p_out);
		return p_out;
	}

	template<class STATE>
	void CompoundPoissonProcess<STATE>::operator<<(std::ostream& p_out) const
	{
		p_out	<< "CPP"							<< "\t"
				<< m_expectedRateChangeIntensity	<< "\t"
				<< m_lambda							<< "\t" 
				<< m_alpha							<< "\t" 
				<< m_beta							<< "\t"
				<< m_nChangePoints;
	}

	template<class STATE>
	CompoundPoissonProcess<STATE>::CompoundPoissonProcess(double
			p_expectedRateChangeIntensity, double p_lambda, double p_alpha,
			double p_nSpecies, double p_changeCalibrationIntensity,
			double p_changeCalibrationDistribution, double p_proposalFactor,
			int p_maxChangePoints)
	: m_lambda						(p_lambda)
	, m_alpha						(p_alpha)
	, m_beta						(0.0)
	, m_nChangePoints				(0)
	, m_expectedRateChangeIntensity	(p_expectedRateChangeIntensity)
	, m_vvWeigths					()
	, m_changeCalibrationIntensity	(p_changeCalibrationIntensity)
	, m_changeCalibrationDistribution(p_changeCalibrationDistribution)
	{
		setBeta(p_alpha);
		m_vvWeigths = vector<vector<double> >
				(p_maxChangePoints, vector<double>(5, 1.0));
		for(int i = 0; i < p_maxChangePoints; ++i)
		{
			double temp = exp( (p_proposalFactor * (-i)) / sqrt(p_nSpecies) );
			m_vvWeigths[i][0] = (10.0/12.0) *  temp;
			m_vvWeigths[i][1] = (10.0/12.0) * (temp +     (1.0-temp)/3.0);
			m_vvWeigths[i][2] = (10.0/12.0) * (temp + 2.0*(1.0-temp)/3.0);
			m_vvWeigths[i][3] = (10.0/12.0);
			m_vvWeigths[i][4] = (11.0/12.0);
		}
		m_vvWeigths[p_maxChangePoints-1][0] = 0;
	}

	template<class STATE>
	CompoundPoissonProcess<STATE>::CompoundPoissonProcess(
			const CompoundPoissonProcess<STATE>& p_Rhs)
	: m_lambda						(p_Rhs.getLambda())
	, m_alpha						(p_Rhs.getAlpha())
	, m_beta						(p_Rhs.getBeta())
	, m_nChangePoints				(p_Rhs.m_nChangePoints)
	, m_expectedRateChangeIntensity	(p_Rhs.m_expectedRateChangeIntensity)
	, m_vvWeigths					(p_Rhs.m_vvWeigths)
	, m_changeCalibrationIntensity	(p_Rhs.m_changeCalibrationIntensity)
	, m_changeCalibrationDistribution(p_Rhs.m_changeCalibrationDistribution)
	{
	}

	template<class STATE>
	CompoundPoissonProcess<STATE>::~CompoundPoissonProcess()
	{
	}

	template<class STATE>
	CompoundPoissonProcess<STATE>& CompoundPoissonProcess<STATE>::operator=
			(const CompoundPoissonProcess<STATE>& p_Rhs)
	{
		if(this != &p_Rhs)
		{
			m_lambda 					= p_Rhs.getLambda();
			m_alpha  					= p_Rhs.getAlpha();
			m_beta   					= p_Rhs.getBeta();
			m_nChangePoints				= p_Rhs.m_nChangePoints;
			m_expectedRateChangeIntensity= p_Rhs.m_expectedRateChangeIntensity;
			m_vvWeigths					= p_Rhs.m_vvWeigths;
			m_changeCalibrationIntensity=p_Rhs.m_changeCalibrationIntensity;
			m_changeCalibrationDistribution =
					p_Rhs.m_changeCalibrationDistribution;
		}
		return *this;
	}

	template<class STATE>
	void CompoundPoissonProcess<STATE>::setBeta(double p_alpha)
	{
		const double delta = 1.0E-5;
		try
		{
			m_beta = exp( (gsl_sf_lngamma(p_alpha+delta)-gsl_sf_lngamma(p_alpha))/
					delta );
		}
		catch(...)
		{
			std::cout<< "Exception occured in function: CompoundPoissonProcess<STATE>::setBeta(...)" << " ";
			std::cout<< p_alpha << std::endl;
		}
	}
	
	template<class STATE>
	CompoundPoissonProcess<STATE>* CompoundPoissonProcess<STATE>::clone() const
	{
		CompoundPoissonProcess<STATE>* pTemp = new CompoundPoissonProcess<STATE>
				(*this);
		return pTemp;
	}

	template<class STATE>
	TreeTime_Mcmcmc::Change* CompoundPoissonProcess<STATE>::getChange(
			STATE* p_pState, int p_loci, double p_runiform) const
	{
		if(p_runiform < m_vvWeigths[m_nChangePoints][0])
		{//cout << "birth" << endl;
			double ratio = (m_vvWeigths[m_nChangePoints+1][1] -
					m_vvWeigths[m_nChangePoints+1][0]) /
					m_vvWeigths[m_nChangePoints][0];
			return new CppChangeBirth(p_pState, ratio, p_loci);
		}
		else if(p_runiform < m_vvWeigths[m_nChangePoints][1])
		{//cout << "death" << endl;
			double ratio = m_vvWeigths[m_nChangePoints-1][0] /
						(m_vvWeigths[m_nChangePoints][1] -
						m_vvWeigths[m_nChangePoints][0]);
			return new CppChangeDeath(p_pState, ratio, p_loci);
		}
		else if(p_runiform < m_vvWeigths[m_nChangePoints][2])
		{//cout << "loci" << endl;
			return new CppChangeRateLocation(p_pState, 1.0, p_loci);
		}
		else if(p_runiform < m_vvWeigths[m_nChangePoints][3])
		{//cout << "value" << endl;
			return new CppChangeRateValue(p_pState, 1.0, p_loci);
		}
		else if(p_runiform < m_vvWeigths[m_nChangePoints][4])
		{//cout << "inten" << endl;
			return new CppChangeRateIntensity(p_pState, 1.0, p_loci,
					1.0/m_expectedRateChangeIntensity,
					m_changeCalibrationIntensity);
		}
		else
		{//cout << "dist" << endl;
			return new CppChangeRateDistribution(p_pState, 1.0, p_loci,
					m_changeCalibrationDistribution);
		}
	}

	template<class STATE>
	double CompoundPoissonProcess<STATE>::getLogDensity(STATE* p_pState,
			int p_loci) const
	{
		double treeLength = getTreeLength(p_pState);
		treeLength /= p_pState->m_pTree->getRoot()->m_timepoint;
		if(m_nChangePoints < 1)
		{
			return -treeLength * m_lambda;
		}
		else
		{
			double tempLL = 0;
			TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> I =
					p_pState->m_pTree->getRoot();
			stack<TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> >
					sIterator;
			sIterator.push(I);
			while(!sIterator.empty())
			{
				I = sIterator.top();
				sIterator.pop();
				list<pair<double,double> >& cp = I->m_mlChangepoints[p_loci];
				for(list<pair<double,double> >::iterator mI = ++cp.begin();
						mI != --cp.end(); ++mI)
				{
					tempLL += log(TreeTime_Core::RandomNumberGenerator::getRNG().
							getGammaDensity(mI->second,m_alpha,m_beta));
				}
				if(I.hasChildNodes())
				{
					sIterator.push(I.getFirstChild());
					sIterator.push(I.getLastChild());
				}
			}
			return tempLL + log( pow(1.0/treeLength, m_nChangePoints)) +
					log(TreeTime_Core::RandomNumberGenerator::getRNG().
					getPoissonDensity(m_nChangePoints,treeLength*m_lambda));
		}
	}

	template<class STATE>
	void CompoundPoissonProcess<STATE>::initRateChanges(STATE* p_pState,
			int p_loci) const
	{
			initRateChanges(p_pState->m_pTree->getRoot(), p_loci);
	}

	template<class STATE>
	void CompoundPoissonProcess<STATE>::initRateChanges(TreeTime_MolecularTree::
			iterator<TreeTime_Core::NodeData> p_I, int p_loci) const
	{
		p_I->m_mlChangepoints[p_loci].push_back(pair<double,double>(0.0,1.0));
		p_I->m_mlChangepoints[p_loci].push_back(pair<double,double>(
				p_I->m_length,1.0));
		if(p_I.hasChildNodes())
		{
			initRateChanges(p_I.getFirstChild(), p_loci);
			initRateChanges(p_I.getLastChild(), p_loci);
		}
	}

	template<class STATE>
	int CompoundPoissonProcess<STATE>::getId() const
	{
		return 1;
	}

	template<class STATE>
	double CompoundPoissonProcess<STATE>::getTreeLength(STATE* p_pState) const
	{
		double treeLength = 0;
		TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> I =
				p_pState->m_pTree->getRoot();
		stack<TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> >
				sIterator;
		sIterator.push(I);
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			treeLength += I->m_length;
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		return treeLength;
	}


}//TreeTime_RateChangeModel
