/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef COMPOUNDPOISSONPROCESS_HPP_
#define COMPOUNDPOISSONPROCESS_HPP_

#include "Settings.hpp"
#include "RateChangeModel.hpp"
#include "../TreeTime_MolecularTree/iterator.hpp"

namespace TreeTime_RateChangeModel
{

	template<class STATE>
	class CompoundPoissonProcess;

	template<class STATE>
	std::ostream& operator<<(std::ostream&,
			const CompoundPoissonProcess<STATE>&);

	template<class STATE>
	class CompoundPoissonProcess : public RateChangeModel<STATE>
	{

		friend ostream& operator<< <STATE>(ostream&,
				const CompoundPoissonProcess<STATE>&);

		public:

			CompoundPoissonProcess(double p_expectedRateChangeIntensity,
					double p_lambda, double p_alpha, double p_nSpecies,
					double p_changeCalibrationIntensity,
					double p_changeCalibrationDistribution,
					double p_proposalFactor, int p_maxChangePoints = 100);
			virtual ~CompoundPoissonProcess();
			CompoundPoissonProcess(const CompoundPoissonProcess& p_Rhs);
			CompoundPoissonProcess& operator=
					(const CompoundPoissonProcess& p_Rhs);

			inline double	getLambda() const;
			inline void		setLambda(double p_lambda);
			inline double	getAlpha() const;
			inline void		setAlpha(double p_alpha);
			inline double	getBeta() const;		
			inline int		getNrChangePoints() const;
			inline void		setNrChangePoints(int p_nChangePoints);
//			inline double	getExpectedRateChangeIntensity() const;
//			inline void		setExpectedRateChangeIntensity(double p_ExpRCInt);



			virtual CompoundPoissonProcess<STATE>* clone() const;
			virtual TreeTime_Mcmcmc::Change* getChange(STATE* p_pState,
					int p_loci, double p_runiform) const;
			virtual double getLogDensity(STATE* p_pState, int p_loci) const;
			virtual void initRateChanges(STATE* p_pState, int p_loci) const;
			virtual int getId() const;

		private:

			void setBeta(double p_alpha);

			void initRateChanges(TreeTime_MolecularTree::iterator
					<TreeTime_Core::NodeData> p_I, int p_loci) const;

			/**
			 * Output Operator, which is called by the namespace level output
			 * operator, and is deduced to the concrete classes.
			 */
			 virtual void operator<<(ostream&) const;
			 
			 double getTreeLength(STATE* p_pState) const;
			 
			 //virtual void	initTreeData(STATE* p_pState) const;

		private:
			// exponential prior with parameter = 1 on m_lambda:
			//Events occur according to a poisson process with rate lambda =\mu.
			//p(k) = {\mu^k \over k!} \exp(-\mu), 1/\mu ist mittelwert, k>=0
			//\mu is intensity
			// Uniform on [0,max_m_alpha].
			double	m_lambda;
			//alpha is a parametrer of a gamma distributed random variable
			//(shape),
			//beta can be calculated as: beta = exp(Gamma'(alpha)/Gamma(alpha))
			// kleiner als 0.1 macht kein Sinn!
			double	m_alpha;
			// beta is the other parametrer of the gamma distributed random
			// variable (scale bzw. 1/rate),
			// taken value such that the Produkt of Modulatins equals 1 in
			// expectation.
			//p(x) dx = {1 \over \Gamma(a) b^a} x^{a-1} e^{-x/b} dx
			//gsl_ran_gamma_pdf (double x, double a, double b)
			// Zufallszahl: gsl_ran_gamma(const gsl_rng * r, double a, double b)
			double	m_beta;
			// nr of ChangePoints in the Modell
			int		m_nChangePoints;
			// Exponential Prior to Lambda
			const double	m_expectedRateChangeIntensity;

			vector<vector<double> >	m_vvWeigths;
			
			const double	m_changeCalibrationIntensity;
			const double	m_changeCalibrationDistribution;
			
			//int m_loci;

	};//CompoundPoissonProcess

}//TreeTime_RateChangeModel

#include "CompoundPoissonProcess.cxx"

#endif /*COMPOUNDPOISSONPROCESS_HPP_*/
