/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "Changes/DmChangeAllRates.hpp"
#include "Changes/DmChangeTwoRates.hpp"
#include "Changes/DmChangeVariance.hpp"
#include "../TreeTime_MolecularTree/Tree.hpp"
#include "../TreeTime_MolecularTree/iterator.hpp"
#include "../TreeTime_Core/RandomNumberGenerator.hpp"

namespace TreeTime_RateChangeModel
{

	template<class STATE>
	inline int DirichletModel<STATE>::getNRateMultipliers() const
	{
		return m_nRates;
	}

	template<class STATE>
	inline double DirichletModel<STATE>::getVariance() const
	{
		return m_variance;
	}

	template<class STATE>
	inline void DirichletModel<STATE>::setVariance(double p_variance)
	{
		m_variance = p_variance;
	}

	template<class STATE>
	std::ostream& operator<<(std::ostream& p_out,
			const DirichletModel<STATE>& p_RCM)
	{
		p_RCM.operator<<(p_out);
		return p_out;
	}

	template<class STATE>
	void DirichletModel<STATE>::operator<<(std::ostream& p_out) const
	{
		p_out	<< "DM" 						<< "\t" 
				<< m_expectedVariance			<< "\t"
				<< m_variance;
	}

	template<class STATE>
	DirichletModel<STATE>::DirichletModel(double p_expectedVariance, double
			p_initialVariance, int p_nSpecies, double
			p_changeCalibrationVariance, double p_changeCalibrationAll,
			double p_changeCalibrationTwo)
	: m_variance					(p_initialVariance)
	, m_nRates						(p_nSpecies+p_nSpecies-2)
	, m_expectedVariance			(p_expectedVariance)
	, m_changeCalibrationVariance	(p_changeCalibrationVariance)
	, m_changeCalibrationAll		(p_changeCalibrationAll)
	, m_changeCalibrationTwo		(p_changeCalibrationTwo)
	{
	}

	template<class STATE>
	DirichletModel<STATE>::DirichletModel(const DirichletModel<STATE>& p_Rhs)
	: m_variance					(p_Rhs.m_variance)
	, m_nRates						(p_Rhs.m_nRates)
	, m_expectedVariance			(p_Rhs.m_expectedVariance)
	, m_changeCalibrationVariance	(p_Rhs.m_changeCalibrationVariance)
	, m_changeCalibrationAll		(p_Rhs.m_changeCalibrationAll)
	, m_changeCalibrationTwo		(p_Rhs.m_changeCalibrationTwo)
	{
	}

	template<class STATE>
	DirichletModel<STATE>::~DirichletModel()
	{
	}

	template<class STATE>
	DirichletModel<STATE>& DirichletModel<STATE>::operator=
			(const DirichletModel<STATE>& p_Rhs)
	{
		if(this != &p_Rhs)
		{
			m_variance				= p_Rhs.m_variance;
			m_nRates				= p_Rhs.m_nRates;
			m_expectedVariance		= p_Rhs.m_expectedVariance;
			m_changeCalibrationVariance= p_Rhs.m_changeCalibrationVariance;
			m_changeCalibrationAll	= p_Rhs.m_changeCalibrationAll;
			m_changeCalibrationTwo	= p_Rhs.m_changeCalibrationTwo;
		}
		return *this;
	}
	
	template<class STATE>
	DirichletModel<STATE>* DirichletModel<STATE>::clone() const
	{
		DirichletModel<STATE>* pTemp = new DirichletModel<STATE>(*this);
		return pTemp;
	}

	template<class STATE>
	TreeTime_Mcmcmc::Change* DirichletModel<STATE>::getChange(
			STATE* p_pState, int p_loci, double p_runiform) const
	{
		if(p_runiform < 0.4)
		{
			return new DmChangeAllRates(p_pState, 1.0, p_loci, m_nRates,
					m_changeCalibrationAll);
		}
		else if(p_runiform < 0.8)
		{
			return new DmChangeTwoRates(p_pState, 1.0, p_loci,
					m_changeCalibrationTwo);
		}
		else
		{
			return new DmChangeVariance(p_pState, 1.0, p_loci, m_nRates,
					m_expectedVariance, m_changeCalibrationVariance);
		}
	}

	template<class STATE>
	double DirichletModel<STATE>::getLogDensity(STATE* p_pState,
			int p_loci) const
	{
		double val = (static_cast<double>(m_nRates-1) -m_variance)/
				(static_cast<double>(m_nRates) * m_variance);
		double* meanValues    = new double[m_nRates];
		double* currentValues = new double[m_nRates];
		int counter = 0;
		TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> I =
				p_pState->m_pTree->getRoot();
		stack<TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> >
				sIterator;
		sIterator.push(I.getFirstChild());
		sIterator.push(I.getLastChild());
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(p_loci == -1)
			{
				currentValues[counter] = I->m_globalRateMultiplier/
						static_cast<double>(m_nRates);
			}
			else
			{
				currentValues[counter] = I->m_vLocalRateMultiplier[p_loci]/
						static_cast<double>(m_nRates);
			}
			meanValues[counter] = val;
			++counter;
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		double logDensity =  TreeTime_Core::RandomNumberGenerator::getRNG().
				getLogDirichletDensity(meanValues, currentValues, m_nRates);
		delete[] meanValues;
		delete[] currentValues;
		return logDensity;
	}

	template<class STATE>
	void DirichletModel<STATE>::initRateChanges(STATE* p_pState, int p_loci)
			const
	{
	}

	template<class STATE>
	int DirichletModel<STATE>::getId() const
	{
		return 2;
	}

}//TreeTime_RateChangeModel
