/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef TT_RC_RATECHANGEMODEL_HPP_
#define TT_RC_RATECHANGEMODEL_HPP_

#include "Settings.hpp"
#include "Random.hpp"
#include "../TreeTime_Mcmcmc/Change.hpp"

namespace TreeTime_RateChangeModel
{
	template<class STATE>
	class RateChangeModel;//Forward Declaration

	/**
	 * PrintStream for Output of RateChangeModel,
	 * implemented as a namespace level function.
	 * @param ostream&
	 * @param const DnaEvolutionModel&
	 * @return ostream&
	 */
	template<class STATE>
	ostream& operator<< (ostream&, const RateChangeModel<STATE>&);

	/// Represents the abstract Interface for a RateChangeModel
	/**
	 * The Evolution Model consists of a Stationary Distribution
	 */
	template<class STATE>
	class RateChangeModel
	{

		/**
		 * Declaration of friend ouput operator
		 */
		friend ostream& operator<< <STATE>(ostream&, const 
				RateChangeModel<STATE>& p_rhs);

		public:

			/// Static Factory Method for generating a DnaEvolutionModel
			/**
			 * This Factory Method expects a String Representation of the Dna
			 * Evolution Model, and returns a Pointer to a new allocated Dna
			 * Evolution Model
			 * @param string
			 * @return DnaEvolutionModel* 
			 */
			static RateChangeModel<STATE>* newRateChangeModel(string
					p_stringRepresentation);

			/**
			 * Destructor of DnaEvolutionModel
			 */
			virtual ~RateChangeModel();

			/**
			 * Returns a Clone of the Dna Evolution Model
			 * @return DnaEvolutionModel* 
			 */
			virtual RateChangeModel<STATE>* clone() const = 0;

			/**
			 */
			virtual TreeTime_Mcmcmc::Change* getChange(STATE* p_pTree,
					int p_loci, double p_runiform) const= 0;

			/**
			 * Sets the Random Generator for changing the Model Parametrers
			 */
			virtual double getLogDensity(STATE* p_pTree, int p_loci= -1) const
					= 0;

			/**
			 * Sets the Random Generator for changing the Model Parametrers
			 */
			static inline void setRandom(Random* p_pRandom);

			virtual void initRateChanges(STATE* p_pState, int p_loci) const = 0;

			virtual int getId() const = 0;

		protected:

			/**
			 * Constructor
			 */
			RateChangeModel();

			/**
			 * Copy Constructor
			 */
			RateChangeModel(const RateChangeModel<STATE>& p_Rhs);

		private:

			/**
			 * Private Copy Constructor, not implemented
			 */
			RateChangeModel<STATE>& operator=(const RateChangeModel<STATE>& 
					p_rhs);

			/**
			 * Output Operator, which is called by the namespace level output
			 * operator, and is deduced to the concrete classes.
			 */
			 virtual void operator<<(ostream&) const = 0;

		protected:

			/** Pointer to Random Generator and Distribution Density Interface */
			static Random*		s_pRandom;

	};//RateChangeModel

}//TreeTime_RateChangeModel

#include "RateChangeModel.cxx"

#endif /*TT_RC_RATECHANGEMODEL_HPP_*/
