/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "Changes/UexChangeAllRates.hpp"
#include "Changes/UexChangeSingleRate.hpp"
#include "../TreeTime_MolecularTree/Tree.hpp"
#include "../TreeTime_MolecularTree/iterator.hpp"
#include "../TreeTime_Core/RandomNumberGenerator.hpp"

namespace TreeTime_RateChangeModel
{

	template<class STATE>
	std::ostream& operator<<(std::ostream& p_out,
			const UncorrelatedExponential<STATE>& p_RCM)
	{
		p_RCM.operator<<(p_out);
		return p_out;
	}

	template<class STATE>
	void UncorrelatedExponential<STATE>::operator<<(std::ostream& p_out) const
	{
		p_out	<< "UEX";
	}

	template<class STATE>
	UncorrelatedExponential<STATE>::UncorrelatedExponential(int p_nSpecies,
			double p_changeCalibrationAll, double p_changeCalibrationSingle)
	: m_nRates					(p_nSpecies+p_nSpecies-2)
	, m_changeCalibrationAll	(p_changeCalibrationAll)
	, m_changeCalibrationSingle	(p_changeCalibrationSingle)
	{
	}

	template<class STATE>
	UncorrelatedExponential<STATE>::UncorrelatedExponential(const
			UncorrelatedExponential<STATE>& p_Rhs)
	: m_nRates					(p_Rhs.m_nRates)
	, m_changeCalibrationAll	(p_Rhs.m_changeCalibrationAll)
	, m_changeCalibrationSingle	(p_Rhs.m_changeCalibrationSingle)
	{
	}

	template<class STATE>
	UncorrelatedExponential<STATE>::~UncorrelatedExponential()
	{
	}

	template<class STATE>
	UncorrelatedExponential<STATE>& UncorrelatedExponential<STATE>::operator=
			(const UncorrelatedExponential<STATE>& p_Rhs)
	{
		m_nRates					= p_Rhs.m_nRates;
		m_changeCalibrationAll		= p_Rhs.m_changeCalibrationAll;
		m_changeCalibrationSingle	= p_Rhs.m_changeCalibrationSingle;
		return *this;
	}
	
	template<class STATE>
	UncorrelatedExponential<STATE>* UncorrelatedExponential<STATE>::clone()
			const
	{
		UncorrelatedExponential<STATE>* pTemp = new UncorrelatedExponential
				<STATE>(*this);
		return pTemp;
	}

	template<class STATE>
	TreeTime_Mcmcmc::Change* UncorrelatedExponential<STATE>::getChange(
			STATE* p_pState, int p_loci, double p_runiform) const
	{
		if(p_runiform < 0.5)
		{
			return new UexChangeAllRates(p_pState, 1.0, p_loci,
					m_changeCalibrationAll);
		}
		else
		{
			return new UexChangeSingleRate(p_pState, 1.0, p_loci, m_nRates,
					m_changeCalibrationSingle);
		}
	}

	template<class STATE>
	double UncorrelatedExponential<STATE>::getLogDensity(STATE* p_pState,
			int p_loci) const
	{
		double logDensity = 0;
		double* pValue;
		TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> I =
				p_pState->m_pTree->getRoot();
		stack<TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> >
				sIterator;
		sIterator.push(I.getFirstChild());
		sIterator.push(I.getLastChild());
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(p_loci == -1)
			{
				pValue = &(I->m_globalRateMultiplier);
			}
			else
			{
				pValue = &(I->m_vLocalRateMultiplier[p_loci]);
			}
			logDensity += log(TreeTime_Core::RandomNumberGenerator::getRNG()
					.getExponentialDensity(*pValue, 1.0));
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		return logDensity;
	}

	template<class STATE>
	void UncorrelatedExponential<STATE>::initRateChanges(STATE* p_pState,
			int p_loci) const
	{
	}

	template<class STATE>
	int UncorrelatedExponential<STATE>::getId() const
	{
		return 3;
	}

}//TreeTime_RateChangeModel
