/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "Changes/UlnChangeAllRates.hpp"
#include "Changes/UlnChangeSingleRate.hpp"
#include "Changes/UlnChangeVariance.hpp"
#include "../TreeTime_MolecularTree/Tree.hpp"
#include "../TreeTime_MolecularTree/iterator.hpp"
#include "../TreeTime_Core/RandomNumberGenerator.hpp"

namespace TreeTime_RateChangeModel
{

	template<class STATE>
	std::ostream& operator<<(std::ostream& p_out,
			const UncorrelatedLognormal<STATE>& p_RCM)
	{
		p_RCM.operator<<(p_out);
		return p_out;
	}

	template<class STATE>
	void UncorrelatedLognormal<STATE>::operator<<(std::ostream& p_out) const
	{
		p_out	<< "ULN"					<< "\t"
				<<	m_priorExpectedVariance	<< "\t"
				<<	m_variance;
	}

	template<class STATE>
	UncorrelatedLognormal<STATE>::UncorrelatedLognormal(double
			p_expectedVariance, double p_initialVariance, int p_nSpecies,
			double p_changeCalibrationVariance, double p_changeCalibrationAll,
			double p_changeCalibrationSingle)
	: m_nRates					(p_nSpecies+p_nSpecies-2)
	, m_changeCalibrationVariance (p_changeCalibrationVariance)
	, m_changeCalibrationAll	(p_changeCalibrationAll)
	, m_changeCalibrationSingle	(p_changeCalibrationSingle)
	, m_priorExpectedVariance	(p_expectedVariance)
	, m_variance				(p_initialVariance)
	, m_normalVariance			(log(m_variance+1))
	, m_normalMean				(-m_normalVariance/2.0)
	, m_normalDeviation			(sqrt(m_normalVariance))
	{
	}

	template<class STATE>
	UncorrelatedLognormal<STATE>::UncorrelatedLognormal(const
			UncorrelatedLognormal<STATE>& p_Rhs)
	: m_nRates					(p_Rhs.m_nRates)
	, m_changeCalibrationVariance (p_Rhs.m_changeCalibrationVariance)
	, m_changeCalibrationAll	(p_Rhs.m_changeCalibrationAll)
	, m_changeCalibrationSingle	(p_Rhs.m_changeCalibrationSingle)
	, m_priorExpectedVariance	(p_Rhs.m_priorExpectedVariance)
	, m_variance				(p_Rhs.m_variance)
	, m_normalVariance			(p_Rhs.m_normalVariance)
	, m_normalMean				(p_Rhs.m_normalMean)
	, m_normalDeviation			(p_Rhs.m_normalDeviation)
	{
	}

	template<class STATE>
	UncorrelatedLognormal<STATE>::~UncorrelatedLognormal()
	{
	}

	template<class STATE>
	UncorrelatedLognormal<STATE>& UncorrelatedLognormal<STATE>::operator=
			(const UncorrelatedLognormal<STATE>& p_Rhs)
	{
		m_nRates					= p_Rhs.m_nRates;
		m_changeCalibrationVariance = p_Rhs.m_changeCalibrationVariance;
		m_changeCalibrationAll		= p_Rhs.m_changeCalibrationAll;
		m_changeCalibrationSingle	= p_Rhs.m_changeCalibrationSingle;
		m_priorExpectedVariance		= p_Rhs.m_priorExpectedVariance;
		m_variance					= p_Rhs.m_variance;
		m_normalVariance			= p_Rhs.m_normalVariance;
		m_normalMean				= p_Rhs.m_normalMean;
		m_normalDeviation			= p_Rhs.m_normalDeviation;
		return *this;
	}
	
	template<class STATE>
	UncorrelatedLognormal<STATE>* UncorrelatedLognormal<STATE>::clone()
			const
	{
		UncorrelatedLognormal<STATE>* pTemp = new UncorrelatedLognormal
				<STATE>(*this);
		return pTemp;
	}

	template<class STATE>
	TreeTime_Mcmcmc::Change* UncorrelatedLognormal<STATE>::getChange(
			STATE* p_pState, int p_loci, double p_runiform) const
	{
		if(p_runiform < 0.3)
		{
			return new UlnChangeAllRates(p_pState, 1.0, p_loci,
					m_changeCalibrationAll);
		}
		else if(p_runiform < 0.8)
		{
			return new UlnChangeSingleRate(p_pState, 1.0, p_loci, m_nRates,
					m_changeCalibrationSingle);
		}
		else
		{
			return new UlnChangeVariance(p_pState, 1.0, p_loci,
					m_priorExpectedVariance, m_changeCalibrationVariance);
		}
	}

	template<class STATE>
	double UncorrelatedLognormal<STATE>::getLogDensity(STATE* p_pState,
			int p_loci) const
	{
		double logDensity = 0;
		double* pValue;
		TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> I =
				p_pState->m_pTree->getRoot();
		stack<TreeTime_MolecularTree::iterator<TreeTime_Core::NodeData> >
				sIterator;
		sIterator.push(I.getFirstChild());
		sIterator.push(I.getLastChild());
		while(!sIterator.empty())
		{
			I = sIterator.top();
			sIterator.pop();
			if(p_loci == -1)
			{
				pValue = &(I->m_globalRateMultiplier);
			}
			else
			{
				pValue = &(I->m_vLocalRateMultiplier[p_loci]);
			}
			logDensity += log(TreeTime_Core::RandomNumberGenerator::getRNG().
					getLognormalDensity(*pValue, m_normalMean,
					m_normalDeviation));
			if(I.hasChildNodes())
			{
				sIterator.push(I.getFirstChild());
				sIterator.push(I.getLastChild());
			}
		}
		return logDensity;
	}

	template<class STATE>
	void UncorrelatedLognormal<STATE>::initRateChanges(STATE* p_pState,
			int p_loci) const
	{
	}

	template<class STATE>
	int UncorrelatedLognormal<STATE>::getId() const
	{
		return 4;
	}

	template<class STATE>
	double UncorrelatedLognormal<STATE>::getNormalMean() const
	{
		return m_normalMean;
	}

	template<class STATE>
	double UncorrelatedLognormal<STATE>::getNormalDeviation() const
	{
		return m_normalDeviation;
	}

	template<class STATE>
	inline double UncorrelatedLognormal<STATE>::getVariance() const
	{
		return m_variance;
	}

	template<class STATE>
	inline void UncorrelatedLognormal<STATE>::setVariance(double p_variance)
	{
		m_variance			= p_variance;
		m_normalVariance	= log(m_variance+1);
		m_normalMean		= -m_normalVariance/2.0;
		m_normalDeviation	= sqrt(m_normalVariance);
	}

}//TreeTime_RateChangeModel
