/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "MemoryAllocator.hpp"

namespace TreeTime_Utilities
{

	MemoryAllocator::MemoryAllocator(int p_nLoci, int* p_nSites, int* p_nRates,
			int p_nStates)
	: m_nLoci(p_nLoci)
	, m_nSites(initConstArray(p_nLoci,p_nSites))
	, m_nRates(initConstArray(p_nLoci,p_nRates))
	, m_nStates(p_nStates)
	, m_pointerSize(sizeof(realf*))
	, m_realSize(sizeof(realf))
	, m_memorySpace(initCalcMemorySpace())
	{
	}

	MemoryAllocator::~MemoryAllocator()
	{
		delete[] m_nSites;
		delete[] m_nRates;
		for(set<realf****>::iterator I = m_UsedMemory.begin();
				I != m_UsedMemory.end(); ++I)
		{
			delete[] *I;// no cast to char[] necessary
		}
		while(!m_UnusedMemory.empty())
		{
			delete[] m_UnusedMemory.top();// no cast to char[] necessary
			m_UnusedMemory.pop();
		}
	}

	const int* const MemoryAllocator::initConstArray(const int p_nr,
			const int* const p_pInts)
	{
		int* const pInts = new int[p_nr];
		for(int i=0; i<p_nr; ++i)
			pInts[i] = p_pInts[i];
		return pInts;
	}

	int MemoryAllocator::initCalcMemorySpace()
	{
		int sum_pointer = m_nLoci;
		int sum_values  = 0;
		for(int i=0; i<m_nLoci; ++i)
		{
			sum_pointer += (m_nRates[i]);
			sum_pointer	+= (m_nRates[i] * m_nSites[i]);
			sum_values	+= (m_nRates[i] * m_nSites[i] * m_nStates);
		}
		return (sum_pointer * m_pointerSize) + (sum_values * m_realSize);
	}

	realf**** MemoryAllocator::allocateMemory()
	{
		if(m_UnusedMemory.empty())
		{
			// Allocate new Memory:
			char* mem = new char[ sizeof(char) * m_memorySpace ];
			int pAdd = 0;
			// Placement new in Memory:
			realf**** probs = new (mem+pAdd) realf***[m_nLoci];
			pAdd += (m_nLoci*m_pointerSize);
			for(int l=0; l<m_nLoci ; ++l)
			{
				probs[l] = new (mem+pAdd) realf**[m_nRates[l]];
				pAdd  += (m_nRates[l]*m_pointerSize);
				for(int r=0; r<m_nRates[l]; ++r)
				{
					probs[l][r] = new (mem+pAdd) realf*[m_nSites[l]];
					pAdd     += (m_nSites[l]*m_pointerSize);
					for(int s=0; s<m_nSites[l]; ++s)
					{
						probs[l][r][s] = new (mem+pAdd) realf[m_nStates];//alloc
						pAdd += (m_nStates*m_realSize);
					}
				}
			}
			// Insert in Used Memory
			m_UsedMemory.insert(probs);
			return probs;
		}
		else
		{
			realf**** probs = m_UnusedMemory.top();
			m_UnusedMemory.pop();
			m_UsedMemory.insert(probs);
			return probs;
		}
	}

	realf**** MemoryAllocator::allocateMemory(realf**** p_pMemory)
	{
		if(m_UnusedMemory.empty())
		{
			// Allocate new Memory:
			char* mem = new char[ sizeof(char) * m_memorySpace ];
			int pAdd = 0;
			// Placement new in Memory:
			realf**** probs = new (mem+pAdd) realf***[m_nLoci];
			pAdd += (m_nLoci*m_pointerSize);
			for(int l=0; l<m_nLoci ; ++l)
			{
				probs[l] = new (mem+pAdd) realf**[m_nRates[l]];
				pAdd  += (m_nRates[l]*m_pointerSize);
				for(int r=0; r<m_nRates[l]; ++r)
				{
					probs[l][r] = new (mem+pAdd) realf*[m_nSites[l]];
					pAdd     += (m_nSites[l]*m_pointerSize);
					for(int s=0; s<m_nSites[l]; ++s)
					{
						probs[l][r][s] = new (mem+pAdd) realf[m_nStates];//alloc
						for(int n=0; n<m_nStates; ++n) //value assignment
							probs[l][r][s][n] = p_pMemory[l][r][s][n];
						pAdd += (m_nStates*m_realSize);
					}
				}
			}
			// Insert in Used Memory
			m_UsedMemory.insert(probs);
			return probs;
		}
		else
		{
			realf**** probs = m_UnusedMemory.top();
			m_UnusedMemory.pop();
			m_UsedMemory.insert(probs);
			for(int l=0; l<m_nLoci; ++l)
			{
				for(int r=0; r<m_nRates[l]; ++r)
				{
					for(int v=0; v<m_nSites[l]; ++v)
					{
						for(int n=0; n<m_nStates; ++n)
						{
							probs[l][r][v][n] = p_pMemory[l][r][v][n];
						}
					}
				}
			}
			return probs;
		}
	}

	void MemoryAllocator::freeMemory(realf**** p_pMemory)
	{
		set<realf****>::iterator I = m_UsedMemory.find(p_pMemory);
		if(I != m_UsedMemory.end())
		{
			m_UsedMemory.erase(I);
			m_UnusedMemory.push(p_pMemory);
		}
	}

}//TreeTime_Utilities
