/*
Copyright (C) 2006-2008 Lin Himmelmann

This file is part of TreeTime.
See the NOTICE file distributed with this work for additional
information regarding copyright ownership and licensing.

TreeTime is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

TreeTime is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with TreeTime.  If not, see <http://www.gnu.org/licenses/>.
*/
#ifndef MEMORYALLOCATOR_HPP_
#define MEMORYALLOCATOR_HPP_

#include "Settings.hpp"

namespace TreeTime_Utilities
{

	///MemoryAllocator manages the Memory of molecular Data.
	/**
	 * MemoryAllocator manages the Memory of molecular Data.
	 * The Memory for the molecular Data at each Node of a Tree is managed by
	 * MemoryAllocator. You can demand new Memory, free the Memory and clone
	 * Memory for a givenNode.
	 * @see MemoryAllocator(int p_nDataPartitions, int[] p_nSites,
	 * int[] p_nRates, int p_nStates)
	 */
	class MemoryAllocator
	{
		public:
			/**
			 * The Constructor is given the following information:
			 * @param p_nDataPartitions, the number of different Data
			 * Partitions, which need a different Evolution Modell
			 * @param p_nSites, a Vector of integral Values with length
			 * p_nDataPartitions, gives the Number of Sites for each Data
			 * Partition
			 * @param p_nRates, a Vector of integral Values with length
			 * p_nDataPartitions, gives the Number of Rates in the Evolution
			 * Modell for each Data Partition
			 * @param p_nStates, the number of different States
			 * (for Dna Data: 4, for Protein Data : 20)
			 */
			MemoryAllocator(int p_nDataPartitions, int* p_nSites,
					int* p_nRates, int p_nStates);
			/**
			 * Destructs the MemoryAllocator and all frees the Memory for the
			 * molecular Data.
			 */
			~MemoryAllocator();
			/**
			 * Allocates Memory for molecular Data, and returns a Pointer
			 * @returm Pointer to allocated Memory
			 */
			realf****	allocateMemory();
			/**
			 * Allocates Memory for molecular Data, initialises it with the
			 * molecular Data in the Argument, and returns a Pointer to the
			 * allocated and initialised Memory
			 * @param Pointer to molecular Data which is in the Memory
			 * @returm Pointer to allocated (and initialised) Memory
			 */
			realf****	allocateMemory(realf**** p_pMemory);
			/**
			 * Deallocates Memory for molecular Data
			 * @param Pointer to molecular Data which is in the Memory and
			 * should be deallocated.
			 */
			void		freeMemory(realf**** p_pMemory);

		private:
			/**
			 * Function used to initialise constant Attributes.
			 */
			const int* const initConstArray(const int ,const int* const);
			/**
			 * Function used to initialise constant Attributes.
			 */
			int initCalcMemorySpace();
			/**
			 * Copy Constructor makes no sense, and is not allowed
			 */
			MemoryAllocator(const MemoryAllocator&);
			/**
			 * Copy Constructor makes no sense, and is not allowed
			 */
			MemoryAllocator& operator=(const MemoryAllocator&);
			
		private:
			/** Number of different Data Partitions */
			const int			m_nLoci;
			/** Vector of integral Values with length m_nLoci,
			 * gives the Number of Sites for each Data Partition */
			const int* const	m_nSites;
			/** Vector of integral Values with length m_nLoci,
			 * gives the Number of Rates for each Data Partition */		
			const int* const	m_nRates;
			/** Number of different States (Dna: 4, Protein : 20) */		
			const int			m_nStates;
			/** List of Pointers to used Memory */			
			set<realf****>		m_UsedMemory;
			/** Stack of Pointers to unused Memory */	
			stack<realf****>	m_UnusedMemory;
			/** Size of Pointer to real */
			const int			m_pointerSize;
			/** Size of reak */
			const int			m_realSize;
			/** Size of memory for a 4 dimensional Data Package */
			const int			m_memorySpace;

	};//MemoryAllocator

}//TreeTime_Utilities

#endif /*MEMORYALLOCATOR_HPP_*/
